/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2002 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.
 
    You should have received a copy of the GNU General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to professionally license this library, the author
    can be reached at info@mohawksoft.com
*/ 

// This is a file based serializer plugin. It saves and restores the msessiond 
// state at startup, and shutdown.
// It assume the serializer string is a filename.

#include <stdio.h>
#include "plugapi.h"

MSESSION_EXPORTS *msessiond;

#define SF_VERSION	0x1042
#define STAG_CHK	0xA50000
#define	STAG_SESSION	0xA50001
#define STAG_DATA	0xA50002
#define STAG_NAME	0xA50003
#define BUF_SIZE	8192

#define VALID_TAG(t)	(((t) & STAG_CHK)==STAG_CHK)

// This is written to the file First.
// It gives a version and a time stamp.
typedef struct saveFileHeader
{
	int	version;
	int 	timestamp;
}SAVEHDR;

// This identifies each section of the file.
// The tag member tells what type of info follows
// The cb member says how many bytes follow.
// The param member contains special info as needed.
typedef struct	saveFileTag
{
	int	tag;
	int	cb;
	int 	param;
}SAVETAG;

// A stupid little buffer class maintains malloc'ed 
// memory.
class Buffer
{
	protected:
	int	m_cb;
	public:
	void *data;
	Buffer(int cb = BUF_SIZE);
	~Buffer();
	Boolean setcb(int cb);
};

Buffer::Buffer(int cb)
{
	data = phmalloc(cb);
	m_cb = (data) ? cb : 0;
	((char *)data)[0]=0;
}
Buffer::~Buffer(void)
{
	if(data)
		phmfree(data);
}
Boolean Buffer::setcb(int cb)
{
	if( cb < m_cb)
	{
		((char *)data)[0]=0;
		return TRUE;
	}
	else
	{
		int ncb = cb*2;
		assert(ncb >= cb);
		phmfree(data);
		data = phmalloc(ncb);
		if(data)
		{
			m_cb = ncb;
			((char *)data)[0]=0;
			return TRUE;
		}
	}
	return FALSE;
}

// Sets error contitions.
int fileError(REQB **preq, char *string)
{
	(*preq)->req.stat = REQ_ERR;
	(*preq)->req.param = REQE_UNKNOWN;
	sprintf((*preq)->req.datum, string);
	(*preq)->req.len = strlen((*preq)->req.datum)+1;
	msessiond->errprintf("%s [operation aborted]", (*preq)->req.datum);
	return (*preq)->req.stat;
}

// The PLUGIN_RESTORE routine is called at startup to restore the state of the
// msessiond daemon.
static int PLUGIN_RESTORE(void *p1, void *p2)
{
	MStdioFile mf;
	SAVEHDR	hdr;
	Buffer session;
	Buffer name;
	Buffer value;

	SAVETAG	tag;
	SAVETAG	tagData;
	int count=0;

	REQB **preq = (REQB **) p1;

	if(!mf.open((*preq)->req.datum, "rb"))
		return fileError(preq,"Restore file not found");

	if(mf.read(&hdr, sizeof(hdr)) != sizeof(hdr))
		return fileError(preq,"Could not read file header");

	if(hdr.version != SF_VERSION)
		return fileError(preq,"Wrong version in save file");

	msessiond->errprintf("Restore");

	while(1)
	{
		if(mf.read(&tag, sizeof(tag)) != sizeof(tag))
		{
			msessiond->errprintf("Restored: %d sessions", count);
			break;
		}
		if(!VALID_TAG(tag.tag))
			return fileError(preq, "Corrupt save file");

		if(tag.tag == STAG_SESSION) // New session
		{
			count++;
			char *svalue = NULL;
			session.setcb(tag.cb);
			if(mf.read(session.data, tag.cb) != tag.cb)
				return fileError(preq,"Could not read session");

			if(mf.read(&tagData, sizeof(tagData)) != sizeof(tagData))
				return fileError(preq,"Could not read data tag");
		
			if(tagData.tag != STAG_DATA)
				return fileError(preq,"Error in file, data tag not found");


			value.setcb(tagData.cb);

			if(tagData.cb)
			{
				if(mf.read(value.data, tagData.cb) != tagData.cb)
					return fileError(preq,"Could not read session data");
				svalue = (char *)value.data;
			}
			// Create Session
			msessiond->ExecuteRequest(preq,MREQ_CREATE, (char *)session.data, "",svalue, tag.param);
		}
		else if(tag.tag == STAG_NAME) // New name/value pair
		{
			name.setcb(tag.cb);
			if(mf.read(name.data, tag.cb) != tag.cb)
				return fileError(preq,"Could not read name");

			if(mf.read(&tagData, sizeof(tagData)) != sizeof(tagData))
				return fileError(preq,"Could not read data tag");
		
			if(tagData.tag != STAG_DATA)
				return fileError(preq,"Error in file, value tag not found");

			value.setcb(tagData.cb);

			if(tagData.cb)
				if(mf.read(value.data, tagData.cb) != tagData.cb)
					return fileError(preq,"Could not read session data");
			msessiond->ExecuteRequest(preq, MREQ_SETVAL, 
					(char *)session.data, (char *)name.data, (char *)value.data, 0);
		}
	}
	return (*preq)->req.stat;
}

// Called for each value/pair MNode in a session
// p1 is the SERIALIZE_PARAM structure
// p2 has session name
static Boolean sessionCallBack(MNode *mn, void *p1, void *p2)
{
	SERIALIZE_PARAM *psp = (SERIALIZE_PARAM *) p1;
	MStdioFile *pmf = (MStdioFile *) psp->vp_var1;
	SAVETAG	tag;
	
	char * name = (char *)mn->KeyGet();
	char * value = (char *)mn->DataGet();

	tag.param = 0;

	tag.tag = STAG_NAME;
	tag.cb = strlen(name)+1;
	pmf->write(&tag, sizeof(tag));
	pmf->write(name, tag.cb);

	tag.tag = STAG_DATA;
	tag.cb = strlen(value)+1;
	pmf->write(&tag, sizeof(tag));
	pmf->write(value, tag.cb);

	return TRUE;
}

// The PLUGIN_SERIALIZE routine is called at shutdown, and optionally during
// operation, to save the current state of the msession daemon.
// In the SERIALIZE_PARAM structure:
// n_var1 is the current serialization id
// n_var2 is the previous serialization id
// vp_var1 is the PostgreSQL connection
// vp_var2 is unused.
static int PLUGIN_SERIALIZE(void *p1, void *p2)
{
	SERIALIZE_PARAM *psp = (SERIALIZE_PARAM *) p2;
	REQB **preq = (REQB **) p1;
	MStdioFile *pmf = (MStdioFile *) psp->vp_var1;
	if(psp->cmd == PLUGIN_SERIALIZE_BEGIN)
	{
		char ext[]=".bak";
		SAVEHDR hdr;
		MStdioFile *pmf = new MStdioFile();

		char *fname = (*preq)->req.datum;

		psp->vp_var1 = (void *)pmf;

		// Construct filename.bak to rename old save file.
		int cb = strlen(fname) + sizeof(ext);
		AUTOPSZ(fbak, cb);

		sprintf(fbak, "%s%s", fname, ext);

		// Delete previous backup.
		unlink(fbak);

		// rename backup file.
		rename(fname, fbak);

		// Create new backup file.
		if(!pmf->open((*preq)->req.datum, "wb"))
		{
			// Put the old one back
			rename(fbak,fname);
			fileError(preq, "Could not open save file");
			return FALSE;
		}

		hdr.version = SF_VERSION;
		hdr.timestamp = time(0);

		if(pmf->write(&hdr, sizeof(hdr)) != sizeof(hdr))
		{
			fileError(preq,"Could not write header");
			return FALSE;
		}
		msessiond->errprintf("Serialize");
	}
	else if(psp->cmd == PLUGIN_SERIALIZE_END)
	{
		if(pmf)
		{
			psp->vp_var1 = NULL;
			delete pmf;
		}
	}
	else if(psp->cmd == PLUGIN_SERIALIZE_SESSION)
	{
		SAVETAG tag;
		// Called for each valid session in the daemon. It is the job of the
		// plugin to enumerate and save all the variables in a session.

		assert(psp->session);
		int timeout = (psp->session->TestFlag(MSESS_FIXED)) ? -1 : psp->session->GetTimeout();

		char *session = (char *) psp->session->KeyGet();
		char *data = (char *) psp->session->DataGet();

		tag.tag = STAG_SESSION;
		tag.cb = strlen(session)+1;
		tag.param = timeout;
		if(pmf->write(&tag, sizeof(tag)) != sizeof(tag))
		{
			fileError(preq,"Could not write session tag");
			return FALSE;
		}
		if(pmf->write(session, tag.cb) != tag.cb)
		{
			fileError(preq,"Could not write session name");
			return FALSE;
		}
		tag.tag = STAG_DATA;
		tag.cb = 0;
		if(data)
		{
			tag.cb = strlen(data)+1;
			if(pmf->write(&tag, sizeof(tag)) != sizeof(tag))
			{
				fileError(preq,"Could not write session data tag");
				return FALSE;
			}
			if(pmf->write(data, tag.cb) != tag.cb)
			{
				fileError(preq,"Could not write session data");
				return FALSE;
			}
		}
		else
		{
			if(pmf->write(&tag, sizeof(tag)) != sizeof(tag))
			{
				fileError(preq,"Could not write session data tag");
				return FALSE;
			}
		}
		psp->session->Enum(sessionCallBack, (void *) psp, (void *)session);
	}
	else
	{
		(*preq)->req.stat = REQ_ERR;
		(*preq)->req.param = REQE_UNKNOWN;

		if(pmf)
		{
			psp->vp_var1 = NULL;
			delete pmf;
		}
		msessiond->errprintf("Unknown serialize command %d", psp->cmd);
		return FALSE;
	}
	return TRUE;
}

FN_HANDLER(default)
{
	return MPLUGIN_VER;
}

FN_INIT(default)
{
	PLUGFN_OPEN(entries)
	PLUGFN_DECL(PLUGIN_SERIALIZE)
	PLUGFN_DECL(PLUGIN_RESTORE)
	PLUGFN_CLOSE

	msessiond = (MSESSION_EXPORTS *) exports;
	*imports = (PLUGFN *)entries;

	InitPhoenix(0,NULL);

	if(msessiond->verbose)
		msessiond->errprintf("Save File plugin loaded");
	return NULL;
}

FN_SHUTDOWN(default)
{
}
