/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2002 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.
 
    You should have received a copy of the GNU General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to professionally license this library, the author
    can be reached at info@mohawksoft.com
*/ 

// This is the flex-plugin. It is a personality plugin which
// calls function plugin routines to service personality features.
// It is intended to answer the question, "How do I used more
// than one personality plugin?"
// By using the flex-plugin, "function only" plugins can define
// routines which correspond to plugin message names which will
// then be called from the personality plugin.
// The pgplug, a PostgreSQL serialization plugin, works with the
// flex-plugin.
// The flex-plugin will be loaded if no other plugin is specified


#include <stdio.h>
#include "plugapi.h"

char *fnstrs[]=
{
	"PLUGIN_START",
	"PLUGIN_PULSE",
	"PLUGIN_SESSION_FLUSH",
	"PLUGIN_SESSION_CREATE",
	"PLUGIN_SESSION_INIT",
	"PLUGIN_SESSION_DESTROY",
	"PLUGIN_ESCAPE",
	"PLUGIN_RELAY",
	"PLUGIN_SERIALIZE",
	"PLUGIN_RESTORE",
	"PLUGIN_SESSION_SETDATA",
	"PLUGIN_SESSION_SETVALUE"
};
static MSESSION_EXPORTS *msessiond=NULL;

static PlugFn handlers[MREQ_PLUGIN_LAST];

static int plugDefault(void *p1, void *p2)
{
	return MPLUGIN_VER;
}
FN_HANDLER(default)
{
	PlugFn fn = (msg>=0 && msg < MREQ_PLUGIN_LAST)?handlers[msg]: NULL;

	if(!fn)
	{
		if(msg>=0 && msg < MREQ_PLUGIN_LAST) 
			fn = msessiond->GetPluginFunction(fnstrs[msg], NULL);
		if(fn)
			handlers[msg] = fn;
		else
			fn = plugDefault;
	}
	return (fn)(p1,p2);
}

FN_INIT(default)
{
	msessiond = (MSESSION_EXPORTS *) exports;

	for(int i=0; i < MREQ_PLUGIN_LAST; i++)
	{
		handlers[i] = msessiond->GetPluginFunction(fnstrs[i], NULL);
	}

	// If there is a PLUGIN_INIT function, call it.
	PlugFn fn = (PlugFn)msessiond->GetPluginFunction("PLUGIN_INIT", NULL);

	if(fn)
		(fn)(NULL,NULL);

	// Return the pointer, to the PLUGIN_SHUTDOWN, if there is one.
	return (void *) msessiond->GetPluginFunction("PLUGIN_SHUTDOWN", NULL);
}

FN_SHUTDOWN(default)
{
	if(pdata)
		((PlugFn)pdata)(NULL,NULL);
}
