/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to professionally license this library, the author
    can be reached at info@mohawksoft.com
*/ 

// Enable this flag if you want extensive tracing
// #define TRACE_DEBUG 1
#define MSESSION_INTERNAL
#define SESSION_INTERNAL
#include "msessiond.h"

#define SET_STAT(preqb) (*preq)->req.stat = REQ_ERR;

// Walks the tree of variables for a session.
static Boolean statwalker(MNode *mn, void *param1, void *param2)
{
	TRACE_FUNCT("statwalker");
	LISTWALK *plw = (LISTWALK *) param1;
	char *name = (char *)mn->KeyGet();
	char *value = (char *)mn->DataGet();
	TRACE_STR(name);
	TRACE_STR(value);
	int cbname = strlen(name)+1;
	int cbvalue = strlen(value)+1;
	int cb_so_far = plw->cb + cbname + cbvalue;

	REQB *reqb = SIZEREQB(plw->reqb, cb_so_far);
	if(!reqb)
		return FALSE;
	else
		plw->reqb = reqb;
	
	char * var = &reqb->req.datum[plw->cb];

	memcpy(var,name,cbname);
	var+= cbname;
	memcpy(var, value,cbvalue);

	var+= cbvalue;

	plw->cb += cbname+cbvalue;
	plw->count++;
	
	return TRUE;
}
	
/** Finds all sessions with a name value pair */
static Boolean listwalker(MNode *mn, void *param1, void *param2)
{
	TRACE_FUNCT("listwalker");
	LISTWALK *plw = (LISTWALK *)param1;
	Session * sess = NODESESSION(mn);
	if(!sess)
	{
		TRACE_STR("Bad Session node!");
		return FALSE;
	}
	if(sess->CompareValue(plw->name, plw->value)==0)
	{
		char *szsession = (char *)mn->KeyGet();
		int cbsession = strlen(szsession)+1;
		int cb_so_far = plw->cb + cbsession;

		REQB *reqb = SIZEREQB(plw->reqb, cb_so_far);
		if(!reqb)
			return FALSE;
		else
			plw->reqb = reqb;
	
		char * var = &reqb->req.datum[plw->cb];

		memcpy(var, szsession, cbsession);
		plw->cb += cbsession;
		plw->count++;
	}
	return TRUE;
}
/** Finds all sessions */
static Boolean sessionwalker(MNode *mn, void *param1, void *param2)
{
	LISTWALK *plw = (LISTWALK *)param1;
	TRACE_FUNCT("sessionwalker");
	Session * sess = NODESESSION(mn);
	if(!sess)
	{
		TRACE_STR("Bad Session node!");
		return FALSE;
	}
	char *szsession = (char *)mn->KeyGet();
	int cbsession = strlen(szsession)+1;
	int cb_so_far = plw->cb + cbsession;

	REQB *reqb = SIZEREQB(plw->reqb, cb_so_far);
	if(!reqb)
		return FALSE;
	else
		plw->reqb = reqb;

	char * var = &reqb->req.datum[plw->cb];
	memcpy(var, szsession, cbsession);
	plw->cb += cbsession;
	plw->count++;
	return TRUE;
}
/** lists all sessions to serializer code */
static Boolean sessionserialize(MNode *mn, void *param1, void *param2)
{
	SERIALIZE_PARAM * psp = (SERIALIZE_PARAM *)param2;
	LISTWALK *plw = (LISTWALK *)param1;
	TRACE_FUNCT("sessionserialize");
	if(!mn)
	{
		TRACE_STR("Bad Session node!");
		return FALSE;
	}

	psp->cmd = PLUGIN_SERIALIZE_SESSION;
	psp->session = (Session *) mn;

	Boolean retval = (Boolean) globals.session_plugin.pcall(MREQ_PLUGIN_SERIALIZE,(void *)&plw->reqb, param2);

	plw->count++;
	return retval;
}
/** Finds all sessions with a specific variable*/
static Boolean sessionvarwalker(MNode *mn, void *param1, void *param2)
{
	LISTWALK *plw = (LISTWALK *)param1;
	TRACE_FUNCT("sessionvarwalker");
	Session * sess = NODESESSION(mn);
	if(!sess)
	{
		TRACE_STR("Bad Session node!");
		return FALSE;
	}
	char *szsession = (char *)mn->KeyGet();
	int cbsession = strlen(szsession)+1;
	int datasize=0;
	if(sess->GetValue(plw->name,NULL,&datasize, FALSE))
	{
		int cb_so_far = plw->cb + cbsession + datasize;
		REQB *reqb = SIZEREQB(plw->reqb, cb_so_far);
		if(!reqb)
			return FALSE;
		else
			plw->reqb = reqb;
		char * var = &reqb->req.datum[plw->cb];
		sess->GetValue(plw->name,var,&datasize, FALSE);
		var+=datasize;
		memcpy(var, szsession, cbsession);
		plw->cb += cbsession+datasize;
		plw->count++;
	}
	return TRUE;
}

void RemoveSession(Session *sess)
{
	globals.session_array->Remove(sess);
}
void DeleteSession(Session *sess)
{
	sess->DeleteSessionData(globals.session_array->GetFilePath());
	delete sess;
}
void FlushSession(FLUSHWALK *pflw)
{
	if(pflw->mn)
	{
		Session * sess= NODESESSION(pflw->mn);
		if(sess)
		{
			if(globals.session_plugin.pcall(MREQ_PLUGIN_SESSION_FLUSH, (void *)pflw, sess))
			{
				RemoveSession(sess);
				DeleteSession(sess);
				pflw->count++;
			}
		}
		pflw->mn = NULL;
	}
}

// Walks the list of sessions and flushes them
// Must flush a session AFTER we have found the next in the list!
static Boolean flushwalker(MNode *mn, void *param1, void *param2)
{
	FLUSHWALK *plw = (FLUSHWALK *)param1;
	TRACE_FUNCT("fluswalker");

	FlushSession(plw);

	if(plw->count >= plw->max)
		return FALSE;

	Session * sess = NODESESSION(mn);
	if(!sess)
	{
		TRACE_STR("Bad Session node!");
		return FALSE;
	}
	if(!sess->TestFlag(MSESS_FIXED)) // Fixed sessions do not get flushed.
	{
		time_t lastused = sess->last_used();
		if(sess->HasExpired(plw->now, plw->age))
			plw->mn = mn; // Mark for delete
		else if(lastused < plw->oldest)
			plw->oldest = lastused;
	}
	return TRUE;
}
// Adds multiple variable/value pairs to a session
// Assumes data starts of offset preq->req.name, and
// count of pairs in preq->req.value
static int intSetMulti(Session *sess, REQB * *preq)
{
	SESSION_DATA_PARAM	dp;
	int ndx = (*preq)->req.name;
	if(sess)
	{
		dp.session_obj = sess;
		dp.session_str = (char *)sess->KeyGet();

		char *name ;
		for(int i=0; i < (*preq)->req.value; i++)
		{
			name = &(*preq)->req.datum[ndx];
			int len = strlen(name)+1;
			ndx+=len;
			char *value = &(*preq)->req.datum[ndx];

			if(value)
			{
				dp.name = name;
				dp.value = value;

				len = strlen(value)+1;
				ndx+=len;
				if(globals.session_plugin.pcall(MREQ_PLUGIN_SESSION_SETVALUE, (void*)&dp, NULL))
					sess->SetValue(name,value);
			}
			else
				break;
		}
		(*preq)->req.stat = REQ_OK;
	}
	return (*preq)->req.stat;
}
// Returns the status of a session.
// And all its data.
static int reqCtl(REQB **preq)
{
	SET_STAT(preq);
	TRACE_FUNCT("reqCtl");
	char *session	=(char *)&(*preq)->req.datum[(*preq)->req.session];
	HASH hash 	= globals.session_array->strhash(session);

	globals.session_mutex.LockRead();
	Session *sess 	= (Session *) globals.session_array->FindStr(hash,session);

	if(sess)
	{
		TRACE_STRSTR("CTL Session exists", session);
		(*preq)->req.stat = REQ_OK;
		int ctlfn = (*preq)->req.param;
		
		switch(ctlfn)
		{
			case REQ_CTL_EXIST:
				(*preq)->req.param = sess->GetCreateTime();
				break;
			case REQ_CTL_TTL:
				(*preq)->req.param = sess->GetTTL(time(0), globals.req_timeout);
				break;
			case REQ_CTL_AGE:
				(*preq)->req.param = time(0) - sess->GetCreateTime();
				break;
			case REQ_CTL_TLA:
				(*preq)->req.param = time(0) - sess->GetLastTime();
				break;
			case REQ_CTL_CTIM:
				(*preq)->req.param = sess->GetCreateTime();
				break;
			case REQ_CTL_TOUCH:
				sess->SetLast();
			case REQ_CTL_NOW:
				(*preq)->req.param = time(0);
				break;
		}
		// Make sure values are in text
		sprintf((*preq)->req.datum, "%d", (*preq)->req.param);
		(*preq)->req.len = strlen((*preq)->req.datum)+1;
		TRACE_STRVAL("CTL_FN", ctlfn);
		TRACE_STRVAL("CTL_VAL", (*preq)->req.param);
	}
	else
	{
		TRACE_STRSTR("CTL Session does not exist", session);
		(*preq)->req.param = REQE_NOSESSION;
	}
	globals.session_mutex.UnlockRead();
	return (*preq)->req.stat;
}
// Returns the status of a session.
// And all its data.
static int reqGetAll(REQB * *preq)
{
	SET_STAT(preq);
	TRACE_FUNCT("reqGetAll");
	LISTWALK lw;
	char *session	=(char *)&(*preq)->req.datum[(*preq)->req.session];
	HASH hash 	= globals.session_array->strhash(session);

	globals.session_mutex.LockRead();
	Session *sess 	= (Session *) globals.session_array->FindStr(hash,session);

	TRACE_STR(session);
	lw.cb		= 0;
	lw.count	= 0;
	lw.reqb 	= *preq;

	if(sess)
	{
		sess->mtxlock();
		sess->Enum(statwalker, (void *) &lw, NULL, TRUE);
		sess->mtxunlock();
		*preq= lw.reqb;
		(*preq)->req.len = lw.cb;
		(*preq)->req.param = lw.count;
		(*preq)->req.stat = REQ_OK;
	}
	else
		(*preq)->req.param = REQE_NOSESSION;
	globals.session_mutex.UnlockRead();
	return (*preq)->req.stat;
}
// Sets a name/value pair or pairs
static int reqSet(REQB * *preq)
{
	SET_STAT(preq);
	TRACE_FUNCT("reqSet");
	SESSION_DATA_PARAM	dp;
	dp.session_str	=(char *)&(*preq)->req.datum[(*preq)->req.session];
	dp.name		=(char *)&(*preq)->req.datum[(*preq)->req.name];
	dp.value	=(char *)&(*preq)->req.datum[(*preq)->req.value];
	HASH hash 	= globals.session_array->strhash(dp.session_str);

	globals.session_mutex.LockRead();
	dp.session_obj 	= (Session *)globals.session_array->FindStr(hash,dp.session_str);
	if((*preq)->req.version == REQV_VMULTI)
		intSetMulti(dp.session_obj, preq);
	else if(dp.session_obj)
	{
		if(globals.session_plugin.pcall(MREQ_PLUGIN_SESSION_SETVALUE, (void*)&dp, NULL))
			dp.session_obj->SetValue(dp.name,dp.value);
		(*preq)->req.stat = REQ_OK;
	}
	else
	{
		TRACE_STRSTR("reqSet No Session found", dp.session_str);
		(*preq)->req.param = REQE_NOSESSION;
	}
	globals.session_mutex.UnlockRead();
	return (*preq)->req.stat;
}
// Gets a value from a name out of a session
static int reqGet(REQB **preq)
{
	SET_STAT(preq);
	TRACE_FUNCT("reqGet");
	char *session	=(char *)&(*preq)->req.datum[(*preq)->req.session];
	char *name	=(char *)&(*preq)->req.datum[(*preq)->req.name];
	HASH hash 	= globals.session_array->strhash(session);

	globals.session_mutex.LockRead();
	Session *sess	= (Session *)globals.session_array->FindStr(hash,session);
	if(sess)
	{
		int datasize=(*preq)->size;
		int len = sess->GetValue(name,(*preq)->req.datum, &datasize);
		if(len)
		{
			REQB *reqb = *preq;
			if(len < 0)
			{
				reqb = SIZEREQB(reqb, datasize);
				if(reqb)
				{
					len = sess->GetValue(name,reqb->req.datum, &datasize);
					*preq = reqb;
				}
			}
			if(reqb)
			{
				TRACE_STR(reqb->req.datum);
				reqb->req.len = len;
				reqb->req.param = atoi((*preq)->req.datum);
				reqb->req.stat = REQ_OK;
			}
			else
				(*preq)->req.param = REQE_BUFFER;
		}
		else
			(*preq)->req.param = REQE_NOVALUE;
	}
	else
		(*preq)->req.param = REQE_NOSESSION;
	globals.session_mutex.UnlockRead();
	return (*preq)->req.stat;
}

static int intCreateLow(REQB **preq, char *session, Session **newSession)
{
	TRACE_FUNCT("intCreateLow");
	if(globals.session_plugin.pcall(MREQ_PLUGIN_SESSION_CREATE, session))
	{
		Session *sess =new (session) Session(globals.session_array); 
		if(sess)
		{
			TRACE_STRSTR("intCreateLow:Session created", session);
			if(!globals.session_plugin.pcall(MREQ_PLUGIN_SESSION_INIT, session, sess))
			{	// The session is locked!!!
				TRACE_STR("Could not init session");
				sess->DeleteSessionData(globals.session_array->GetFilePath());
				delete sess;
				(*preq)->req.param = REQE_LOCKED;
				return FALSE;
			}
			if(!globals.session_mutex.LockWrite())
			{
				TRACE_STR("Could not Write Lock");
				sess->DeleteSessionData(globals.session_array->GetFilePath());
				delete sess;
				(*preq)->req.param = REQE_NOWLOCK;
				return FALSE;
			}
			globals.session_array->Insert(sess);

			globals.session_mutex.UnlockWriteLockRead();

			(*preq)->req.stat = REQ_OK;
			if((*preq)->req.version == REQV_VMULTI)
				intSetMulti(sess, preq);

			if((*preq)->req.param < 0)
				sess->SetFlag(MSESS_FIXED);
			else if((*preq)->req.param > 0)
				sess->SetTimeout((time_t)(*preq)->req.param);

			globals.session_mutex.UnlockRead();
			*newSession = sess;
		}
		else
		{
			TRACE_STRSTR("intCreateLow:Session denied", session);
			(*preq)->req.param = REQE_NOSESSION;
		}
	}
	else
	{
		TRACE_STRSTR("intCreateLow:Session failed", session);
		(*preq)->req.param = REQE_DENIED;
	}

	return (*preq)->req.stat;
}
// Creates a session.
// UNDOCUMENTED: Passing a value to create overides what
// plugin does.
static int reqCreate(REQB **preq)
{
	SET_STAT(preq);
	TRACE_FUNCT("reqCreate");
	char *session	=(char *)&(*preq)->req.datum[(*preq)->req.session];
	char *value	=(char *)&(*preq)->req.datum[(*preq)->req.value];

	if(value && strlen(value)==0)
		value = NULL;

	HASH hash 	= globals.session_array->strhash(session);

	Session *sess = (Session *)globals.session_array->FindStr(hash,session);

	TRACE_STR(session);

	if(!sess)
	{
		TRACE_STRSTR("Session does not exist", session);
		intCreateLow(preq, session, &sess);
		if(value)
			sess->SetSessionData(value, globals.session_array->GetFilePath());
	}
	else
	{
		TRACE_STRSTR("Session exists", session);
		(*preq)->req.param = REQE_DUPSESSION;
	}

	return (*preq)->req.stat;
}
// Deletes a session.
static int reqDrop(REQB * *preq)
{
	SET_STAT(preq);
	TRACE_FUNCT("reqDrop");
	char *session	=(char *)&(*preq)->req.datum[(*preq)->req.session];
	HASH hash 	= globals.session_array->strhash(session);

	Session *sess = (Session *)globals.session_array->FindStr(hash,session);

	if(!sess)
	{
		(*preq)->req.param = REQE_NOSESSION;
	}
	else if(globals.session_plugin.pcall(MREQ_PLUGIN_SESSION_DESTROY, session, sess))
	{
		TRACE_STR("Drop OK");
		(*preq)->req.param = REQE_UNKNOWN;

		if(!globals.session_mutex.LockWrite())
		{
			(*preq)->req.param = REQE_NOWLOCK;
			return FALSE;
		}

		globals.session_array->Remove(hash,sess);

		globals.session_mutex.UnlockWrite();

		sess->DeleteSessionData(globals.session_array->GetFilePath());

		delete sess;

		(*preq)->req.stat = REQ_OK;
	}
	else
	{
		TRACE_STR("Drop Denied");
		(*preq)->req.param = REQE_DENIED;
	}

	return (*preq)->req.stat;
}
// Returns all sessions with a name/value pair. (memory limited)
static int reqFind(REQB * *preq)
{
	SET_STAT(preq);
	LISTWALK lw;
	TRACE_FUNCT("reqFind");
	lw.name		=(char *)&(*preq)->req.datum[(*preq)->req.name];
	lw.value	=(char *)&(*preq)->req.datum[(*preq)->req.value];

	int cbname = strlen(lw.name)+1;
	int cbval = strlen(lw.value)+1;

	AUTOPSZ(szname, cbname);
	AUTOPSZ(szval,cbval);

	memcpy(szname, lw.name, cbname);
	memcpy(szval, lw.value, cbval);

	lw.name = szname;
	lw.value = szval;

	lw.cb		= 0;
	lw.count	= 0;
	lw.reqb 	= *preq;

	globals.session_mutex.LockRead();
	globals.session_array->Enum(listwalker, (void *) &lw, NULL, TRUE);
	globals.session_mutex.UnlockRead();

	*preq= lw.reqb;
	(*preq)->req.len = lw.cb;
	(*preq)->req.stat = REQ_OK;
	(*preq)->req.param = lw.count;
	TRACE_VAL(lw.count);
	return (*preq)->req.stat;
}
// Returns all session:value array of all sessions
// with a variable set
static int reqListVar(REQB * *preq)
{
	SET_STAT(preq);
	LISTWALK lw;
	TRACE_FUNCT("reqListVar");
	lw.name		=(char *)&(*preq)->req.datum[(*preq)->req.name];
	int cbname = strlen(lw.name)+1;
	AUTOPSZ(szname, cbname);
	memcpy(szname, lw.name, cbname);
	lw.name 	= szname;
	lw.value 	= 0;
	lw.cb		= 0;
	lw.count	= 0;
	lw.reqb 	= *preq;
	globals.session_mutex.LockRead();
	globals.session_array->Enum(sessionvarwalker, (void *) &lw, NULL, TRUE);
	globals.session_mutex.UnlockRead();
	*preq= lw.reqb;
	(*preq)->req.len = lw.cb;
	(*preq)->req.stat = REQ_OK;
	(*preq)->req.param = lw.count;
	TRACE_VAL(lw.count);
	return (*preq)->req.stat;
}
// Returns all sessions 
static int reqList(REQB * *preq)
{
	SET_STAT(preq);
	LISTWALK lw;
	TRACE_FUNCT("reqList");

	lw.cb		= 0;
	lw.count	= 0;
	lw.reqb 	= *preq;
	globals.session_mutex.LockRead();
	globals.session_array->Enum(sessionwalker, (void *) &lw, NULL, TRUE);
	globals.session_mutex.UnlockRead();
	*preq= lw.reqb;
	(*preq)->req.len = lw.cb;
	(*preq)->req.stat = REQ_OK;
	(*preq)->req.param = lw.count;
	TRACE_VAL(lw.count);
	return (*preq)->req.stat;
}
// Returns count of sessions.
static int reqCount(REQB * *preq)
{
	SET_STAT(preq);
	TRACE_FUNCT("reqCount");
	(*preq)->req.stat = REQ_OK;
	(*preq)->req.param = globals.session_array->count();
	return (*preq)->req.stat;
}
// Finds all the sessions which have expired and deletes them.
static int reqFlush(REQB * *preq)
{
	SET_STAT(preq);
	FLUSHWALK lw;
	TRACE_FUNCT("reqFlush");

	memset(&lw, 0, sizeof(lw));

	lw.now = time(0);
	lw.mn = NULL;
	lw.age = (*preq)->req.param;
	lw.count=0;
	lw.oldest=lw.now;
	lw.max = globals.maxflush;

	if(!globals.session_mutex.LockWrite())
	{
		(*preq)->req.param = REQE_NOWLOCK;
		return FALSE;
	}

	// Enumerate sessions. If they are expired, flush them.
	globals.session_array->Enum(flushwalker, (void *) &lw, NULL, TRUE);

	// Flush the last session
	FlushSession(&lw);

	globals.session_mutex.UnlockWrite();

	// This is where the plugin should save its sessions, not during enumeration.
	globals.session_plugin.pcall(MREQ_PLUGIN_SESSION_FLUSH, (void *)&lw, NULL);

	(*preq)->req.stat = REQ_OK;
	(*preq)->req.param = lw.count;
	(*preq)->req.session = lw.oldest;
	(*preq)->req.name = lw.max;
	TRACE_VAL(lw.count);
	return (*preq)->req.stat;
}
// Marks a session as locked.
// Creates a session if one does not exist.
// Returns session's data string and key
static int reqSlock(REQB **preq)
{
	SET_STAT(preq);
	TRACE_FUNCT("reqSlock");
	char *session=(char *)&(*preq)->req.datum[(*preq)->req.session];
	HASH hash = globals.session_array->strhash(session);
	int locked = 0;

	globals.session_mutex.LockRead();
	Session *sess	= (Session *)globals.session_array->FindStr(hash,session);
	if(!sess)
	{
		TRACE_STR("reqSunlock:new session");
		sess =new (session) Session(globals.session_array); 
		globals.session_mutex.UnlockRead();
		if(!globals.session_mutex.LockWrite())
		{
			delete sess;
			(*preq)->req.param = REQE_NOWLOCK;
			return FALSE;
		}
		globals.session_array->Insert(hash,sess);
		globals.session_mutex.UnlockWriteLockRead();
	}
	if(sess)
	{
		locked = sess->lock();
		if(locked)
		{
			int datasize=(*preq)->size;
			int len = sess->GetSessionData((*preq)->req.datum, &datasize, globals.session_array->GetFilePath());
			if(len < 0)
			{
				REQB *reqb = *preq;
				reqb = SIZEREQB(reqb, datasize);
				if(reqb)
				{
					len = sess->GetSessionData(reqb->req.datum, &datasize, globals.session_array->GetFilePath());
					*preq =reqb;
				}
			}
			(*preq)->req.len = (len < 0) ? 0 : len;
			(*preq)->req.param = locked;
			(*preq)->req.stat = REQ_OK;
		}
		else
			(*preq)->req.param = REQE_LOCKED;
	}
	else
		(*preq)->req.param = REQE_NOSESSION;

	globals.session_mutex.UnlockRead();
	return (*preq)->req.stat;
}
// Unlocks a previously locked session.
// Updates data string if one is sent.
static int reqSunlock(REQB **preq)
{
	SET_STAT(preq);
	TRACE_FUNCT("reqSunlock");
	char *session	=(char *)&(*preq)->req.datum[(*preq)->req.session];
	char *value	=(char *)&(*preq)->req.datum[(*preq)->req.value];

	HASH hash 	= globals.session_array->strhash(session);

	globals.session_mutex.LockRead();
	Session *sess	= (Session *)globals.session_array->FindStr(hash,session);

	if(sess)
	{
		Boolean unlocked = sess->unlock((*preq)->req.param);

		if(unlocked)
		{
			if(strlen(value))
				sess->SetSessionData(value, globals.session_array->GetFilePath());
			(*preq)->req.stat = REQ_OK;
			(*preq)->req.param = 0;
		}
		else
			(*preq)->req.param = REQE_LOCKED;
	}
	else
		(*preq)->req.param = REQE_NOSESSION;

	globals.session_mutex.UnlockRead();
	return (*preq)->req.stat;
}
// Fixes a session in memory
static int reqTimeout(REQB **preq)
{
	SET_STAT(preq);
	TRACE_FUNCT("reqFix");
	char *session	=(char *)&(*preq)->req.datum[(*preq)->req.session];
	HASH hash 	= globals.session_array->strhash(session);
	globals.session_mutex.LockRead();
	Session *sess	= (Session *)globals.session_array->FindStr(hash,session);
	if(sess)
	{
		if((*preq)->req.param < 0)
		{	// Never clean up session
			sess->SetFlag(MSESS_FIXED);
		}
		else if((*preq)->req.param == 0)
		{	// Reset session to default
			sess->ResetFlag(MSESS_FIXED);
			sess->SetTimeout((time_t)0);
		}
		else
		{	// Set for timeout
			sess->ResetFlag(MSESS_FIXED);
			sess->SetTimeout((time_t)(*preq)->req.param);
		}
		(*preq)->req.param = REQE_NOERROR;
		(*preq)->req.stat = REQ_OK;
	}
	else
	{
		TRACE_STR("No Session");
		TRACE_STR(session);
		(*preq)->req.param = REQE_NOSESSION;
	}

	globals.session_mutex.UnlockRead();
	return (*preq)->req.stat;
}
// Implement a "sequence-like" method for a counter.
static int reqInc(REQB * *preq)
{
	SET_STAT(preq);
	SESSION_DATA_PARAM dp;

	TRACE_FUNCT("reqInc");
	dp.session_str	=(char *)&(*preq)->req.datum[(*preq)->req.session];
	dp.name		=(char *)&(*preq)->req.datum[(*preq)->req.name];

	HASH hash 	= globals.session_array->strhash(dp.session_str);

	globals.session_mutex.LockRead();
	dp.session_obj = (Session *)globals.session_array->FindStr(hash,dp.session_str);
	if(dp.session_obj)
	{
		int lenval;
		char buffer[16];
		(*preq)->req.param = dp.session_obj->nextval(dp.name, buffer);
		lenval = strlen(buffer)+1;
		(*preq)->req.len = lenval;

		dp.session_str = (char *)dp.session_obj->KeyGet();
		dp.value = (char *)buffer;

		if(globals.session_plugin.pcall(MREQ_PLUGIN_SESSION_SETVALUE, (void*)&dp, NULL))
		{
			memcpy((*preq)->req.datum, buffer, lenval);
			(*preq)->req.stat = REQ_OK;
		}
		else
			(*preq)->req.stat = REQ_ERR;

	}
	else
		(*preq)->req.param = REQE_NOSESSION;

	globals.session_mutex.UnlockRead();
	return (*preq)->req.stat;
}
// Sets data string for a session.
// if the session does not exist, it is created.
static int reqDataSet(REQB **preq)
{
	SET_STAT(preq);
	TRACE_FUNCT("reqDataSet");
	SESSION_DATA_PARAM dp;

	dp.session_str	=(char *)&(*preq)->req.datum[(*preq)->req.session];
	dp.value	=(char *)&(*preq)->req.datum[(*preq)->req.value];

	HASH hash = globals.session_array->strhash(dp.session_str);

	globals.session_mutex.LockRead();
	dp.session_obj 	= (Session *)globals.session_array->FindStr(hash,dp.session_str);

	TRACE_STR(dp.session_str);
	TRACE_STR(dp.value);
	
	if(!dp.session_obj)
	{
		(*preq)->req.param = REQE_NOSESSION;
		(*preq)->req.stat = REQ_ERR;
	}
	else
	{
		if(globals.session_plugin.pcall(MREQ_PLUGIN_SESSION_SETDATA, (void *)&dp,NULL))
			dp.session_obj->SetSessionData(dp.value, globals.session_array->GetFilePath());

		(*preq)->req.stat = REQ_OK;
	}
	globals.session_mutex.UnlockRead();
	return (*preq)->req.stat;
}
// Gets the data string for a session.
static int reqDataGet(REQB * *preq)
{
	SET_STAT(preq);
	TRACE_FUNCT("reqDataGet");
	char *session	=(char *)&(*preq)->req.datum[(*preq)->req.session];
	HASH hash = globals.session_array->strhash(session);

	globals.session_mutex.LockRead();
	Session *sess	= (Session *)globals.session_array->FindStr(hash,session);

	if(!sess)
	{
		TRACE_STR("Create Session for New Data");
		globals.session_mutex.UnlockRead();
		intCreateLow(preq, session, &sess);
		globals.session_mutex.LockRead();
	}
	TRACE_STRSTR("DataGet Session", session);
	if(sess)
	{
		int datasize=(*preq)->size;
		int len = sess->GetSessionData((*preq)->req.datum, &datasize, globals.session_array->GetFilePath());
			
		if(len < 0)
		{
			REQB *reqb = *preq;
			reqb = SIZEREQB(reqb, datasize);
			if(reqb)
			{
				len = sess->GetSessionData(reqb->req.datum, &datasize, globals.session_array->GetFilePath());
				*preq =reqb;
			}
		}
		if(len)
		{
			TRACE_STR((*preq)->req.datum);
			(*preq)->req.param =0;
			(*preq)->req.stat = REQ_OK;
		}
		else
			(*preq)->req.param = REQE_NOVALUE;

		(*preq)->req.len = (len < 0) ? 0 : len;
	}
	else
		(*preq)->req.param = REQE_NOSESSION;
	globals.session_mutex.UnlockRead();
		
	return (*preq)->req.stat;
}

static int reqUniq(REQB **preq)
{
	SET_STAT(preq);
	TRACE_STR("Unique Session request");
	int nchar = (*preq)->req.param;
	REQB *reqb = SIZEREQB(*preq, nchar+1);

	Session *sess = NULL;
	if(reqb)
	{
		char *session = (char *)reqb->req.datum;
		do
		{
			HASH hash;
			TimeRandString(session, nchar);
			TRACE_STRSTR("Unique Session request:", session);
			hash = globals.session_array->strhash(session);
			TRACE_STR(session);
			sess = (Session *)globals.session_array->FindStr(hash,session);
		}
		while(sess != NULL);

		reqb->req.param = 0; // Use default
		*preq = reqb;

		if(intCreateLow(preq, session, &sess) == REQ_OK)
		{
			reqb->req.len = nchar+1;
			TRACE_STR("Unique Session request: SUCCESS");
		}
		else
		{
			TRACE_STR("Unique Session request: FAILED in intCreateLow");
		}
	}
	else
	{
		TRACE_STR("Unique Session request: FAILED");
		(*preq)->req.param = REQE_NOSESSION;
	}
		
	return (*preq)->req.stat;
}
static int reqRandStr(REQB **preq)
{
	SET_STAT(preq);
	TRACE_FUNCT("reqRandStr");
	int nchar = (*preq)->req.param;
	int cb = nchar+1;
	REQB *reqb = SIZEREQB(*preq, cb);
	char *retstr = (char *)reqb->req.datum;
	RandString(retstr, nchar);
	*preq = reqb;
	(*preq)->req.stat = REQ_OK;
	(*preq)->req.len = cb;
	return (*preq)->req.stat;
}
static int reqPlugin(REQB **preq)
{
	SET_STAT(preq);
	TRACE_FUNCT("reqPlugin");
	// This is the default error 
	// The plugin should change the error

	int result = globals.session_plugin.pcall(MREQ_PLUGIN_ESCAPE, preq);

	if(result == NO_PLUGIN_FUNCTION || result == MPLUGIN_VER)
	{
		(*preq)->req.param = REQE_NOFN;
		(*preq)->req.len=0;
	}

	return (*preq)->req.stat;
}
static int reqFunction(REQB **preq)
{
	SET_STAT(preq);
	TRACE_FUNCT("reqFunction");
	PlugFn fn = globals.session_plugmgr.FindFn((*preq)->req.datum);

	if(fn)
		return globals.session_plugmgr.CallPluginFn((*preq)->req.datum, (void *)preq, NULL);
	else
	{
		(*preq)->req.param = REQE_NOFN;
		return (*preq)->req.stat;
	}
}
// Returns all sessions 
static int reqSerialize(REQB * *preq)
{
	SET_STAT(preq);
	LISTWALK lw;
	TRACE_FUNCT("reqSerialize");

	SERIALIZE_PARAM	sp;

	memset(&sp, 0, sizeof(sp));

	sp.cmd = PLUGIN_SERIALIZE_BEGIN;

	if(globals.session_plugin.pcall(MREQ_PLUGIN_SERIALIZE,(void *)preq, (void *)&sp))
	{
		lw.cb		= 0;
		lw.count	= 0;
		lw.reqb 	= *preq;
		sp.cmd 		= PLUGIN_SERIALIZE_SESSION;
		sp.shutdown 	= (*preq)->req.param;
		globals.session_mutex.LockRead();
		globals.session_array->Enum(sessionserialize, (void *) &lw, (void *) &sp, TRUE);
		globals.session_mutex.UnlockRead();
		*preq= lw.reqb;
		(*preq)->req.stat = REQ_OK;
		(*preq)->req.param = lw.count;
		TRACE_VAL(lw.count);

		sp.cmd = PLUGIN_SERIALIZE_END;

		globals.session_plugin.pcall(MREQ_PLUGIN_SERIALIZE,(void *)preq, (void *) &sp);
	}
	return (*preq)->req.stat;
}
// Called on startup, restores the session manager
static int reqRestore(REQB * *preq)
{
	SET_STAT(preq);
	globals.session_plugin.pcall(MREQ_PLUGIN_RESTORE,(void *)preq, (void *) NULL);
	return (*preq)->req.stat;
}

// Called to exec
static int reqExec(REQB * *preq)
{
	SET_STAT(preq);
	sprintf((*preq)->req.datum, "Execute function disallowed by default, use plugin\n");
	(*preq)->req.len = strlen((*preq)->req.datum)+1;
	(*preq)->req.stat = REQ_ERR;
	(*preq)->req.param = REQE_DENIED;
	return (*preq)->req.stat;
}

static int reqNofunct(REQB * *preq)
{
	SET_STAT(preq);
	(*preq)->req.param = REQE_UNKNOWN;
	return REQ_ERR;
}

char *MReqErrText(int param)
{
	switch(param)
	{
		case REQE_NOERROR:
			return "No error";
		case REQE_NOSESSION:
			return "No session was found";
		case REQE_DUPSESSION:
			return "Session already exists";
		case REQE_NOWLOCK:
			return "Could not obtain write lock";
		case REQE_NOVALUE:
			return "No value was found in session";
		case REQE_UNKNOWN:
			return "Unknown error";
		case REQE_LOCKED:
			return "Already locked, or you are not the owner of the lock.";
		case REQE_NOSEND:
			return "Could not send request to server";
		case REQE_BADCONN:
			return "Bad connection value used";
		case REQE_BADHDR:
			return "Bad header in request";
		case REQE_BADVER:
			return "Bad version in request";
		case REQE_BADLEN:
			return "Bad datum length in request";
		case REQE_NOFN:
			return "Function not found";
		case REQE_BUFFER:
			return "Buffer size error";
		case REQE_DENIED:
			return "Operation denied";
		default:
			return "Error value out of range";
	}
}
ReqHandler reqHandlers[]=
{
	NULL,		//REQ_ERR
	NULL,		//REQ_OK
	reqCtl, 	//REQ_CTL
	reqSet,		//REQ_SETVAL
	reqGet,		//REQ_GETVAL
	reqCreate,	//REQ_CREATE
	reqDrop,	//REQ_DROP
	reqGetAll,	//REQ_GETALL
	reqFind,	//REQ_FIND
	reqCount,	//REQ_COUNT
	reqFlush,	//REQ_FLUSH
	reqSlock,	//REQ_SLOCK
	reqSunlock,	//REQ_SUNLOCK
	reqTimeout,	//REQ_TIMEOUT
	reqInc,		//REQ_INC
	reqDataGet,	//REQ_DATAGET
	reqDataSet,	//REQ_DATASET
	reqList,	//REQ_LIST
	reqListVar,	//REQ_LISTVAR
	reqUniq,	//REQ_UNIQ
	reqRandStr,	//REQ_RANDSTR
	reqPlugin,	//REQ_PLUGIN
	reqFunction,	//REQ_CALL
	reqSerialize,	//REQ_SERIALIZE
	reqRestore,	//REQ_RESTORE
	reqExec,	//REQ_EXEC
	NULL
};

extern "C" ReqHandler MsessionReplaceHandler(ReqHandler handler, int ndx)
{
	ReqHandler ret = NULL;
	if(ndx > REQ_OK && ndx < MREQ_LAST)
	{
		ret = reqHandlers[ndx];
		reqHandlers[ndx] = handler;
	}
	return ret;
}
extern "C" int MsessionHandleRequest(REQB **preq)
{
	TRACE_FUNCT("MsessionHandleRequest");
	int stat = (*preq)->req.stat;

#ifdef ENABLE_RELAY
	if(globals.session_plugin.pcall(MREQ_PLUGIN_RELAY, preq)) {
#endif
	(*preq)->req.len = 0;

	if(stat > REQ_OK && stat < MREQ_LAST)
	{
		stat = (reqHandlers[stat])(preq);

		// See if fnction was even handled!!!
		if((*preq)->req.stat != REQ_OK && (*preq)->req.stat != REQ_ERR)
		{
			(*preq)->req.stat = REQ_ERR;
			(*preq)->req.param = REQE_NOFN;
		}
		if(stat == REQ_ERR && (*preq)->req.len == 0)
		{
			char *perr = MReqErrText((*preq)->req.param);
			if(perr)
			{
				int len = strlen(perr)+1;
				memcpy((*preq)->req.datum, perr, len);
				(*preq)->req.len = len;
			}
		}
	}
	else
		(*preq)->req.stat = REQ_ERR;
	return stat;
#ifdef ENABLE_RELAY
	} else return REQ_ERR;
#endif

}
extern "C" int ExecReqDirect(REQB **ppreq, int stat, char *session, char *name, char *value, int param)
{
	FormatRequest(ppreq, stat, session, name, value, param);
	return MsessionHandleRequest(ppreq);
}
extern "C" int ExecReqMultiDirect(REQB **ppreq, int stat, char *session, int n, char **pairs, int param)
{
	FormatRequestMulti(ppreq, stat, session, n, pairs, param);
	return MsessionHandleRequest(ppreq);
}
