/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2002 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.
 
    You should have received a copy of the GNU General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to professionally license this library, the author
    can be reached at info@mohawksoft.com
*/ 


// Alow debug information.
// #define TRACE_DEBUG

// Allow headers to present internal information
#define MSESSION_INTERNAL
// Allow definitions from session internals
#define SESSION_INTERNAL
// Allow definitions from plugins.
#define MSESSION_PLUGIN
// Allow mreq internals.
#define MREQ_INTERNAL

#include "msessiond.h"
#include "signal.h"

MSESSION_GLOBALS	globals;

char g_szOptions[]="Va:?p:d:i:t:q:vl:e:F:h:g:n:A:D:P:F:R:S:T:";

void errout(char *str)
{
	mferr.mfprintf("%s %s\n", timestr(0),str);
}
void errprintf(char *fmt, ...)
{
	va_list arglist;
	char buffer[MAX_STRING];
	va_start(arglist, fmt);
	VSNPRINTF(buffer, sizeof(buffer), fmt, arglist);
	va_end(arglist);
	mferr.mfprintf("%s %s\n", timestr(0),buffer);
}


// Default signal handler
void signal_handler(int sig)
{
	if(globals.verbose > 1)
		errprintf("Msession received signal, shutting down [%d]", getpid());
	globals.smgr->SetContinue(FALSE);
}

// Load a personality plugin.
Boolean LoadPersonality(char *optarg)
{
	char name[PATH_MAX];
	char path[PATH_MAX];

	globals.session_plugmgr.GetPluginInfo(optarg, name, path);

	globals.session_plugin.Open(path, name);
	
	PLUGFN *plgf = globals.session_plugin.getimports();

	if(plgf)
		globals.session_plugmgr.MapFunctions(plgf);
	return TRUE;
}

int CallPluginFunction(char *name, void *p1, void *p2)
{
	return globals.session_plugmgr.CallPluginFn(name,p1,p2);
}

PlugFn	GetPluginFunction(char *fnname, PlugFn fnDefault)
{
	return globals.session_plugmgr.GetFn(fnname, fnDefault);
}
SessionMgr *	GetSessionMgr(void)
{
	return globals.session_array;
}

void InitGlobals(int argc, char **argv)
{
	globals.serialize = phmstrdup(DEF_SERIALIZE);
	globals.version = sizeof(globals);
	globals.argc=argc;
	globals.argv=argv;
	globals.dropped=0;
	globals.autodrop=TRUE;
	globals.autosave=DEF_AUTOSAVE;
	globals.verbose=DEF_VERBOSE;	
	globals.maxThreads=DEF_MAXTH;
	globals.listeners = DEF_LIST;
	globals.globals=DEF_GLOBALS;
	globals.req_timeout= DEF_EXP;
	globals.name = NULL;
	globals.port = DEF_PORT;
	globals.dead=DEF_DEAD;
	globals.interval=DEF_INTR;
	globals.queue = DEF_QUE;
	globals.nhash= DEF_HASH;
	globals.lock = DEF_LOCK;
	globals.personality=0;
	globals.szfilePath=NULL;
	globals.szPersonality=NULL;
	globals.smgr=NULL;
	globals.mcfg=NULL;

	globals.SizeRequestBuffer = SizeRequestBuffer;
	globals.MsessionReplaceHandler = MsessionReplaceHandler;
	globals.MsessionHandleRequest = MsessionHandleRequest;
	globals.AllocateRequestBuffer = AllocateRequestBuffer;
	globals.FreeRequestBuffer = FreeRequestBuffer;
	globals.FormatRequest = FormatRequest;
	globals.FormatRequestMulti= FormatRequestMulti;
	globals.CallPluginFunction = CallPluginFunction;
	globals.GetPluginFunction = GetPluginFunction;
	globals.ExecuteRequest = ExecReqDirect;
	globals.ExecuteRequestMulti = ExecReqMultiDirect;
	globals.GetSessionMgr = GetSessionMgr;
	globals.errout = errout;
	globals.errprintf = errprintf;
	globals.DeleteSession=DeleteSession;
	globals.RemoveSession=RemoveSession;
	globals.session_plugin.SetExports((PLUGFN *)&globals);
	globals.session_plugmgr.SetExports((PLUGFN *)&globals);
}

void InitCmdOptGlobals(int argc, char **argv)
{
	int opt=1;
	int optind=1;
	char *optarg;
	char *safe_optarg;

	// Process options
	for(optind=1; 1; )
	{	
		opt = GetOption(argc,argv,g_szOptions, &optarg, &optind);
		safe_optarg = (optarg) ? optarg : (char *)"";

		switch(opt)
		{
			case 'a': 
				if(!MSock::allowhba(safe_optarg))
				{
					errprintf("Invalid HBA format \"%s\"\n", safe_optarg);
					errprintf("Should be \"255.255.255.0/192.168.1.0\"\n");
					exit(0);
				}
				break;
			case 'd':
				globals.dead = atoi(safe_optarg);
				break;
			case 'e':
				globals.req_timeout= atoi(safe_optarg);
				break;
			case 'g':
				if(safe_optarg)
					globals.globals = phmstrdup(safe_optarg);
				break;
			case 'i':
				globals.interval = atoi(safe_optarg);
				break;
			case 'l':
				globals.lock = atoi(safe_optarg);
				break;
			case 't':
				globals.listeners = atoi(safe_optarg);
				break;
			case 'p':
				globals.port = atoi(safe_optarg);
				break;
			case 'q':
				globals.queue = atoi(safe_optarg);
				break;
			case 'v':
				globals.verbose++;
				break;
			case 'h':
				globals.nhash = atoi(safe_optarg);
				break;
			case 'n':
				globals.name = safe_optarg;
				break;
			case 'A':
				globals.autosave = atoi(safe_optarg);
				break;
			case 'D':
				if(*safe_optarg == 'Y' || *safe_optarg == 'y')
					globals.autodrop = TRUE;
				else if (*safe_optarg == 'N' || *safe_optarg == 'n')
					globals.autodrop = FALSE;
				break;
			case 'R':
				globals.szfilePath=safe_optarg;
				break;
			case 'S':
				phmfree(globals.serialize);
				globals.serialize = phmstrdup(safe_optarg);
				break;
			case 'T':
				globals.maxThreads = atoi(safe_optarg);
				break;
			case 'V':
				printf("%s\nCompiled %s\n", MSESSION_VERSION,__DATE__);
			case '?':
				printf("Usage msessiond [options]\n");
				printf("\t-a hba,  Add allowed HBA in the form: \"255.255.255.0/192.168.2.0\"\n");
				printf("\t-d n,    dead session timeout in seconds (%d)\n", globals.dead);
				printf("\t-e n,    Number of seconds a persistent connection takes to expire (%d)\n", DEF_EXP);
				printf("\t-g str,  Create a fixed session for global variables\n");
				printf("\t-h n,    size of session hash table (%d)\n", globals.nhash);
				printf("\t-i n,    interval between dead sesson garbage collection (%d)\n", globals.interval);
				printf("\t-l n,    timeout value of a lock (%d)\n", globals.lock);
				printf("\t-t n,    Number of listener threads to start (%d)\n", globals.listeners);
				printf("\t-p n,    TCP/IP port on which to listen (%d)\n", globals.port);
				printf("\t-q n,    number of backlogged connections in queue (%d)\n", globals.queue);
				printf("\t-v,      verbose execution\n");
				printf("\t-A n,    Autosave every [n] minutes\n");
				printf("\t-D y/n,  Automatically drop connections when there are no free threads [%s]\n",(globals.autodrop) ? "Yes" : "No");
				printf("\t-F str,  Function name:module. Must be Full path, or name in LD_LIBRARY_PATH\n");
				printf("\t-P str,  Personality name:module. Must be Full path, or name in LD_LIBRARY_PATH.\n");
				printf("\t-R path, Use file based raw session data\n");
				printf("\t-T n,    Maximum number of threads (%d)\n", globals.maxThreads);
				exit(0);
				break;
		}
		if(opt <= 0)
			break;
	}
}

void InitCmdOptPlugins(int argc, char **argv)
{
	int opt=1;
	int optind=1;
	char *optarg;

	for (optind=1; 1; )
	{	opt = GetOption(argc,argv,g_szOptions, &optarg, &optind);
		switch (opt)
		{
			case 'F':
				errprintf("Loading Function Plugin: %s", optarg);
				if(!globals.session_plugmgr.LoadFunctionMod(optarg))
					exit(0);
				break;
			case 'P':
				errprintf("Loading Personality %s\n", optarg);
				if(globals.personality++)
				{
					mferr.mfprintf("Warning: Only one personality module can be loaded, replacing previous definition\n");
					phmfree(globals.szPersonality);
					
				}
				globals.szPersonality = phmstrdup(optarg);
				break;
		}
		if(opt <= 0)
			break;
	}
}
#define SAFESTRDUP( F )	 strT = F; (strT) ? phmstrdup(strT) : NULL

void InitFileOptGlobals(int argc, char **argv)
{
	char *strT;
	globals.mcfg = new MConfigFile (argc, argv);
	assert(globals.mcfg);

	globals.port 		= globals.mcfg->GetIntValue(CFG_PORT,DEF_PORT);
	globals.listeners 	= globals.mcfg->GetIntValue(CFG_LIST,DEF_LIST);
	globals.dead		= globals.mcfg->GetIntValue(CFG_DEAD,DEF_DEAD);
	globals.queue 		= globals.mcfg->GetIntValue(CFG_QUE,DEF_QUE);
	globals.interval	= globals.mcfg->GetIntValue(CFG_INTR,DEF_INTR);
	globals.lock 		= globals.mcfg->GetIntValue(CFG_LOCK,DEF_LOCK);
	globals.req_timeout	= globals.mcfg->GetIntValue(CFG_EXP,DEF_EXP);
	globals.nhash		= globals.mcfg->GetIntValue(CFG_HASH,DEF_HASH);
	globals.maxThreads	= globals.mcfg->GetIntValue(CFG_MAXTH,DEF_MAXTH);
	globals.autosave	= globals.mcfg->GetIntValue(CFG_AUTOSAVE,DEF_AUTOSAVE);
	globals.verbose		= globals.mcfg->GetIntValue(CFG_VERBOSE,DEF_VERBOSE);	
	globals.sessionlen	= globals.mcfg->GetIntValue(CFG_SESSIONLEN, DEF_SESSIONLEN);

	// String values
	globals.globals		= SAFESTRDUP( globals.mcfg->GetValue(CFG_GLOBALS, DEF_GLOBALS) );
	globals.name 		= SAFESTRDUP( globals.mcfg->GetValue(CFG_NAME, DEF_NAME) );
	globals.szfilePath	= SAFESTRDUP( globals.mcfg->GetValue(CFG_FILEPATH,DEF_FILEPATH) );
	globals.szPersonality	= SAFESTRDUP( globals.mcfg->GetValue(CFG_PERSONALITY, DEF_PERSONALITY) );

	// Integer values
	globals.autodrop	= globals.mcfg->GetIntValue(CFG_AUTODROP, TRUE);
	globals.maxflush	= globals.mcfg->GetIntValue(CFG_MAXFLUSH, DEF_MAXFLUSH);
	
	for(int i=0; 1; i++)
	{
		char buffer[32];
		sprintf(buffer, CFGF_HBA, i);

		char *hba = globals.mcfg->GetValue(buffer, NULL);

		if(!hba)
			break;

		if(!MSock::allowhba(hba))
		{
			errprintf("Invalid HBA format \"%s\"\n", hba);
			errprintf("Should be \"255.255.255.0/192.168.1.0\"");
			exit(0);
		}
	}
}

void InitFileOptPlugins(int argc, char **argv)
{
	assert(globals.mcfg);
	for(int i=0; 1; i++)
	{
		char buffer[32];
		sprintf(buffer, CFGF_PLUGIN, i);

		char *plugin = globals.mcfg->GetValue(buffer, NULL);

		if(!plugin)
			break;

		if(!globals.session_plugmgr.LoadFunctionMod(plugin))
			exit(0);
	}
}

int MAIN(msessiond)(int argc, char **argv)
{
	// Initialize the phoenix library
	InitPhoenix(argc, argv);

	// Init globals struct
	InitGlobals(argc, argv);

	// Read options from file
	InitFileOptGlobals(argc, argv);

	// Read options from command line
	InitCmdOptGlobals(argc,argv);

	// Create the session manager
	globals.session_array = new SessionMgr(globals.nhash);

	// Process the config file plugins
	InitFileOptPlugins(argc, argv);

	// Process the Command line plugins.
	InitCmdOptPlugins(argc,argv);

	MSock msserv(globals.port);
	globals.smgr = new SServerMgr(&msserv);

	globals.smgr->setCleanInterval(globals.interval);
	globals.smgr->setDeadTimeout(globals.dead);

	globals.session_array->SetMaxLockTime(globals.lock);
	globals.session_array->SetFilePath(globals.szfilePath);

	if(!LoadPersonality(globals.szPersonality))
		errprintf("Could not locate plugin %s, running without personality enhancements", 
			globals.szPersonality);
	
	globals.smgr->Init();

	while(1)
	{
		if(msserv.Listen(globals.queue))
			break;

		mferr.mfprintf("Can't listen on port, try again in 1 second\n");
		sleep(1);
	}
#ifdef POSIX
	signal(SIGPIPE, SIG_IGN);
	signal(SIGHUP, signal_handler);
#endif
	signal(SIGINT, signal_handler);
	globals.smgr->PreRun();
	errout("Msession Ready");
	globals.smgr->Run(globals.listeners);
	globals.smgr->PostRun();
	errout("Msession Shutdown");

	delete globals.smgr;
	return 0;
}
