/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2002 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.
 
    You should have received a copy of the GNU General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to professionally license this library, the author
    can be reached at info@mohawksoft.com
*/ 
// #define TRACE_DEBUG

#include "settings.h"
#ifndef _MSESSIOND_
#include "session.h"
#include "reqclient.h"

#ifdef MSESSION_INTERNAL
// Internal request handler prototype.
// Functions may be replaced.
typedef int (*ReqHandler)(REQB **preq);


#endif

typedef int (*SESSIONPLUG_FN)(void *pdata,void*p1,void*p2);

enum REQ_TYPES
{
	MREQ_CTL=2,
	MREQ_SETVAL,
	MREQ_GETVAL,
	MREQ_CREATE,
	MREQ_DROP,
	MREQ_GETALL,
	MREQ_FIND,
	MREQ_COUNT,
	MREQ_FLUSH,
	MREQ_SLOCK,
	MREQ_SUNLOCK,
	MREQ_TIMEOUT,
	MREQ_INC,
	MREQ_DATAGET,
	MREQ_DATASET,
	MREQ_LIST,
	MREQ_LISTVAR,
	MREQ_UNIQ,
	MREQ_RANDSTR,
	MREQ_PLUGIN,
	MREQ_CALL,
	MREQ_SERIALIZE,
	MREQ_RESTORE,
	MREQ_EXEC,
	MREQ_LAST,
// These are handled by the server, not the library.
	MREQ_ADMIN = REQ_INTERNAL_END,
};

enum MREQ_PLUGIN_MSG
{
	MREQ_PLUGIN_PULSE=PLUGMSG_START,
	MREQ_PLUGIN_SESSION_FLUSH,
	MREQ_PLUGIN_SESSION_CREATE,
	MREQ_PLUGIN_SESSION_INIT,
	MREQ_PLUGIN_SESSION_DESTROY,
	MREQ_PLUGIN_ESCAPE,
	MREQ_PLUGIN_RELAY,
	MREQ_PLUGIN_SERIALIZE,
	MREQ_PLUGIN_RESTORE,
	MREQ_PLUGIN_SESSION_SETDATA,
	MREQ_PLUGIN_SESSION_SETVALUE,
	MREQ_PLUGIN_LAST
};
enum MSESSION_ADMIN
{
	MADMIN_SHUTDOWN,
	MADMIN_LOADDLL,
	MADMIN_SETSERIAL,
	MADMIN_SETVERBOSE,
	MADMIN_SETMAXTHREADS,
	MADMIN_SERIALIZE,
	MADMIN_AUTOSAVE,
	MADMIN_AUTODROP,
	MADMIN_STATUS
};

#define REQ_CTL_EXIST  0
#define REQ_CTL_TTL    1
#define REQ_CTL_AGE    2
#define REQ_CTL_TLA    3
#define REQ_CTL_CTIM   4
#define REQ_CTL_TOUCH  5
#define REQ_CTL_NOW    6

/** During a serialize operation, this structure is used to maintain
    state within the plugin */
typedef struct _plugin_serialize
{
	int  		cmd;		// Command to the plugin
	Session * 	session;	// Session, or NULL
	Boolean		shutdown;
	int  		n_var1;		// plugin integer variable 1
	int  		n_var2;		// plugin integer variable 2
	void *		vp_var1;	// plugin pointer variable 1
	void *		vp_var2;	// plugin pointer variable 2
}SERIALIZE_PARAM;

/** During set session data and set session value operatings, this 
    structure is sent to the plugin */
typedef struct _plugin_setdata
{
	Session *session_obj;
	char *	session_str;
	char *	name;
	char *	value;
	void *  plugin;
}SESSION_DATA_PARAM;

/** During set list walking, this structure is passed around. */
typedef struct ListWalk
{
	REQB	*reqb;
	char	*name;
	char 	*value;
	int 	cb;
	int	count;
}LISTWALK;

/** Structure to walk session list for flushing. */
typedef struct _flushWalk
{
	MNode *mn;
	int count;
	int now;
	int age;
	int oldest;
	int max;
	
}FLUSHWALK;


#define PLUGIN_SERIALIZE_BEGIN		1
#define PLUGIN_SERIALIZE_SESSION	2
#define PLUGIN_SERIALIZE_END		3

// Private session flags
#define MSESS_FIXED	0x01


#ifdef MREQ_INTERNAL
extern "C" ReqHandler MsessionReplaceHandler(ReqHandler handler, int ndx);
extern "C" int MsessionHandleRequest(REQB **preq);
extern "C" int ExecReqDirect(REQB **ppreq, int stat, char *session, char *name, char *value, int param);
extern "C" int ExecReqMultiDirect(REQB **ppreq, int stat, char *session, int n, char **pairs, int param);
#endif

#ifdef MSESSION_INTERNAL

// GROW_THREADS allows the thread pool to grow under load.
// Comment out this setting to disable this feature.
#define GROW_THREADS	TRUE

// SHRINK_THREADS allows the thread pool to shrink when
// the load reduces. 
// Comment out this setting to disable this feature.
#define SHRINK_THREADS	TRUE

#define MSESSION_VERSION	"1.21 [" __DATE__ "]"


extern void InitPluginAPI(void);
extern Boolean LoadFunctionMod(char *optarg);
extern void RemoveSession(Session *sess);
extern void DeleteSession(Session *sess);


extern	Boolean LoadFunctionMod(char *optarg);
extern	Boolean LoadPersonality(char *optarg);
extern	void errout(char *str);
extern	void errprintf(char *fmt, ...);



class SessionServer;
// The server thread manager. Only one of these is created.
class SServerMgr : public MThreadMgr
{
	protected:
	friend class SessionServer;
	MSock *m_sock;
#ifdef ENABLE_MTSVR
	MMTServer m_mtsvr;
#endif
	time_t	m_lastClean;
	int	m_cleanInterval;
	int 	m_deadTimeout;
	int	m_oldestSession;
	int 	m_last;
	int 	m_lastrpt;
	int	m_freeThreads;
	int	m_threadStat;
	int	m_pps;
	int 	m_exit_stat;

	REQB	*m_reqb;
	AtomicCounter	m_ac;			// Count home many threads are busy
	AtomicCounter	m_extraThreads;		// How many extra threads are running
	AtomicCounter	m_count;		// Count of connections

	public:
	SServerMgr(MSock *sock);
	~SServerMgr();
	MThread *NewThread(void);
	Boolean Pulse(void);
	void	setCleanInterval(int secs);
	void	setDeadTimeout(int secs);
	int	Interp(MThread *mth);
	void	Exit(void);
	void	Init(void);
	void	PreRun(void);
	void	PostRun(void);
	int	GetFreeThreads(void)	{ return m_freeThreads; 	}
	void	AddFreeThread(void)	{ ATOMIC_INC(m_freeThreads);	}
	int	GetThreadStat(void)	{ return m_threadStat;		}
	int	GetLastCount(void)	{ return m_last;		}
	int	GetPPS(void)		{ return m_pps; 		};
};


// Shared variables in msession server

typedef struct _msessiond_globals_
{
	
	int		version;		// The version of the globals structure
	int		argc;			// Command line argument count
	char **		argv;			// Command line argument values
	MConfigFile *	mcfg;			// Configuration file parameters
	SServerMgr *	smgr;			// Server manager
	SessionMgr *	session_array;		// Where we keep all the session nodes
	MPlugin	   	session_plugin;		// Personality Plugin
	MPlugMgr   	session_plugmgr;	// Functions plugins
	MTArray<MPlugin *> functions;		// A list of all the function plugin modules
	MRWMutex 	session_mutex;		// System Read/Write Mutex
	int		req_timeout;		// Default timeout value for ReadRequestTimeout
	int		verbose;		// Do we spit out lots of errors?
	int		maxThreads;		// Maximum number of threads
	int		listeners;		// Minimum number of threads
	int		autosave;		// Number of minutes between saves
	int		autodrop;		// Autmoatically drop socket conections
	int		dropped;		// Number of dropped connections
	int		fAllowExec;		// Allow execute function
	int		dead;			// Timeout value for dead sessions
	int		interval;		// Cleanup interval
	int		port;			// Port on which system listens
	int		queue;			// The listen(queue) value
	int		nhash;			// The size of the session hash table
	int		lock;			// timout for locked sessons
	int		maxflush;		// Maximum flush per session
	int		personality;		// Personality count
	int		sessionlen;		// Length of the session string
	char *		szfilePath;		// File path
	char *		szPersonality;		// Personality module
	char *		globals;		// The name of the "global" session, if one.
	char *		serialize;		// This string is passed to the personality plugin
	char *		name;			// Name of msession daemon
	int 		(*MsessionHandleRequest)(REQB **preq);
	ReqHandler 	(*MsessionReplaceHandler)(ReqHandler newHandler, int ndx);
	REQB *		(*AllocateRequestBuffer)(unsigned  size);
	void 		(*FreeRequestBuffer)(REQB *req);
	int 		(*FormatRequest)(REQB **ppreq, int stat, char *session, char *name, char *value, int param);
	int 		(*FormatRequestMulti)(REQB **ppreq, int stat, char *session, int n, char **pairs, int param);
	int		(*CallPluginFunction)(char *fnname, void *p1, void *p2);
	PlugFn		(*GetPluginFunction)(char *fname, PlugFn fnDefault);
	int 		(*ExecuteRequest)(REQB **ppreq, int stat, char *session, char *name, char *value, int param);
	int 		(*ExecuteRequestMulti)(REQB **ppreq, int stat, char *session, int n, char **pairs, int param);
	REQB *		(*SizeRequestBuffer)(REQB *req, unsigned int size);
	SessionMgr *	(*GetSessionMgr)(void);
	void 		(*errout)(char *str);
	void		(*errprintf)(char *fmt, ...);
	void		(*DeleteSession)(Session *);
	void		(*RemoveSession)(Session *);

}MSESSION_GLOBALS;

extern MSESSION_GLOBALS	globals;
// One of these is created for each thread specified on the command line.

#ifndef ENABLE_MTSVR
class SessionServer : public MServerThread
#else
class SessionServer : public MTServerThread
#endif
{
	protected:
	MSock *		m_sock;
	REQB	*	m_reqb;
	SServerMgr *	m_mgr;
	Boolean HandleAdmin(MSock *ms, REQB ** reqb);
	Boolean HandlePersistent(MSock *ms, REQB **reqb);
	Boolean HandleSocket(MSock *ms);
	public:

#ifndef ENABLE_MTSVR
	SessionServer(SServerMgr *mgr, MSock *ms);
#else
	SessionServer(SServerMgr *mgr, MMTServer *mtsvr);
#endif
	~SessionServer(void);
	Boolean FilterRequest(MSock *ms, REQB ** reqb);
};
#endif
#endif
