/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2002 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.
 
    You should have received a copy of the GNU General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to professionally license this library, the author
    can be reached at info@mohawksoft.com
*/ 

//
// This is a sample plugin to show how to use the plugin features of msessiond
//
#include "plugapi.h"


MTArray <PSTR> g_flush;

//
// The plugin does not share the internal data of msession. Therefor calls made 
// to msession MUST go through the msession function pointers.
MSESSION_EXPORTS *msessiond;

// The plugin defines a jump table for all the handlers.
static SESSIONPLUG_FN handlers[MREQ_PLUGIN_LAST];

// The default plugin function, which does nothing, should return the version.
static int plugDefault(void *pdata, void *p1, void *p2)
{
	return MPLUGIN_VER;
}

// RESTORE
// This call is used at startup of the msession daemon to allow a plugin to
// recreated session information saved from the SERIALIZE function.
// p1 is the request buffer double pointer (REQB **)
// p2 is unused.
static int plugRestore(void *pdata, void *p1, void *p2)
{
	REQB **preq = (REQB **) p1;
	fprintf(stderr,"Restore called\n");
	(*preq)->req.stat = REQ_OK;
	return (*preq)->req.stat;
}

// A session is a container of MNode. An MNode is a class which represents
// a name/value pair. To get all the name/value pairs in a session, they must
// be enumerated with the Session's Enum function, and a callback function
// which gets called for each value in the session.
Boolean sessionCallBack(MNode *mn, void *p1, void *p2)
{
	char * name = (char *)mn->KeyGet();
	char * value = (char *)mn->DataGet();
	fprintf(stderr,"\t%s=%s\n", name, value);
	return TRUE;
}

// SERIALIZE
// This call is used to save the current state of the system.
// p1 is the request buffer double pointer (REQB **)
// p2 is SERIALIZE_PARAM structure.
// The time between PLUGIN_SERIALIZE_BEGIN and PLUGIN_SERIALIZE_END
// the system is locked, so be fast.
// The requst buffer may be used to retrieve parameters passed to the
// msessiond daemon.
static int plugSerialize(void *pdata, void *p1, void *p2)
{
	REQB **preq = (REQB **) p1;
	SERIALIZE_PARAM *psp = (SERIALIZE_PARAM *) p2;

	if(psp->cmd == PLUGIN_SERIALIZE_BEGIN)
	{
		// Initialize your save handler here
		fprintf(stderr,"Begin serialize\n");
	}
	else if(psp->cmd == PLUGIN_SERIALIZE_END)
	{
		// Finish up here
		fprintf(stderr,"Finish serialize\n");
		(*preq)->req.stat = REQ_OK;
	}
	else
	{
		// You will be passed each and every session class.
		// Write it to a file.
		char *session = (char *) psp->session->KeyGet();
		char *data = (char *) psp->session->DataGet();
		fprintf(stderr,"Session Name: %s=%s\n", session, data);
		psp->session->Enum(sessionCallBack, (void *)session, NULL);
	}
	return TRUE;
}
// PULSE
// This call is used to pass time to the plugin on
// a regular basis.
// p1 and p2 are not currently used
// Pulse is called about once a second, AFTER house
// cleaning (flushing dead sessions, if any) is done.
// You can do some processing here, nothing is locked.
// Returning FALSE here will stop the server
// If you wish to do extensive processing on flushed
// sessions, you should make a queue of sessions
// to be flushed, and return FALSE from plugFlush(...)
// Then, in this call doo all the analysis of the sessions
// and delete them here.
static int plugPulse(void *pdata, void *p1, void *p2)
{
	return TRUE;
}
// FLUSH
// This call is made to check with the plugin if it is OK
// to flush an expired session.
// When p1 is the flush structure, p2 is the session object
// This is called when the session list is write locked!!
// No sessions can be created or destroyed, and all web servers
// will be waiting for you. Don't be a hero, get in and get out.
// return TRUE if it is OK to flush session, FALSE to forbid it.
// 
// When p1 is the flush structure and p2 is NULL, all the items 
// have been checked and you are free to flush the objects.
//
static int plugFlush(void *pdata, void *p1, void *p2)
{
	if(p1)
	{
		PSTR p = phmstrdup((char *)p1);
		g_flush.Add(p);
	}
	else
	{
		if(g_flush.count())
		{
			int count = g_flush.count();
			for(int i=0; i < count; i++)
				fprintf(stderr,"%s flushed\n", g_flush[i]);
			g_flush.trunk(0);
		}
	}
	return TRUE;
}

// CREATE
// This call checks if a session can be created.
// p1 is the session name, p2 is not currently used
// This is called prior to locking the session list.
// A web server is waiting for you to finish.
// return TRUE if it is OK to create the  session, FALSE to forbid it.
static int plugCreate(void *pdata, void *p1, void *p2)
{
	return TRUE;
}
// INIT
// This call allows a plugin to initialize a new session.
// p1 is the session name, p2 is the session object
// This is called right after session creation.
// The session list is read locked during this call. 
// No session creations or deletions can run
// Make it quick.
// return value is ignored
static int plugInit(void *pdata, void *p1, void *p2)
{
	return TRUE;
}

// DESTROY
// This call checks if a session can be deleted.
// p1 is the session name, p2 is the session object
// This is called prior to locking the session list.
// A web server is waiting for you to finish.
// return TRUE if it is OK to destroy the  session, FALSE to forbid it.
static int plugDestroy(void *pdata, void *p1, void *p2)
{
	return TRUE;
}

// ESCAPE
// This is a generic entry point into the plugin from
// a client.
// p1 is a REQB **
// The php function call is:
// msesssion_plugin(sesssion, param1, param2)
// param1 is sent as name
// (param1 is converted to an integer and stored as req->param)
// param2 is sent as value
// The session table is not locked, however
// a web server is waiting for you.
// This is different than the other plugin calls in that it is a core
// msession request handler.
// param1 should be a number if you intend to use the plugin call to 
// implement more than one function because it is converted to an int
// on the web server.
static int plugEscape(void *pdata, void *p1, void *p2)
{
	
	REQB **preq = (REQB**)p1;
	char *session	=(char *)&(*preq)->req.datum[(*preq)->req.session];
	char *param1	=(char *)&(*preq)->req.datum[(*preq)->req.name];
	char *param2	=(char *)&(*preq)->req.datum[(*preq)->req.value];
	int nparam1	= (*preq)->req.param;

	fprintf(stderr,"Call from sesion %s, %s [%d], %s\n", session, param1, nparam1, param2);

	sprintf((*preq)->req.datum, "Time = %ld\n", time(0));

	(*preq)->req.len = strlen((*preq)->req.datum)+1;
	(*preq)->req.stat = REQ_OK;
	return (*preq)->req.stat;
}

// RELAY
// This is called prior to all msessiond actions. It allows
// a plugin to take complete control over the daemon
// p1 is a REQB **
// This function is intended to manage a hot-spare msessiond
// It is not built yet, it is just in the definition stage.
// This function should be FAST!! it will be called on every request
// prior to being performed on the local server.
// returning false here, means that the remote msessiond system failed
// to perform this request and should not be performed on the local
// system.
// functions like msession_randstr, need not be executed on the
// local system, the remote system already did it.
//
// functions like msession_uniq should be translated to an msession_create
// call. Let the slave system create the uniq session name, and let the 
// local system have a copy.
//
static int plugRelay(void *pdata, void *p1, void *p2)
{
	return TRUE;
}

// FN_HANDLER()
// This is called by msessiond for every plugin call.
// The handlers table is a thunking layer.
//
// This is called for all the plugin events.
// Parameters: 
//	void *pdata	This is what was returned from init
// 	int msg		This is the message to act on
//	void *p1, 	Extra data
//	void *p2	More data
FN_HANDLER(default)
{
	return (msg>=0 && msg < MREQ_PLUGIN_LAST)?(handlers[msg])(pdata,p1,p2) :
				plugDefault(pdata,p1,p2);
}
// FN_INIT()
// This is called to initialize the plugin.
//
FN_INIT(default)
{
	int i;

	for(i=0; i < MREQ_PLUGIN_LAST; i++)
		handlers[i] = plugDefault;

	handlers[MREQ_PLUGIN_PULSE]=plugPulse;
	handlers[MREQ_PLUGIN_SESSION_FLUSH]=plugFlush;
	handlers[MREQ_PLUGIN_SESSION_CREATE]=plugCreate;
	handlers[MREQ_PLUGIN_SESSION_INIT]=plugInit;
	handlers[MREQ_PLUGIN_SESSION_DESTROY]=plugDestroy;
	handlers[MREQ_PLUGIN_ESCAPE]=plugEscape;
	handlers[MREQ_PLUGIN_RELAY]=plugRelay;
	handlers[MREQ_PLUGIN_SERIALIZE]=plugSerialize;
	handlers[MREQ_PLUGIN_RESTORE]=plugRestore;
	
	msessiond = (MSESSION_EXPORTS *) exports;

	fprintf(stderr,"Sample Personality plugin loaded\n");
	return NULL;
}
// FN_SHUTDOWN()
// This is called at shutdown.
// This is NOT reliable, if the user kills the program, this may
// not be called.
FN_SHUTDOWN(default)
{
	fprintf(stderr,"Sample Personality plugin unloaded\n");
}
