/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2002 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.
 
    You should have received a copy of the GNU General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to professionally license this library, the author
    can be reached at info@mohawksoft.com
*/ 

// This is a SQL Cache plugin. It provides a very active connection
// between msession and a SQL database. It is designed to create 
// persistent sessions which are stored in a SQL database


#include <stdio.h>
#include "plugapi.h"

// Set this to TRUE if you wish the database will always
// provide upto data session information.
// If you define this, then msession provides little
// more than a read-only cache and session controler.
Boolean g_full_mirror = 0;

// Change this to change the default number of SQL connections
// allocated. Either that or set it in the serializer parameter
#define DEF_QUEUE_SIZE	4

// Used to call back into msessiond
MSESSION_EXPORTS *msessiond;

// Use a SQL connection queue to manage connections.
MSqlQueue g_sql_queue;

// If TRUE, msession does a hard SQL syncronization (SLOW!)
Boolean fSerial=FALSE;

// If TRUE, msession does reports "ownership" of sessions
Boolean fOwner=FALSE;

// If system is serializing, wait
static void WaitForSerial()
{
	while(fSerial == TRUE)
		YieldTimeSlice();
}

#define DEF_NON_OWNER	"none"

// Sets the owner of a session in the msession_status table
// session is the session name.
// owner is the name of the session server, or NULL to release 
// ownership
Boolean SetOwner(MSql *conn, char *session, char *owner)
{
	static char szQueryOwner[]=
	"select owner from msession_status where session = '%s'";
	static char szUpdStat[]=
	"update msession_status set logtime = now(), owner = '%s' where session = '%s'";
	static char szInsStat[]=
	"insert into msession_status (session, logtime, owner) values ('%s', now(), '%s')";

	// If the feature is not enabled, don't even try
	if(!fOwner)
		return TRUE;
	
	// Get the right owner
	char *sqlowner = (owner) ? owner : (char *)DEF_NON_OWNER;
	// Default is that we can't set the owner
	Boolean fCanOwn = FALSE;

	// I need to check of I can own this session
	if (sqlowner != DEF_NON_OWNER)
	{
		fCanOwn = FALSE;
		SQLRESULT res = conn->ExecfResult(szQueryOwner, session);
		if(res)
		{
			int n = conn->RowsCount(res);
			if(n)
			{	// It was found in the status table
				char * pg_owner = conn->ColumnValue(res,0,0);
				if(!strcasecmp(pg_owner, DEF_NON_OWNER) || !strcasecmp(pg_owner, owner))
					fCanOwn = TRUE;
			}	
			else
			{	// It was not found in the status table
				fCanOwn = TRUE;
			}
			conn->FreeResult(res);	
		}
		else
			fCanOwn = TRUE;

	}
	else
		fCanOwn = TRUE;

	if(fCanOwn)
	{
		if(msessiond->verbose > 1)
			msessiond->errprintf("Setting owner of %s to %s", session, sqlowner);
		if(conn->Execf(szUpdStat, sqlowner, session) != SQLERR_OK_AFFECTED)
			conn->Execf(szInsStat, session, sqlowner);
	}
	return fCanOwn;
}


// Typical SQL strategy, try first to update, if that succeeds you are done.
// if it does not succeed, that means the record is not in the database, so
// we insert it.
//
/* Saves a session into the msession_sessions table */
static void savesession(MSql *conn, char *session,char *value, int timeout, Boolean fActive=TRUE)
{
	static char szUpdfmt[]=
	"update msession_sessions set value='%s', timeout= %d,  lasthit=now() where session='%s'";
	static char szInsfmt[]=
	"insert into msession_sessions (session,value,timeout, lasthit) values ('%s','%s',%d, now());";

	if(conn->Execf(szUpdfmt, value, timeout, session) >=  SQLERR_OK_NONE)
		conn->Execf(szInsfmt, session, value, timeout);

	// If we record owner, we need to make sure 
	if(fOwner && fActive == FALSE)
	{
	}
}

/* Saves session values to the msession_values table */
static void savesessionvalue(MSql *conn, char *session, char * name, char *value)
{
	static char szUpdfmt[]=
		"update msession_values set value = '%s' where name= '%s' and session='%s';";
	static char szInsfmt[]=
		"insert into msession_values (session,name,value) values ('%s','%s','%s');";

	if(conn->Execf(szUpdfmt, value, name, session) >= SQLERR_OK_NONE)
		conn->Execf(szInsfmt, session, name, value);
}

/* Get session values from the database for a session */
static int getsessionvalues(MSql *conn, char *session_str, Session *session_obj)
{
	char szfmtData[]="select value, timeout from msession_sessions where session = '%s'";
	char szfmtValue[]="select name, value from msession_values where session = '%s'";

	SQLRESULT res=NULL;

	res = conn->ExecfResult(szfmtData, session_str);

	if(!res)
		return REQ_ERR;

	if(conn->RowsCount(res) < 1)
	{
		conn->FreeResult(res);
		return REQ_ERR;
	}

	char *value = conn->ColumnValue(res, 0, 0);

	if(!value)
	{
		conn->FreeResult(res);
		return REQ_ERR;
	}

	session_obj->SetSessionData(value, NULL);

	conn->FreeResult(res);

	res =conn->ExecfResult(szfmtValue, session_str);

	if(res)
	{
		int n = conn->RowsCount(res);

		for(int i=0; i < n; i++)
		{
			char *name = conn->ColumnValue(res,i, 0);
			char *value = conn->ColumnValue(res,i, 1);

			if(!name || !value)
				break;

			session_obj->SetValue(name,value);
		}
		conn->FreeResult(res);
	}
	return REQ_OK;
}

// Called for each value/pair MNode in a session
// p1 is SQL connection
// p2 has session name
static Boolean sessionCallBack(MNode *mn, void *p1, void *p2)
{
	char * name = (char *)mn->KeyGet();
	char * value = (char *)mn->DataGet();
	savesessionvalue((MSql *)p1, (char *)p2, name, value) ;
	return TRUE;
}

// The PLUGIN_SERIALIZE routine is called at shutdown, and optionally during
// operation, to save the current state of the msession daemon.
// In the SERIALIZE_PARAM structure:
// n_var1 is the current serialization id
// n_var2 is the previous serialization id
// vp_var1 is the PostgreSQL connection
// vp_var2 is unused.
static int PLUGIN_SERIALIZE(void *p1, void *p2)
{
	SERIALIZE_PARAM *psp = (SERIALIZE_PARAM *) p2;
	REQB **preq = (REQB **) p1;
	MSql *conn = (MSql *)psp->vp_var1;

	if(psp->cmd == PLUGIN_SERIALIZE_BEGIN)
	{
		// Called first to initialize the plugin's serialization function.
		// In this case, setup the database.
		conn = g_sql_queue.Get();

		if(!conn)
		{
			(*preq)->req.len = strlen((*preq)->req.datum)+1;
			return FALSE;
		}
		psp->vp_var1 = (void *)conn;

		psp->n_var2 = 0;
		// conn->Exec("BEGIN");
		fSerial = TRUE;
		return TRUE;
	}
	else if(psp->cmd == PLUGIN_SERIALIZE_END)
	{
		// Called last to finish the plugin's serialization function.
		// In this case, commit the transaction, and if all is well, delete the old data.
		//
		// conn->Exec("COMMIT");
		fSerial = FALSE;
		if(psp->n_var2)
		{
			// delete old values
			conn->Exec("insert into msession_log select * from msession_status");
			(*preq)->req.stat = REQ_OK;
		}
		g_sql_queue.Release(conn);
	}
	else if(psp->cmd == PLUGIN_SERIALIZE_SESSION)
	{
		// Called for each valid session in the daemon. It is the job of the
		// plugin to enumerate and save all the variables in a session.

		assert(psp->session);
		int timeout = (psp->session->TestFlag(MSESS_FIXED)) ? -1 : psp->session->GetTimeout();

		char *session = (char *) psp->session->KeyGet();
		char *data = (char *) psp->session->DataGet();

		if(!data) 
			data = "";
		savesession(conn, session, data, timeout);
		psp->session->Enum(sessionCallBack, (void *) psp->vp_var1, (void *)session);
		if(psp->shutdown)
			SetOwner(conn, session, NULL);
	}
	else
	{
		fSerial = FALSE;
		(*preq)->req.stat = REQ_ERR;
		(*preq)->req.param = REQE_UNKNOWN;
		if(psp->vp_var1)
			g_sql_queue.Release(conn);
		msessiond->errprintf("Unknown serialize command %d", psp->cmd);
		return FALSE;
	}
	return TRUE;
}

// The PLUGIN_RESTORE routine is called at startup to restore the state of the
// msessiond daemon.
static int PLUGIN_RESTORE(void *p1, void *p2)
{
	static char szSqlPlugCmd[]="SQLP:";

	REQB **preq = (REQB **) p1;
	int nque = msessiond->listeners+1;

	int len = strlen((*preq)->req.datum)+1;
	char *sqlstr = (char *)ALLOCA(len);
	memset(sqlstr, 0, len);
	strcpy(sqlstr, (*preq)->req.datum);

	assert(sqlstr);

	char *param = strstr(sqlstr, szSqlPlugCmd);

	if(param)
	{
		*param=0; 	// Cap off our prameters from the SQL connect string;
		param += 5;	// Skip over key string

		while(*param)
		{
			char	p = *param++;
			int 	n = 0;
			if(*param == ':')
			{
				param++;
				n = atoi(param);
				while(isdigit(*param) && param)
					param++;
			}
			switch (p)
			{
				case 'm': // Mirror mirror level
					g_full_mirror = n;
					break;
				case 'q': // Queue size
					nque = n;
					break;
				case 'o': // Set owner
					fOwner = n;
					break;
			}
		}
	}
	if(!g_sql_queue.Init(sqlstr, nque))
	{
		msessiond->errprintf("No SQL connection!!");
		(*preq)->req.stat = REQ_ERR;
		(*preq)->req.param = REQE_UNKNOWN;
		return (*preq)->req.stat;
	}
	else
		(*preq)->req.stat = REQ_OK;

	if(msessiond->verbose)
	{
		msessiond->errprintf("Mirror level %d", g_full_mirror);
		msessiond->errprintf("Queue size: %d (%d)", nque, g_sql_queue.GetCount());
		msessiond->errprintf("Owner Symantics %d", fOwner);
	}
	return (*preq)->req.stat;
}

static int PLUGIN_SESSION_INIT(void *p1, void *p2)
{
	int result = 0;
	char *session_str = (char *)p1;
	Session * session_obj = (Session *)p2;

	MSql *conn = g_sql_queue.Get();

	if(conn != NULL)
	{
		if(SetOwner(conn, session_str, msessiond->name))
		{
			getsessionvalues(conn, session_str, session_obj);
			result = REQ_OK;
		}
		else
			msessiond->errprintf("Couldn't set owner on %s", session_str);
		g_sql_queue.Release(conn);

	}
	return result;
}
// This is a time sync and will affect performance.
// Called for each node prior to deletion but after
// removal.
// p1 is the session string
// p2 is the session object
//
static MTArray <Session *> s_flushed_sessions(DEF_MAXFLUSH);

static int PLUGIN_SESSION_FLUSH(void *p1, void *p2)
{
	Session *session_obj = (Session *)p2;
	FLUSHWALK *pflw = (FLUSHWALK *)p1;
	if(!session_obj)
	{
		// This is called when the session tree is unlocked
		// you can take time some time.
		int n = s_flushed_sessions.count();
		if(n)
		{
			struct tms tmsdata;
			clock_t ct_start = times(&tmsdata);

			MSql *conn = g_sql_queue.Get();
			for(int i=0; i < n; i++)
			{
				session_obj = s_flushed_sessions[i];
				char *session_str = (char *)session_obj->KeyGet();
				int timeout =(session_obj->TestFlag(MSESS_FIXED))? -1 : session_obj->GetTimeout();
				char *data = (char *) session_obj->DataGet();

				if(!data) 
					data = "";
				savesession(conn, session_str, data, timeout);
				session_obj->Enum(sessionCallBack, conn, (void *)session_str);
				SetOwner(conn, session_str, NULL);
			}
			g_sql_queue.Release(conn);
			s_flushed_sessions.trunk(0);
			clock_t ct_end = times(&tmsdata);
			clock_t elapsed = (ct_end - ct_start)*10;
			double dtime = elapsed;
			dtime /= n;

			if(msessiond->verbose)
				msessiond->errprintf("Saved %d sessions to database in %dms, (%.2fms per session)", 
					n,elapsed, dtime);
		}
		return TRUE;
	}
	else
	{
		msessiond->RemoveSession(session_obj);
		s_flushed_sessions.Add(session_obj);
		pflw->count++;
		return FALSE; // Do not flush!
	}
}

// This is called to delete a session
// p1 is the session string
// p2 is the session object
static int PLUGIN_SESSION_DESTROY(void *p1, void *p2)
{
	WaitForSerial();
	MSql *conn = g_sql_queue.Get();
	char *session_str = (char *)p1;
	Session *session_obj = (Session *)p2;
	int timeout = (session_obj->TestFlag(MSESS_FIXED)) ? -1 : session_obj->GetTimeout();

	char *data = (char *) session_obj->DataGet();

	if(!data) 
		data = "";
		
	savesession(conn, session_str, data, timeout);
	session_obj->Enum(sessionCallBack, conn, (void *)session_str);
	SetOwner(conn, session_str, NULL);
	g_sql_queue.Release(conn);
	return REQ_OK;
}
// This is called to set the sessions main data string
// p1 is the SESSION_DATA_PARAM
// p2 is NULL
static int PLUGIN_SESSION_SETDATA(void *p1, void *p2)
{
	if(g_full_mirror > 0)
	{
		WaitForSerial();
		SESSION_DATA_PARAM *pdp = (SESSION_DATA_PARAM *)p1;
		MSql *conn = g_sql_queue.Get();

		int timeout = (pdp->session_obj->TestFlag(MSESS_FIXED)) ? 
				-1 : pdp->session_obj->GetTimeout();

		if(!pdp->value) 
			pdp->value = "";

		savesession(conn, pdp->session_str, pdp->value, timeout);

		g_sql_queue.Release(conn);
	}
	return REQ_OK;
}

// This is called to set the session name/value pairs
// p1 is the SESSION_DATA_PARAM
// p2 is NULL
static int PLUGIN_SESSION_SETVALUE(void *p1, void *p2)
{
	if(g_full_mirror > 0)
	{
		WaitForSerial();
		SESSION_DATA_PARAM *pdp = (SESSION_DATA_PARAM *)p1;
		MSql *conn = (MSql *)g_sql_queue.Get();

		if(!pdp->value) 
			pdp->value = "";

		savesessionvalue(conn, pdp->session_str, pdp->name, pdp->value);

		g_sql_queue.Release(conn);
	}
	return REQ_OK;
}

FN_HANDLER(default)
{
	return MPLUGIN_VER;
}

FN_INIT(default)
{
	PLUGFN_OPEN(entries)
	PLUGFN_DECL(PLUGIN_SESSION_INIT)
	PLUGFN_DECL(PLUGIN_SERIALIZE)
	PLUGFN_DECL(PLUGIN_RESTORE)
	PLUGFN_DECL(PLUGIN_SESSION_FLUSH)
	PLUGFN_DECL(PLUGIN_SESSION_DESTROY)
	PLUGFN_DECL(PLUGIN_SESSION_SETDATA)
	PLUGFN_DECL(PLUGIN_SESSION_SETVALUE)
	PLUGFN_CLOSE

	msessiond = (MSESSION_EXPORTS *) exports;
	*imports = (PLUGFN *)entries;

	if(msessiond->verbose)
		msessiond->errprintf("SQL Cache Persistent Session plugin loaded");
	return NULL;
}

FN_SHUTDOWN(default)
{
}
