/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2002 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.
 
    You should have received a copy of the GNU General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to professionally license this library, the author
    can be reached at info@mohawksoft.com
*/ 

#define MSESSION_INTERNAL
#define SVR_INTERNAL
#define SESSION_INTERNAL
#define MREQ_INTERNAL

#include "msessiond.h"


SServerMgr::SServerMgr(MSock *sock) 
#ifdef ENABLE_MTSVR
		: m_mtsvr(sock)
#endif

{
	m_sock = sock;
	m_lastClean = time(0);
	m_cleanInterval = DEF_INTR;
	m_deadTimeout = DEF_DEAD;
	m_oldestSession = m_lastClean;
	m_reqb = AllocateRequestBuffer(DEF_REQBLEN);
	m_last=0;
	m_pps = 0;
	m_lastrpt=0;
	m_fContinue=TRUE;
	m_freeThreads = globals.maxThreads;
	m_exit_stat = 0;
}
SServerMgr::~SServerMgr()
{
	FreeRequestBuffer(m_reqb);
}

MThread *SServerMgr::NewThread(void)
{
	if(globals.verbose>1)
		mferr.mfprintf("New Thread\n");
#ifndef ENABLE_MTSVR
	MThread *thread = (MThread *) new SessionServer(this,m_sock);
#else
	MThread *thread = (MThread *) new SessionServer(this,&m_mtsvr);
#endif

	// Needs a HUGE, stack because ODBC is a pig!
	thread->SetStackSize(1024*1024);
	return thread;
}

void SServerMgr::setCleanInterval(int secs)
{
	m_cleanInterval = secs;
}
void SServerMgr::setDeadTimeout(int secs)
{
	m_deadTimeout = secs;
}
// Let server threads die
int SServerMgr::Interp(MThread *mth)
{
	return TINTERP_THEND;
}

void SServerMgr::Init(void)
{
	ExecReqDirect(&m_reqb, MREQ_RESTORE, globals.serialize, "", "", 0);
	if(globals.globals)
		ExecReqDirect(&m_reqb, MREQ_CREATE, globals.globals, "", "", -1);

}

void SServerMgr::PreRun(void)
{
#ifdef ENABLE_MTSVR
	m_mtsvr.run(FALSE);
#endif
}

void SServerMgr::PostRun(void)
{
#ifdef ENABLE_MTSVR
//	m_mtdvr.stop();
#endif
}

void SServerMgr::Exit(void)
{
	if(!m_exit_stat)
	{
		sleep (1); // Let system settle down
		ExecReqDirect(&m_reqb, MREQ_SERIALIZE, globals.serialize, "", "", TRUE);
		globals.session_plugin.Close();
		globals.session_plugmgr.CloseAll();
		m_exit_stat = 1;
	}
}

// Called once a second 
Boolean SServerMgr::Pulse(void)
{
	struct tms tmsdata;
	time_t now = time(0);
	time_t elapsed = now - m_lastClean;
	int busythreads = m_ac.Get();
	static time_t last_save = 0;

	if(!m_fContinue)
	{
		Exit();
		return FALSE;
	}

	int rpt = now - m_lastrpt;

	int gcount = m_count.Get();
	int n = gcount - m_last;
	m_pps = n;
	m_last = gcount;
	m_lastrpt = now;

	if(globals.verbose)
	{
		if(rpt)
		{
			if(n)
			{

				double persec = n;
				persec /= rpt;
#ifdef ENABLE_MTSVR
	m_threadStat = m_mtsvr.NumWaiting();
	mferr.mfprintf("%s %d (%.2f per second) %d total, %d waiting, %d sessions\n",
		timestr(0),n,persec,m_last,m_threadStat,
		globals.session_array->count());
#else
	int numthreads = count();
	m_threadStat = busythreads;
	mferr.mfprintf("%s %d (%.2f per second) %d total, %d busy, %d sessions\n",
		timestr(0),n, persec,m_last,m_threadStat,
		globals.session_array->count());
#endif
			}
		}
	}
#ifndef ENABLE_MTSVR
	// No need to grow/shring threads with NTSVR
#ifdef GROW_THREADS
	if((numthreads - busythreads) < 2)
	{
		// Make sure there are a couple free threads at all times
		if(numthreads >= globals.maxThreads)
		{
			if(globals.verbose>1)
				mferr.mfprintf("You should increase the maximum number of threads, busy=%d, max=%d\n", 
					busythreads, globals.maxThreads);
		}
		else
		{
			/* Scale up number of threads, double the number as long as there is a load */
			int numnew = numthreads /* /2 */;
			int newcount = numthreads + numnew;
				
			if(newcount >= globals.maxThreads)
				numnew = globals.maxThreads - numthreads;
			if(numnew > 0)
				RunThreads(numnew);
		}
	}
#endif
#ifdef SHRINK_THREADS
	if(busythreads < (numthreads/2)) // Too many running
	{
		/* Scale down number of threads when there is no load, go slowly so as not to thrash too much */
#if 0
		int extra = numthreads/4;
#else
		int extra = 2;
#endif
		m_extraThreads.Set(((numthreads - extra) >= globals.listeners ) ? extra: 0);
	}
#endif
#endif

	// Calculate number of free threads.
	m_freeThreads = globals.maxThreads - busythreads;

	if(elapsed >= m_cleanInterval)
	{
		if(globals.verbose>1)
			mferr.mfprintf("%s Flushing Dead Sessions\n", timestr(0));

		if(now > (m_oldestSession + m_deadTimeout))
		{
			clock_t ct_start = times(&tmsdata);
			ExecReqDirect(&m_reqb, MREQ_FLUSH, "", "", "", m_deadTimeout);
			clock_t ct_end = times(&tmsdata);
			if(globals.verbose)
				mferr.mfprintf("%s %d sessions removed, %d active %d ms\n", timestr(0),
					m_reqb->req.param, globals.session_array->count(), 10*(ct_end - ct_start));
			if(m_reqb->req.param >= m_reqb->req.name)
			{
				if(globals.verbose)
					mferr.mfprintf("%s Max sessions flushed, start again in 1 sec.\n", timestr(0) );
			}
			else
			{
				m_oldestSession = m_reqb->req.session;
				m_lastClean = now;
			}
		}
		else if(globals.verbose)
		{
			m_lastClean = now;
			mferr.mfprintf("%s No expired sessions, %d active\n", timestr(0), globals.session_array->count());
		}
	}

	if(globals.autosave)
	{
		now = time(0);
		if(last_save == 0)
			last_save = time(0);
		else if((now - last_save) > (globals.autosave * 60))
		{
			ExecReqDirect(&m_reqb, MREQ_SERIALIZE, globals.serialize, "", "", FALSE);
			last_save = now;
		}
	}
	return globals.session_plugin.pcall(MREQ_PLUGIN_PULSE);
}
