/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2002 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.
 
    You should have received a copy of the GNU General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to professionally license this library, the author
    can be reached at info@mohawksoft.com
*/ 

#define MSESSION_INTERNAL
#define SESSION_INTERNAL
#define SVR_INTERNAL
#define MREQ_INTERNAL

#include "msessiond.h"

#ifndef ENABLE_MTSVR
SessionServer::SessionServer(SServerMgr *mgr, MSock *ms) : MServerThread(ms)
#else
SessionServer::SessionServer(SServerMgr *mgr, MMTServer *mtsvr) : MTServerThread(mtsvr)
#endif
{
	m_mgr = mgr;
	m_reqb = AllocateRequestBuffer(2048);
	assert(m_reqb);
}
SessionServer::~SessionServer()
{
	FreeRequestBuffer(m_reqb);
}

// Called for a persistent connection.
Boolean SessionServer::HandlePersistent(MSock *ms, REQB **reqb)
{
	int expire;
	TRACE_FUNCT("HandlePersistent");
	assert((*reqb)->req.stat == REQ_POPEN);

	expire = ((*reqb)->req.param) ? (*reqb)->req.param : globals.req_timeout;

	(*reqb)->req.stat = REQ_OK;

	WriteRequest(reqb, ms);

	// If using ENABLE_MTSVR, all connections are assumed 
	// persistent.
#ifndef ENABLE_MTSVR

	Boolean fContinue = TRUE;
	while(fContinue && m_mgr->GetContinue())
	{
		if(!ReadRequestTimeout(reqb, ms, expire))
		{
			if(globals.verbose)
				mferr.mfprintf("%s Socket timed out or closed\n", timestr(0));
			break;
		}
		if(globals.autodrop && m_mgr->GetFreeThreads() < 2)
		{
			ms->Shutdown(SHUT_RD);
			m_mgr->AddFreeThread();
			fContinue = FALSE;
			ATOMIC_INC(globals.dropped);
		}
		if(!FilterRequest(ms, reqb))
		{
			if((*reqb)->req.stat == REQ_PCLOSE)
			{
				TRACE_FUNCT("HandlePersistent:exit");
				fContinue = FALSE;
			}
			else
			{
				MsessionHandleRequest(reqb);
			}
			WriteRequest(reqb, ms);
		}
	}
	ms->Shutdown(SHUT_WR);
#endif
	return TRUE;
}
// Admin functions head here, and are handled based on param.
Boolean SessionServer::HandleAdmin(MSock *ms, REQB **reqb)
{
	int pn = atoi((*reqb)->req.datum);
	char *p = (*reqb)->req.datum;

	// Be optimistic, assume success
	(*reqb)->req.stat = REQ_OK;
	(*reqb)->req.len = 0;

	TRACE_FUNCT("HandleAdmin");
	switch((*reqb)->req.param)
	{
		case MADMIN_SHUTDOWN:
			mferr.mfprintf("%s Shutdown in %d seconds\n", timestr(0),pn);
			if(pn>1)
				sleep(pn-1);
			m_mgr->SetContinue(FALSE);
			return TRUE;
		case MADMIN_LOADDLL:
			mferr.mfprintf("%s Loading plugin %s\n", timestr(0),p);
			return globals.session_plugmgr.LoadFunctionMod(p);
		case MADMIN_SETSERIAL:
			mferr.mfprintf("%s Setting serialization parameter:\n\t\"%s\"\n",timestr(0),p);
			if(globals.serialize)
				phmfree(globals.serialize);
			globals.serialize = phmstrdup(p);
			return TRUE;
		case MADMIN_SETVERBOSE:
			mferr.mfprintf("%s Set verbose level: %d\n", timestr(0),pn);
			globals.verbose = pn;
			return TRUE;
		case MADMIN_SETMAXTHREADS:
			mferr.mfprintf("%s Set maximum number of threads: %d\n",timestr(0), pn);
			globals.maxThreads = pn;
			return TRUE;
		case MADMIN_SERIALIZE:
			mferr.mfprintf("%s Serializing %s\n", timestr(0),(p)?p:"");
			if(!p || strlen(p) < 1)
				p = globals.serialize;
			return ExecReqDirect(&m_reqb, MREQ_SERIALIZE, p, "", "", FALSE) == REQ_OK;
		case MADMIN_AUTOSAVE:
			mferr.mfprintf("%s Set autosave: %d\n",timestr(0), pn);
			globals.autosave = pn;
			return TRUE;
		case MADMIN_AUTODROP:
			if(*p == 'Y' || *p == 'y')
				globals.autodrop = TRUE;
			else if (*p == 'N' || *p == 'n')
				globals.autodrop = FALSE;
			mferr.mfprintf("%s Set autodrop: %d\n",timestr(0), globals.autodrop ? "Yes" : "No");
			break;
		case MADMIN_STATUS:
			mferr.mfprintf("%s Status Request\n",timestr(0));
			sprintf((*reqb)->req.datum, 
					"MSession Version %s\n"  
#ifdef ENABLE_MTSVR
					"Thread Pool:         yes\n"
					"Sockets Waiting:     %d\n"
#else
					"Thread Pool:         no\n"
					"Threads Waiting:     %d\n"
#endif
					"Total dialogs:       %d\n"
					"Dialogs per second:  %d\n"
					"Total Sessions:      %d\n"
					"Dropped Connections: %d\n"
					"Total Threads:       %d\n"
					"Max Threads:         %d\n"
					"Active Threads:      %d\n"
					"Autosave:            %d\n"
					"Memory handles:      %d\n"
					"Phoenix Memory:      %d\n"
					"System Memory:       %d\n"
					"System Frag:         %d\n",
					MSESSION_VERSION ,		// Version
					m_mgr->GetThreadStat(),		// Waiting
					m_mgr->GetLastCount(),		// Last count m_last
					m_mgr->GetPPS(),		// m_pps
					globals.session_array->count(),
					globals.dropped,
					m_mgr->count(),
					globals.maxThreads,
					m_mgr->m_ac.Get(),
					globals.autosave,
					phmcount(),
					phmtotal(),
					phmsystotal(),
					phmfrag()
					);
			(*reqb)->req.len = strlen(m_reqb->req.datum)+1;
			break;

		default:
			(*reqb)->req.stat = REQ_ERR;
			(*reqb)->req.param = REQE_DENIED;
			break;
	}
	return (*reqb)->req.stat == REQ_OK;
}

// If here, see if it is a request on which we operate.
Boolean SessionServer::FilterRequest(MSock *ms, REQB ** reqb)
{
	assert(reqb);
	if((*reqb)->req.stat == MREQ_ADMIN)
	{	
		HandleAdmin(ms,reqb);
		WriteRequest(reqb, ms);
		return TRUE;
	}
	else if((*reqb)->req.stat == REQ_POPEN)
	{
		return HandlePersistent(ms, &m_reqb);
	}
	else if((*reqb)->req.stat == REQ_PCLOSE)
	{
		ms->Close();
		return TRUE;
	}
	else if((*reqb)->req.stat == REQ_PING)
	{
		m_mgr->m_count.Inc();
		(*reqb)->req.stat = REQ_OK;
		(*reqb)->req.param = m_mgr->m_count.Get();
		(*reqb)->req.len = 0;
		WriteRequest(reqb, ms);
		return TRUE;
	}
	return FALSE;
}
Boolean SessionServer::HandleSocket(MSock *ms)
{
	AutoCounter ac( &m_mgr->m_ac);
	Boolean retval = TRUE;

	TRACE_FUNCT("HandleSocket");
	if(globals.verbose > 1)
		mferr.mfprintf("%s Handle Socket\n", timestr(0));

	if(ReadRequest(&m_reqb, ms)!= REQ_ERR)
	{
		if(!FilterRequest(ms, &m_reqb))
		{
			MsessionHandleRequest(&m_reqb);
			WriteRequest(&m_reqb, ms);
		}
	}
#ifdef ENABLE_MTSVR
	else
		ms->Close();
#else
#ifdef SHRINK_THREADS
	if(retval)
	{
		if(m_mgr->m_extraThreads.Get() > 0)
		{
			m_mgr->m_extraThreads.Dec();
			retval = FALSE;
		}
	}
#endif
#endif
	return (retval) ? m_mgr->GetContinue() : retval;
}
