/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to commercially license this library, the author
    can be reached at info@mohawksoft.com
*/ 
#include "Phoenix.h"

#define VERBOSE
#define height(n)	((n) ? n->m_meta : 0)
#if 0
#define maxheight(r, n1,n2)\
	{ 	int h1 = height(n1);\
		int h2 = height(n2);\
		r = MAX(h1,h2);	\
	} r

#else
#define maxheight(r, n1, n2)\
	{ r = (n1) ? ((n2) ? MAX(n1->m_meta, n2->m_meta) : n1->m_meta) : height(n2); } r
#endif

		

MAVL::MAVL()
{
	m_last=NULL;
	m_top=NULL;
#ifdef MNODE_HAS_TYPE
	m_type |= MNODE_DATA_MAVL;
#endif
}
MAVL::~MAVL()
{
	// Don't delete nodes, someone else should!
}

void MAVL::clear(void)
{
	m_top=0;
}

MNode *MAVL::Find(NKey key)
{
	int res;
	MNode *node = m_top;

	while(node)
	{
		res = node->cmp(key);
		if(!res) return node;
		node = (res < 0) ?  (MNode *) node->m_prev : (MNode *) node->m_next;	
	}
	return NULL;
}

// Classic rotate functions for AVL 

// Single Rotate Left
MNode * MAVL::SRL( MNode * cn2 )
{
	MNode * cn1 	= cn2->m_prev;
	cn2->m_prev 	= cn1->m_next;	
	cn1->m_next 	= cn2;
	maxheight(cn2->m_meta,cn2->m_prev, cn2->m_next)+=1;
	maxheight(cn1->m_meta,cn1->m_prev, cn2)+=1;
	return cn1; 
}
// Single Rotate Right
MNode * MAVL::SRR( MNode * cn1 )
{
	MNode * cn2 	= cn1->m_next;
	cn1->m_next 	= cn2->m_prev;
	cn2->m_prev 	= cn1;
	maxheight(cn1->m_meta,cn1->m_prev, cn1->m_next)+=1;
	maxheight(cn2->m_meta,cn2->m_next, cn1)+=1;
	return cn2; 
}
// Double Rotate Left
MNode * MAVL::DRL( MNode * cn3 )
{
	cn3->m_prev = (MNode *) SRR(cn3->m_prev);
	return SRL(cn3);
}
// Double Rotate Right
MNode * MAVL::DRR( MNode * cn1 )
{
	cn1->m_next = (MNode *) SRL(cn1->m_next);
	return SRR(cn1);
}
Boolean	MAVL::Insert(MNode *insert)
{
	if(insert)
	{
		m_top = Insert(m_top, insert);
		return ((int)insert->m_meta) != NODE_INSERT_FAILED;
	}
	else
		return FALSE;
}
Boolean	MAVL::InsertStr(MNode *insert)
{
	if(insert)
	{
		m_top = InsertStr(m_top, insert);
		return ((int)insert->m_meta) != NODE_INSERT_FAILED;
	}
	else
		return FALSE;
}
MNode *MAVL::Insert(MNode *cnT, MNode *insert)
{
	if( cnT )
	{
		int cmpresult = cnT->cmp(insert);
		if(cmpresult < 0)
		{
			cnT->m_prev = (MNode *)Insert(cnT->m_prev, insert);
			if(height((cnT->m_prev)) - height((cnT->m_next)) == 2)
				cnT = (cnT->m_prev->cmp(insert) < 0)  ?  SRL( cnT ) : DRL( cnT );
		}
		else if(cmpresult > 0)
		{
			cnT->m_next = Insert(cnT->m_next, insert);
                	if(height((cnT->m_next)) - height((cnT->m_prev)) == 2)
				cnT = (cnT->m_next->cmp(insert) > 0 ) ?  SRR( cnT ) : DRR( cnT );
		}
		else 
			insert->m_meta = (NMeta) NODE_INSERT_FAILED;
	}
	else
                cnT = insert;

	maxheight(cnT->m_meta, cnT->m_prev, cnT->m_next)+=1;
	return cnT;
}
MNode *MAVL::InsertStr(MNode *cnT, MNode *insert)
{
	if( cnT )
	{
		int cmpresult = cnT->scmp(insert);
		if(cmpresult < 0)
		{
			cnT->m_prev = (MNode *)Insert(cnT->m_prev, insert);
			if(height((cnT->m_prev)) - height((cnT->m_next)) == 2)
				cnT = (cnT->m_prev->cmp(insert) < 0)  ?  SRL( cnT ) : DRL( cnT );
		}
		else if(cmpresult > 0)
		{
			cnT->m_next = Insert(cnT->m_next, insert);
                	if(height((cnT->m_next)) - height((cnT->m_prev)) == 2)
				cnT = (cnT->m_next->cmp(insert) > 0 ) ?  SRR( cnT ) : DRR( cnT );
		}
		else 
			insert->m_meta = (NMeta) NODE_INSERT_FAILED;
	}
	else
                cnT = insert;
	
	maxheight(cnT->m_meta, cnT->m_prev, cnT->m_next)+=1;
	return cnT;
}

Boolean MAVL::TreeWalkUp(MNode *mn, EnumFn fn, void *p1, void *p2)
{
	Boolean retval=TRUE;
	if(mn->m_prev)
		retval = TreeWalkUp(mn->m_prev,fn,p1,p2);
	if(retval)
		retval = (*fn)(mn,p1,p2);
	if(retval && mn->m_next)
		retval = TreeWalkUp(mn->m_next,fn,p1,p2);
	return retval;
}

Boolean MAVL::TreeWalkDn(MNode *mn, EnumFn fn, void *p1, void *p2)
{
	Boolean retval=TRUE;
	if(mn->m_next)
		retval = TreeWalkDn(mn->m_next,fn,p1,p2);
	if(retval)
		retval = (*fn)(mn,p1,p2);
	if(retval && mn->m_prev)
		retval = TreeWalkDn(mn->m_prev,fn,p1,p2);
	return retval;
}

Boolean MAVL::defwalker(MNode *mn, void *param1, void *param2)
{
	printf("%s\n", (char *)mn->KeyGet());
	return TRUE;
}
void MAVL::Enum(EnumFn fn, void *p1, void *p2, Boolean ascending)
{
	if(!fn)
		fn = (EnumFn)defwalker;

	if(m_top)
	{
		if(ascending)
			TreeWalkUp(m_top,fn,p1,p2);
		else
			TreeWalkDn(m_top,fn,p1,p2);
	}

	m_last = NULL;
}
