/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to professionally license this library, the author
    can be reached at info@mohawksoft.com
*/ 
/*
	Notes:
	Bitmap Save routines save an integer which specifies the type
	of bitmap it is, however, the load routines operate on the 
	assumption that this value was extracted first, then the new
	routine is called with the file pointer positioned after the
	type specifier.
*/

#include "Phoenix.h"
#include "mbitmap.h"
#include "bmtbl.h"

// #define TRACK_PROC

// Change these for different archetectures.
#define BYTE_BITS 8
#define LONG_BITS 	(BYTE_BITS * sizeof(unsigned long))


#define BMAP_VERSION	'01WM'


#define BMLOGOP( OP )							\
{       register unsigned long *d = (unsigned long *)bm->m_data;	\
       	register unsigned long *s = (unsigned long *)this->m_data;	\
	register int count = MIN(bm->m_clongs, m_clongs);		\
       	while(count--)							\
	{								\
		*d = *d OP *s++;                              		\
		d++;							\
	}								\
}  

int calclongs(int cbits)
{
	int clongs = ((cbits/BYTE_BITS)/sizeof(unsigned long));

	if((clongs * LONG_BITS) < (unsigned int)cbits)
		clongs++;
	return clongs;
}

MBitmap::MBitmap()
{
}

MBitmap *MBitmap::Load(int type, MRandomFile *mf)
{
	MBitmap *mbm;
	switch(type)
	{
		case MBM_TYPE_BIT:
			mbm = new (mf) MBitmapBits();
			break;
		case MBM_TYPE_LIST:
			mbm = new (mf) MBitmapList(mf);
			break;
		default:
			mbm = NULL;
	}
	return mbm;
}
MBitmap::~MBitmap()
{
}
void	MBitmap::setoffset(FOFFSET offset)
{
	m_offset = offset;
}
FOFFSET	MBitmap::getoffset(void)
{
	return m_offset;
}
void	MBitmap::setfsize(int fsize)
{
	m_fsize = fsize;
}
int	MBitmap::getfsize(void)
{
	return m_fsize;
}

void MBitmapBits::AND(MBitmap *bm)
{
	unsigned int ndx;
	unsigned int dndx=0;
	unsigned long *bitmap = (unsigned long *) m_data;
	for(ndx =0; ndx < m_clongs; ndx++)
	{
		unsigned char *p = (unsigned char *)&bitmap[ndx];
		for(unsigned int i=0; i < sizeof(unsigned long); i++)
		{
			for(int j=0; j < 8; j++)
			{
				unsigned char b1 = char_bits[j] & *p;
				unsigned char b2 = bm->get(dndx);
				if(b1 && b2)
					bm->set(dndx);
				else
					bm->reset(dndx);
				dndx++;
			}
			p++;
		}
	}
}

void MBitmapBits::OR(MBitmap *bm)
{
	unsigned int ndx;
	unsigned int dndx=0;
	unsigned long *bitmap = (unsigned long *) m_data;
	for(ndx =0; ndx < m_clongs; ndx++)
	{
		unsigned char *p = (unsigned char *)&bitmap[ndx];
		for(unsigned int i=0; i < sizeof(unsigned long); i++)
		{
			for(int j=0; j < 8; j++)
			{
				unsigned char b1 = char_bits[j] & *p;
				unsigned char b2 = bm->get(dndx);
				if(b1 || b2)
					bm->set(dndx);
				dndx++;
			}
			p++;
		}
	}
}

void MBitmapBits::XOR(MBitmap *bm)
{
	unsigned int ndx;
	unsigned int dndx=0;
	unsigned long *bitmap = (unsigned long *) m_data;
	for(ndx =0; ndx < m_clongs; ndx++)
	{
		unsigned char *p = (unsigned char *)&bitmap[ndx];
		for(unsigned int i=0; i < sizeof(unsigned long); i++)
		{
			for(int j=0; j < 8; j++)
			{
				unsigned char b1 = char_bits[j] & *p;
				unsigned char b2 = bm->get(dndx);
				if(b1 && b2)
					bm->reset(dndx);
				else if(!b1 && !b2)
					bm->reset(dndx);
				else
					bm->set(dndx);
				dndx++;
			}
			p++;
		}
	}
}
void MBitmapBits::NAND(MBitmap *bm)
{
	unsigned int ndx;
	unsigned int dndx=0;
	unsigned long *bitmap = (unsigned long *) m_data;
	for(ndx =0; ndx < m_clongs; ndx++)
	{
		unsigned char *p = (unsigned char *)&bitmap[ndx];
		for(unsigned int i=0; i < sizeof(unsigned long); i++)
		{
			for(int j=0; j < 8; j++)
			{
				unsigned char b1 = char_bits[j] & *p;
				unsigned char b2 = bm->get(dndx);
				if(b1 && b2)
					bm->reset(dndx);
				else
					bm->set(dndx);
				dndx++;
			}
			p++;
		}
	}
}
void MBitmapBits::NOR(MBitmap *bm)
{
	unsigned int ndx;
	unsigned int dndx=0;
	unsigned long *bitmap = (unsigned long *) m_data;
	for(ndx =0; ndx < m_clongs; ndx++)
	{
		unsigned char *p = (unsigned char *)&bitmap[ndx];
		for(unsigned int i=0; i < sizeof(unsigned long); i++)
		{
			for(int j=0; j < 8; j++)
			{
				unsigned char b1 = char_bits[j] & *p;
				unsigned char b2 = bm->get(dndx);
				if(b1 || b2)
					bm->reset(dndx);
				else
					bm->set(dndx);
				dndx++;
			}
			p++;
		}
	}
}

void MBitmapBits::shrink(void)
{
	// No need to do anything.
}
void MBitmapBits::AND(MBitmapBits *bm)
	BMLOGOP( & )

void MBitmapBits::OR(MBitmapBits *bm)
	BMLOGOP( | )

void MBitmapBits::XOR(MBitmapBits *bm)
	BMLOGOP( ^ )

void MBitmapBits::NAND(MBitmapBits *bm)
	BMLOGOP( &~ )

void MBitmapBits::NOR(MBitmapBits *bm)
	BMLOGOP( |~ )

void *MBitmapBits::operator new(size_t size, int cbits, char *name)
{
	void *p=NULL;
	MBitmapBits *mbm;
	char *	pc;
	unsigned long cb;
	unsigned long cbstr;
	unsigned long cbbits ;
	int clongs = calclongs(cbits);

	cbbits = clongs*sizeof(unsigned long);

	cbstr = (name) ? strlen(name)+1 : 0;

	cb = cbbits + cbstr + size;

	p = phmalloc(cb);
	memset(p, 0, cb);

	mbm = (MBitmapBits *) p;
	mbm->m_cbits  = cbits;
	mbm->m_clongs = clongs;

	pc = (char *)p;
	pc += size;

	if(cbstr)
	{
		mbm->m_key = (NKey) pc;
		strcpy(pc, name);
		pc+= cbstr;	
	}
	
	mbm->m_data = (NData)pc;
	mbm->m_offset=-1;
	mbm->m_fsize =0;
	return p;
}

void *MBitmapBits::operator new(size_t size, MRandomFile *mf)
{
	struct _bmdisk bm;
	void *p=NULL;
	MBitmapBits *mbm;
	unsigned long cb;
	unsigned long cbbits;
	FOFFSET	offset;

	char *	pc;

	offset = mf->tell();

	mf->read(&bm, sizeof(bm));

	cbbits = bm.clongs * sizeof(unsigned long);	

	cb = cbbits + bm.cbstr + size;

	p = phmalloc(cb);
	memset(p, 0, cb);
	mbm = (MBitmapBits *) p;

	pc = (char *)p;
	pc += size;

	if(bm.cbstr)
	{
		mbm->m_key = (NKey)pc;
		mf->read(pc, bm.cbstr);
		pc += bm.cbstr;
	}

	int cbcmp;
	int cbdest = cbbits;
	void *pcmp;

	mf->read(&cbcmp, sizeof(cbcmp));

	mbm->m_fsize = cbcmp;

	pcmp = Decompress(mf, &cbdest, pc);

#ifdef ASSERT_DEBUG
	assert((unsigned int)cbdest == cbbits);
	assert(pcmp);
#endif

	mbm->m_cbits  = bm.cbits;
	mbm->m_clongs = bm.clongs;
	mbm->m_data = (NData)pc;

	mbm->m_offset = offset;

	return p;
}

MBitmapBits::MBitmapBits(void)
{
	m_blkmask = ~0;
	m_blkadd = 0;
}
void MBitmapBits::SetInto(MBitmapBits *mbm)
{
	OR(mbm);
}

MBitmapBits::MBitmapBits(MBitmapList *mbml)
{
	mbml->SetInto(this);
}
MBitmapBits::~MBitmapBits()
{
}
Boolean MBitmapBits::Save(MRandomFile *mf, FOFFSET *offset)
{
	struct _bmtypedisk bmt;
	ssize_t cb;

	bmt.bmtype = this->type();
	bmt.bm.cbits = m_cbits;
	bmt.bm.clongs = m_clongs;
	bmt.bm.cbstr = (m_key) ? strlen((char *)m_key)+1 : 0;
	cb = m_clongs * sizeof(unsigned long);

	void *pt;
	int cbbuf = cb + CMPEXTRA;
#ifdef GCC
	char buffer[cbbuf];
#else
	char *buffer = (char *)_alloca(cbbuf);
#endif

	pt = (void *)buffer;

	void *p = Compress(m_data,(int *)&cb, pt);
	if(!p)
	{
		mferr.mfprintf("failure in compressor\n");
		assert(p);
	}
	if(m_blkmask)
		cb = (cb & m_blkmask) + m_blkadd;// Round up a few

	if(m_fsize && (cb <= m_fsize))
	{
		mf->seek(m_offset, SEEK_SET);
	}
	else
	{
		m_fsize = 0;
		m_offset = mf->seek(0, SEEK_END);
	}
	*offset = m_offset;

	if(mf->write((void *)&bmt, sizeof(bmt)) != sizeof(bmt))
	{
		mferr.mfprintf("Can't write bmt %s, %d", __FILE__, __LINE__);
		return FALSE;
	}

	if(bmt.bm.cbstr)
	{
		if(mf->write(m_key, bmt.bm.cbstr) != (int)bmt.bm.cbstr)
		{
			TRACE_ERROR("Can't write bmt.bm.cbstr");
			return FALSE;
		}
	}

	if(mf->write(&cb, sizeof(cb)) != sizeof(cb))
	{
		TRACE_ERROR("Can't write cb");
		return FALSE;
	}

	if(mf->write(p, cb) != cb)
	{
		TRACE_ERROR("Can't write p");
		return FALSE;	
	}

	m_fsize = cb;
	mf->flush();
	return TRUE;
}

void MBitmapBits::NOT(void)
{
	unsigned long *d = (unsigned long *)m_data;
	int count = m_clongs;
	while(count--)
	{
		*d = ~*d;
		d++;
	}
}

int MBitmapBits::count(void)
{
	unsigned int ndx;
	int count =0;
	unsigned long *bitmap = (unsigned long *) m_data;
	for(ndx =0; ndx < m_clongs; ndx++)
	{
		if(bitmap[ndx])
		{
			unsigned char *p = (unsigned char *)&bitmap[ndx];
			for(unsigned int i=0; i < sizeof(unsigned long); i++)
				count += char_bitcounts[*p++];
		}
	}
	return count;
}
void MBitmapBits::setall()
{
	memset(m_data, 0xFF, sizeof(long)*m_clongs);
}
void MBitmapBits::resetall()
{
	memset(m_data, 0, sizeof(long)*m_clongs);
}
Boolean MBitmapBits::get(int bit)
{
	return ( ((unsigned long *)m_data)[OFFSET(bit)] & char_ones[INDEX(bit)]) ? 1 :0;
}

void MBitmapBits::set(int bit)
{
	SETBMBIT(bit);
}
void MBitmapBits::reset(int bit)
{
	((unsigned long *)m_data)[OFFSET(bit)] &= char_zeros[INDEX(bit)];
}
int MBitmapBits::getbits(MTArray <int> &mta, int maxcount)
{
	int count=0;
	unsigned long *bitmap = (unsigned long *) m_data;
	for(unsigned int ndx =0; ndx < m_clongs; ndx++)
	{
		unsigned long chunk = bitmap[ndx];
		if(chunk)
		{
			int position = ndx * LONG_BITS;
			for(unsigned int i=0; i < LONG_BITS; i++)
			{
				if(chunk & 1)
				{
					int bit = position + i;
					mta.Add(bit);
					if(++count >= maxcount)
						goto done;
				}
				chunk = chunk >> 1;
			}
		}
	}
	done:
	return mta.count();
}
int MBitmapBits::getbits(MITuple &mitpl, int maxcount)
{
	int count=0;
	unsigned long *bitmap = (unsigned long *) m_data;
	for(unsigned int ndx =0; ndx < m_clongs; ndx++)
	{
		unsigned long chunk = bitmap[ndx];
		if(chunk)
		{
			int position = ndx * LONG_BITS;
			for(unsigned int i=0; i < LONG_BITS; i++)
			{
				if(chunk & 1)
				{
					ITUPLE tpl;
					tpl.a = position + i;
					tpl.b = 0;

					mitpl.Add(tpl);
					if(++count >= maxcount)
						goto done;
				}
				chunk = chunk >> 1;
			}
		}
	}
	done:
	return mitpl.count();
}
int MBitmapBits::getbits(MITuple *mitpl, int maxcount)
{
	int count=0;
	unsigned long *bitmap = (unsigned long *) m_data;
	for(unsigned int ndx =0; ndx < m_clongs; ndx++)
	{
		unsigned long chunk = bitmap[ndx];
		if(chunk)
		{
			int position = ndx * LONG_BITS;
			for(unsigned int i=0; i < LONG_BITS; i++)
			{
				if(chunk & 1)
				{
					ITUPLE tpl;
					tpl.a = position + i;
					tpl.b = 0;

					mitpl->Add(tpl);
					if(++count >= maxcount)
						goto done;
				}

				chunk = chunk >> 1;
			}
		}
	}
	done:
	return mitpl->count();
}
int MBitmapBits::getbits(int *bits, int cbits)
{
	int count=0;
	unsigned long *bitmap = (unsigned long *) m_data;
	for(unsigned int ndx =0; ndx < m_clongs; ndx++)
	{
		unsigned long chunk = bitmap[ndx];
		if(chunk)
		{
			int position = ndx * LONG_BITS;
			for(unsigned int i=0; i < LONG_BITS; i++)
			{
				if(chunk & 1)
				{
					*bits++ = position + i;
					if(++count >= cbits)
						return count;
				}
				chunk = chunk >> 1;
			}
		}
	}
	return count;
}
MBitmapList::MBitmapList()
{
	m_lastadd = -1;
}
MBitmapList::MBitmapList(MRandomFile *mf) : m_bits(mf)
{
	m_lastadd = -1;
}

void *MBitmapList::operator new (size_t size, char *str)
{
	MBitmapList *mbml;
	char *p;

	int cbstr = strlen(str)+1;
	int cb = size + cbstr;
	p = (char *)phmalloc(cb);
	mbml = (MBitmapList *)p;
#ifdef ASSERT_DEBUG
	assert(p);
#endif
	if(mbml)
	{
		if(str)
		{
			mbml->m_key = &p[size];
			strcpy((char *)mbml->m_key, str);
		}
		else
		{
			mbml->m_key = NULL;
		}
	
		mbml->m_fsize = 0;
		mbml->m_offset = -1;
	}
	return (void *)mbml;	
}
// If here, bmtype has already been read.
void *MBitmapList::operator new (size_t size, MRandomFile *mf)
{
	MBitmapList *mbml;
	char *p;

	FOFFSET offset = mf->tell();
	int cbstr;
	int cb;
	mf->read(&cbstr, sizeof(cbstr));

	cb = size + cbstr;
	p = (char *)phmalloc(cb);
	mbml = (MBitmapList *)p;
#ifdef ASSERT_DEBUG
	assert(p);
#endif
	if(mbml)
	{
		if(cbstr)
		{
			mbml->m_key = &p[size];
			mf->read(mbml->m_key, cbstr);
		}
		else
		{
			mferr.mfprintf("no key in bitmap file\n");
			// Don't read an empty key
			mbml->m_key = NULL;
		}
		mbml->m_fsize = 0;
		mbml->m_offset = offset;
	}
	return (void *)mbml;	
}

Boolean MBitmapList::Save(MRandomFile *mf, FOFFSET *poff)
{
	int cbstr = strlen((char *)m_key)+1;
	int bmtype = this->type();

	// Always make a new entry
	m_offset = mf->seek(0, SEEK_END);

	*poff = m_offset;

	if(mf->write(&bmtype, sizeof(bmtype)) != sizeof(bmtype))
	{
		TRACE_ERROR("cant write bmtype");
		return FALSE;
	}
	if(cbstr > 1)
	{
		mf->write(&cbstr, sizeof(cbstr));
		mf->write(m_key, cbstr);
	}
	else
	{
		// Don't write an empty string!!!
		cbstr = 0;
		mf->write(&cbstr, sizeof(cbstr));
	}
	m_bits.save(mf);
	m_fsize = 0;
	return TRUE;
}

void MBitmapList::SetInto(MBitmapBits *mbmb)
{
	int count = m_bits.count();
	for(int i = 0; i < count; i++)
		mbmb->bmset(m_bits[i]);
}
void MBitmapList::AND(MBitmapBits *bm)
{
	assert(0);	// Can't AND a list!
}
void MBitmapList::OR(MBitmapBits *bm)
{
	int count = m_bits.count();
	for(int i = 0; i < count; i++)
		bm->bmset(m_bits[i]);
}
void MBitmapList::XOR(MBitmapBits *bm)
{
	int count = m_bits.count();
	for(int i = 0; i < count; i++)
	{
		Boolean f = bm->bmget(m_bits[i]);
		if(f)
			bm->bmreset(m_bits[i]);
		else
			bm->bmset(m_bits[i]);
	}
}
void MBitmapList::NAND(MBitmapBits *bm)
{
	int count = m_bits.count();
	for(int i = 0; i < count; i++)
	{
		Boolean f = bm->bmget(m_bits[i]);
		if(f)
			bm->bmreset(m_bits[i]);
	}
}
void MBitmapList::NOR(MBitmapBits *bm)
{
}
void MBitmapList::AND(MBitmap *bm)
{
}
void MBitmapList::OR(MBitmap *bm)
{
}
void MBitmapList::XOR(MBitmap *bm)
{
}
void MBitmapList::NAND(MBitmap *bm)
{
}
void MBitmapList::NOR(MBitmap *bm)
{
}
void MBitmapList::NOT(void)
{
}
Boolean	MBitmapList::get(int bit)
{
	int count = m_bits.count();
	for(int i=0; i < count; i++)
	{
		if(m_bits[i] == bit)
			return TRUE;
	}
	return FALSE;
}
void 	MBitmapList::set(int bit)
{
	if(bit == m_lastadd)
		return; // Been there, done that.
#ifdef CHECK_PREV  // Not correct, but should improve performance.
	if(m_bits.find(bit) == -1)
#endif
		m_bits.Add(bit);
	m_lastadd = bit;
}
void 	MBitmapList::reset(int bit)
{
}
void 	MBitmapList::setall()
{
}
void 	MBitmapList::resetall()
{
}
int 	MBitmapList::count(void)
{
	return m_bits.count();
}

int MBitmapList::getbits(MTArray <int> &mta, int maxcount)
{
	int count = m_bits.count();

	count = MIN(count,maxcount);
	
	for(int i=0; i < count ; i++)
		mta.Add( m_bits[i] );
	return mta.count();
}

int 	MBitmapList::getbits(int *bits, int cbitsMax)
{
	int count = m_bits.count();
	
	for(int i=0; (i < count) && (i < cbitsMax); i++)
		*bits++ = m_bits[i];
	return MIN(cbitsMax, count);
}

void MBitmapList::shrink(void)
{
	m_bits.shrink();
}

// THIS IS EVIL, IF the FILE format for MBitmapBits, MBitmapList
// and Marray<int> change, this MUST BE rewritten!!!!!!
// UGLY UGLY UGLY
Boolean MBitmapBits::SetInto(int type, FOFFSET offset, MRandomFile *mf)
{
	Boolean fresult = FALSE;
	int bmtype;
	mf->seek(offset, SEEK_SET);
	if(mf->read(&bmtype, sizeof(bmtype)) == sizeof(bmtype))
	{
		if(type == bmtype)
		{
			fresult = SetInto(type, mf);
#ifdef TRACK_PROC
			printf("SetInto: %d\n", count());
#endif
		}
	}
	return fresult;
}
Boolean MBitmapBits::SetInto(int type, MRandomFile *mf)
{
	if(type == MBM_TYPE_BIT)
	{
		struct _bmdisk bm;
		int cbcmp;

		int cbdest = m_clongs*sizeof(unsigned long);
		unsigned int cbbits = cbdest;

		mf->read(&bm, sizeof(bm));

		// Incompatible bitmaps!
		if((bm.cbits != m_cbits)||(bm.clongs != m_clongs))
		{
			mferr.mfprintf("Incompatible bitmaps\n");
			return FALSE;
		}

		// Skip the string
		if(bm.cbstr)
			mf->seek(bm.cbstr, SEEK_CUR);

		mf->read(&cbcmp, sizeof(cbcmp));

		if(!Decompress(mf, &cbdest, m_data))
		{
			mferr.mfprintf("Failed to decompress\n");
			return FALSE;
		}
		if((unsigned int) cbdest != cbbits)
		{
			mferr.mfprintf("Failure in decompressor\n");
			return FALSE;
		}

		return TRUE;
	}
	else if(type == MBM_TYPE_LIST)
	{
		struct _madisk ma;

		int chunk[8192];
		int cbstr;

		mf->read(&cbstr, sizeof(cbstr));

		// ignore string
		if(cbstr)
			mf->seek(cbstr, SEEK_CUR);

		// These are from marray
		mf->read(&ma, sizeof(ma));
#ifdef TRACK_PROC
		printf("%d bits in bitmap\n", ma.num);
#endif
#ifndef MAGIC_SUM
#ifdef ASSERT_DEBUG
		assert(ma.magic == MARRAY_MAGIC);
#endif
#else
		int sum=0;
#endif
#ifdef ASSERT_DEBUG
		assert(ma.num <= ma.max);
#endif
		int num = ma.num;

		while(num)
		{
			int to_read = MIN(num, (int)(sizeof(chunk)/sizeof(int)));
			int cb = to_read * sizeof(int);
			if(mf->read((void *)chunk, cb) != cb)
			{
				mferr.mfprintf("Failed to read int section\n");
				return FALSE;
			}

			num -= to_read;

			for(int i=0; i < to_read; i++)
			{
				SETBMBIT(chunk[i]);
#ifdef MAGIC_SUM
				sum += chunk[i];
#endif
			}
		}
#ifdef MAGIC_SUM
#ifdef ASSERT_DEBUG
		assert(ma.magic == sum);
#endif
#endif
		return TRUE;
	}
	mferr.mfprintf("Unknown bitmap type %d\n", type); 
	return FALSE;
}

MBitmapFile::MBitmapFile(unsigned int blkmask, unsigned int blkadd)
{
	m_fCloned = FALSE;
	m_file = NULL;
	m_bits = 0;
	m_fwriteable = FALSE;
	m_blkmask = blkmask;
	m_blkadd = blkadd;
}
MBitmapFile::MBitmapFile(MBitmapFile *mbmf)
{
	m_fCloned	= TRUE;
	m_fwriteable 	= FALSE;
	m_file 		= mbmf->m_file->clone();
	m_bits 		= mbmf->m_bits;
	m_name		= mbmf->m_name;
	m_blkmask = 0;
	m_blkadd = 0;
}
MBitmapFile *MBitmapFile::clone(void)
{
	return new MBitmapFile(this);
}

MBitmapFile::~MBitmapFile()
{
	if(m_file)
		delete m_file;

	if(!m_fCloned && m_name)
		phmfree(m_name);
}

Boolean MBitmapFile::Write(MBitmap *mbm, FOFFSET *loc)
{
	Boolean fresult=FALSE;
	if(m_fwriteable)
	{
		fresult = mbm->Save(m_file, loc);
#ifdef ASSERT_DEBUG
		assert(fresult);
#endif
		if(!fresult)
			mferr.mfprintf("Can't save bitmap: %s, %d", __FILE__, __LINE__);
	}
	else
	{
		mferr.mfprintf("Bitmap file is readonly\n");
	}
	return fresult;
}

MBitmap *MBitmapFile::Read(FOFFSET loc)
{
	int bmtype;

	m_file->seek(loc, SEEK_SET);
	
	if(m_file->read(&bmtype, sizeof(bmtype)) == sizeof(bmtype))
		return MBitmap::Load(bmtype, m_file);
	return 0;
}

Boolean MBitmapFile::SetWorkspace(MBitmapBits *bmwsp, FOFFSET loc)
{
	Boolean fResult=FALSE;
	int bmtype;
	m_file->seek(loc, SEEK_SET);
	bmwsp->resetall();

	if(m_file->read(&bmtype, sizeof(bmtype)) == sizeof(bmtype))
		fResult = bmwsp->SetInto(bmtype, m_file);
	else
		mferr.mfprintf("Failed to read bitmap type\n");

#ifdef TRACK_PROC
	printf("SetWorkspace: %d\n", bmwsp->count());
#endif
	return fResult;
}

MBitmapBits *MBitmapFile::CreateWorkspace(void)
{
	return new (m_bits) MBitmapBits();
}

Boolean MBitmapFile::Open(char *fname)
{
	MMapFile *file = new MMapFile(fname);

	file->seek(0, SEEK_SET);

	HDR *hdr = (HDR *) file->getptr(sizeof(HDR));

	if(hdr)
	{
		if(hdr->magic == (int)MBMF_MAGIX)
		{
			char *str = (char *)file->getptr(hdr->cbstr);

			m_file = file;
			m_bits = hdr->bits;

			if(str)
				m_name = phmstrdup(str);

			file->freeptr(hdr);
			return TRUE;
		}
		file->freeptr(hdr);
	}
	delete file;
	return FALSE;
}

void MBitmapFile::Close(void)
{
	delete m_file;
	m_file = NULL;
	m_bits = 0;
	m_fwriteable = 0;

	if(m_name)
	{
		phmfree(m_name);
		m_name = 0;
	}
}

Boolean MBitmapFile::Create(char *fname, char *name, int bits)
{
	MStdioFile *file = new MStdioFile();

	if(file)
	{
		if(file->open(fname, "w+"))
		{
			HDR hdr;

			hdr.magic	= MBMF_MAGIX;
			hdr.bits	= bits;
			hdr.cbstr	= strlen(name)+1;
			m_name		= phmstrdup(name);
			m_bits		= bits;
			m_file		= file;
			m_fwriteable	= TRUE;

			if(file->write(&hdr, sizeof(hdr))==sizeof(hdr))
				return file->bwrite(m_name, hdr.cbstr);
		}	
	}
	return FALSE;
}
