/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to professionally license this library, the author
    can be reached at info@mohawksoft.com
*/ 
#ifndef __BITMAP__
#define __BITMAP__

/**
A bitmap class that allows binary operations on
arbitrarily large bitmaps 
Logic operations operate on the bitmap passed into the function */

#define CALCBMBSIZE(n)	(sizeof(MBitmapBits) + (calclongs(n)*sizeof(unsigned long)))
#define CALCBMLSIZE(c)	(sizeof(MBitmapList) + (sizeof(unsigned long)*c))	
#define SETBMBIT(bit) 	((unsigned long *)m_data)[OFFSET(bit)] |= char_ones[INDEX(bit)];
#define OFFSET(bit)	((bit) >> 5)
#define INDEX(bit)	((bit) & 0x1f)
extern unsigned char char_bits[];
extern unsigned long char_zeros[];
extern unsigned long char_ones[];


class MBitmapBits;
class MBitmapList;
class MBitmapFile;

/** Calculate the number of longs needed to store cbits */
int calclongs(int cbits);

#define MBM_TYPE_BIT	0
#define MBM_TYPE_LIST	1

struct _bmdisk
{
	unsigned int	cbits;
	unsigned int	clongs;
	unsigned int	cbstr;
};

struct _bmtypedisk
{
	int	bmtype;
	struct 	_bmdisk bm;
};

/** A base class for bitmap indexes */
class MBitmap : public MNode
{
	friend class MBitmapFile;
	protected:
	FOFFSET	m_offset;
	int	m_fsize;

	static MBitmap * Load(int type, MRandomFile *mf);
	virtual Boolean Save(MRandomFile *, FOFFSET *poff)=0;
	virtual void 	SetInto(MBitmapBits *mbmb)=0;

	public:
	MBitmap();
	virtual ~MBitmap();


	virtual void AND(MBitmapBits *bm)=0;
	virtual void OR(MBitmapBits *bm)=0;
	virtual void XOR(MBitmapBits *bm)=0;
	virtual void NAND(MBitmapBits *bm)=0;
	virtual void NOR(MBitmapBits *bm)=0;

	virtual void AND(MBitmap *bm)=0;
	virtual void OR(MBitmap *bm)=0;
	virtual void XOR(MBitmap *bm)=0;
	virtual void NAND(MBitmap *bm)=0;
	virtual void NOR(MBitmap *bm)=0;

	virtual void NOT(void)=0;
	virtual Boolean	get(int bit)=0;
	virtual void 	set(int bit)=0;
	virtual void 	reset(int bit)=0;
	virtual void 	setall()=0;
	virtual void 	resetall()=0;
	virtual int 	count(void)=0;
	virtual int	getbits(MTArray <int> &mta, int maxcount)=0;
	virtual int 	getbits(int *bits, int cbitsMax)=0;
	virtual void	setoffset(FOFFSET offset);
	virtual FOFFSET	getoffset(void);
	virtual void	setfsize(int fsize);
	virtual int	getfsize(void);
	virtual int	type(void)=0;
	virtual	void	shrink(void)=0;
};


/** A true bitmap object */
class MBitmapBits : public  MBitmap
{
	friend class MBitmap;
	friend class MBitmapFile;
	friend class MBitmapList;
	protected:
	unsigned int	m_clongs;
	unsigned int 	m_cbits;
	unsigned int 	m_blkmask;
	unsigned int	m_blkadd;

	virtual Boolean	SetInto(int type, FOFFSET offset, MRandomFile *mf);
	virtual Boolean SetInto(int type, MRandomFile *mf);
	Boolean Save(MRandomFile *, FOFFSET *poff);
	void *operator new(size_t size, MRandomFile *mf);

	inline Boolean bmget(int bit)
	{ return ( ((unsigned long *)m_data)[OFFSET(bit)] & char_ones[INDEX(bit)]) ? 1 :0; }

	inline void bmset(int bit)
	{ SETBMBIT(bit); }

	inline void bmreset(int bit)
	{ ((unsigned long *)m_data)[OFFSET(bit)] &= char_zeros[INDEX(bit)]; }

	public:
	MBitmapBits();
	MBitmapBits(MBitmapList *bmbl);
	~MBitmapBits();

	virtual void AND(MBitmapBits *bm);
	virtual void OR(MBitmapBits *bm);
	virtual void XOR(MBitmapBits *bm);
	virtual void NAND(MBitmapBits *bm);
	virtual void NOR(MBitmapBits *bm);

	virtual void AND(MBitmap *bm);
	virtual void OR(MBitmap *bm);
	virtual void XOR(MBitmap *bm);
	virtual void NAND(MBitmap *bm);
	virtual void NOR(MBitmap *bm);

	virtual void NOT(void);
	virtual Boolean	get(int bit);
	virtual void 	set(int bit);
	virtual void 	reset(int bit);
	virtual void 	setall();
	virtual void 	resetall();
	virtual int 	count(void);
	virtual int 	getbits(int *bits, int cbitsMax);
	virtual int	getbits(MTArray <int> &mta, int maxcount);
	virtual int 	getbits(MITuple &itpl, int maxcount);
	virtual int 	getbits(MITuple *itpl, int maxcount);


	virtual void 	SetInto(MBitmapBits *mbmb);
	void *operator new(size_t size, int cbits, char *name=NULL);
	virtual	void	shrink(void);
	int type(void) { return MBM_TYPE_BIT; }
};

/** A pseudo bitmap which is actually an array of integers */
class MBitmapList : public MBitmap
{
	friend class MBitmap;
	friend class MBitmapFile;
	protected:
	MTArray<int> m_bits;
	int	m_lastadd;

	Boolean Save(MRandomFile *, FOFFSET *poff);
	void *operator new (size_t size, MRandomFile *mf);

	public:
	MBitmapList(void);
	MBitmapList(MRandomFile *mf);
	void *operator new (size_t size, char *name);
	

	virtual void AND(MBitmapBits *bm);
	virtual void OR(MBitmapBits *bm);
	virtual void XOR(MBitmapBits *bm);
	virtual void NAND(MBitmapBits *bm);
	virtual void NOR(MBitmapBits *bm);

	virtual void AND(MBitmap *bm);
	virtual void OR(MBitmap *bm);
	virtual void XOR(MBitmap *bm);
	virtual void NAND(MBitmap *bm);
	virtual void NOR(MBitmap *bm);

	virtual void NOT(void);
	virtual Boolean	get(int bit);
	virtual void 	set(int bit);
	virtual void 	reset(int bit);
	virtual void 	setall();
	virtual void 	resetall();
	virtual int 	count(void);
	virtual int	getbits(MTArray <int> &mta, int maxcount);
	virtual int 	getbits(int *bits, int cbitsMax);
	virtual void 	SetInto(MBitmapBits *mbmb);
	virtual	void	shrink(void);
	int type(void) { return MBM_TYPE_LIST; }
};

/** A file for bitmaps */
class MBitmapFile
{
	#define MBMF_MAGIX	0xfeed1001
	typedef struct _hdr
	{
		int	magic;
		int	bits;
		int	cbstr;
		char	str[0];
		
	}HDR;

	protected:
	Boolean		m_fCloned;
	MRandomFile *	m_file;
	int		m_bits;
	Boolean		m_fwriteable;
	unsigned int	m_blkmask;
	unsigned int	m_blkadd;
	char *		m_name;

	MBitmapFile(MBitmapFile *mbmf);

	public:
	MBitmapFile(unsigned int blkmask = ~0, unsigned int blkadd = 0);
	virtual ~MBitmapFile(void);
	Boolean Write(MBitmap *mbm, FOFFSET *loc);
	MBitmap *Read(FOFFSET loc);
	Boolean SetWorkspace(MBitmapBits *bmwsp, FOFFSET loc);
	MBitmapBits *CreateWorkspace(void);
	Boolean Open(char *filename);
	Boolean	Create(char *filename, char *name, int bits);
	void Close(void);
	int nbits() {return m_bits; }
	char *name() {return m_name; }
	virtual MBitmapFile *clone(void);
	
};

#endif
