/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to commercially license this library, the author
    can be reached at info@mohawksoft.com
*/ 
// #define TRACE_DEBUG

#include "Phoenix.h"

typedef struct _cmp
{
	unsigned int sum;
	int method;
	unsigned int cbpack;
	unsigned int cbunpack;
	unsigned char data[4];
}Packed;

#define CMP_NONE	0x1234
#define CMP_RLE		0x4321

void *Compress(void *m, int *pcb, void *buffer)
{
	int cb = *pcb;
	unsigned int cbn = (sizeof(Packed)+cb)-4;
	unsigned char *p= (unsigned char *)m;
	unsigned char t;
	unsigned char *pt;
	int cbytes;
	Packed *cmp;

	cmp = (Packed *)buffer;

	TRACE_STR("compress");
	TRACE_HEX(*pcb);

	cmp->cbunpack = cb;
	cmp->cbpack = sizeof(Packed) - sizeof(cmp->data);

	pt = cmp->data;
	cmp->sum=0;
	
	for(cbytes=0; cbytes < cb; )
	{
		TRACE_STR("cmp loop");
		// Check for uncompressable
		if(cmp->cbpack >= (unsigned int)cb)
		{
			// Too bad, not worth compressing;
			memcpy(cmp->data, m, cb);
			cmp->sum = 0;
			cmp->method = CMP_NONE;
			cmp->cbpack = cbn;
			*pcb = cmp->cbpack;
			return cmp;
		}
		int count;
		t = *p++; 	// First byte

		TRACE_HEX(t);

		*pt++ = t; 	// Save it;
		cbytes++;	// Take byte out of sum
		cmp->sum += t;	// create sum of bytes;

		for(count = 1; (cbytes < cb) && (t == *p);)
		{
			cbytes ++; // Take byte out of sum
			count ++;  // Bump count
			p++;	   // advanced pointer
		}
		cmp->sum+=count;
		TRACE_HEX(count);
		cmp->cbpack += 2;
		if(count < 256)
		{
			*pt++ = (unsigned char) count;
		}
		else
		{
			*pt++=0;
			*((unsigned int *)pt) = count;
			pt+=sizeof(unsigned int);
			cmp->cbpack += sizeof(unsigned int);	
		}
	}	
	TRACE_STR("done");
	TRACE_HEX(cmp->sum);
	cmp->method = CMP_RLE;
	*pcb = cmp->cbpack;

	return cmp;
}

#define PKSIZE	(sizeof(Packed) - sizeof(((Packed *)0)->data))

void *Decompress(MRandomFile *mf, int *pcb, void *pdest)
{
#ifdef PARANOID
	unsigned int sum=0;
#endif
	void *ret = pdest;
	unsigned char *p =(unsigned char *)ret;
	unsigned char t;
	unsigned int cb=0;
	unsigned int count;
	Packed pk;

	assert(p);

	if(mf->read(&pk, PKSIZE)!= PKSIZE)
	{
		mferr.mfprintf("can't read Packed\n");
		return NULL;
	}
	if(pk.cbunpack > (unsigned int) *pcb)
	{
		mferr.mfprintf("buffer too small\n");
		return NULL;
	}
	if(pk.method == CMP_NONE)
	{
		mf->read(pdest, pk.cbunpack);
		return pdest;
	}

	if(pk.method != CMP_RLE)
	{
		mferr.mfprintf("Bad compression method %d!=%d\n",pk.method, CMP_RLE);
		return NULL;
	} 

	unsigned char *pbm = (unsigned char *) mf->getptr(pk.cbpack);

	void *psave = (void*) pbm;

	assert(pbm);

	for(cb=0; cb < pk.cbunpack;)
	{
		t = *pbm++;
		count = 0;
		count = (unsigned int) *pbm++;
		if(count == 0)
		{
			count = *((unsigned int *)pbm);
			pbm+=sizeof(unsigned int);
		}
		assert(((int)count) >= 0);
#ifdef PARANOID
		sum += t;
		sum += count;
		if(cb+count > pk.cbunpack)
		{
			mferr.mfprintf("incorrect count %d > %d\n", 
						cb+count, pk.cbunpack);
			return NULL;
		}
#endif
		if(count == 1)
			*p++=t;
		else
		{
			memset(p, t, count);
			p+=count;
		}
		cb+= count;	
	}
	mf->freeptr(psave);

#ifdef PARANOID
	if(sum != pk.sum)
	{
		mferr.mfprintf("Error in check sum\n");
		return NULL;
	}
#endif
	*pcb = pk.cbunpack;
	return ret;
}

static void *rlDecompress(Packed *pcmp, int *pcb, void *pdest)
{
	unsigned int sum=0;
	void *ret = pdest;
	unsigned char *p =(unsigned char *)ret;
	unsigned char t;
	unsigned int cb=0;
	unsigned int count;

	assert(p);
	unsigned char *pt = pcmp->data;

	TRACE_STR("Decompress");
	TRACE_HEX(pcmp->cbunpack);

	if(pcmp->cbunpack > (unsigned int) *pcb)
	{
		mferr.mfprintf("buffer too small\n");
		return NULL;
	}

	for(cb=0; cb < pcmp->cbunpack;)
	{
		t = *pt++;
		count = (unsigned int) *pt++;
		TRACE_STR("decomp loop");
		TRACE_HEX(t);
		TRACE_HEX(count);
		
		assert(count < 256);

		sum += t;
		sum+= count;
		if(cb+count > pcmp->cbunpack)
		{
			mferr.mfprintf("incorrect count %d > %d\n", 
						cb+count, pcmp->cbunpack);
			return NULL;
		}
		memset(p, t, count);
		cb+= count;	
		p+=count;
	}
	if(sum != pcmp->sum)
	{
		mferr.mfprintf("Error in check sum\n");
		return NULL;
	}
	*pcb = pcmp->cbunpack;
	return ret;
}
static void *noDecompress(Packed *pcmp, int *pcb, void *pdest)
{
	assert((unsigned int)*pcb >= pcmp->cbunpack);

	if(pdest)
	{
		*pcb = pcmp->cbunpack;
		memcpy(pdest, pcmp->data, pcmp->cbunpack);
	}
	return pdest;
}
void *Decompress(void *pcmp, int *pcb, void *pdest)
{
	switch(((Packed *)pcmp)->method)
	{
		case CMP_RLE:
			return rlDecompress((Packed *)pcmp, pcb, pdest);
			break;
		case CMP_NONE:
			return noDecompress((Packed *)pcmp, pcb, pdest);
			break;
		default:
			return NULL;
	}
}
