/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to commercially license this library, the author
    can be reached at info@mohawksoft.com
*/ 
// #define TRACE_DEBUG

#include "Phoenix.h"

MConfigFile::MConfigFile()
{
	m_filename = NULL;
}

MConfigFile::~MConfigFile()
{
	if(m_filename)
		phmfree(m_filename);
}

char *MConfigFile::GetValue(char *name, char *def)
{
	MNode * mn = Find(name);

	if(!mn)
		return def;
	else
	{
		char *s = (char *) mn->DataGet();
		TRACE_STRSTR(name, s);
		return s;
	}
}

void MConfigFile::CritHandler(char *str)
{
	char szCfgErrFmt[]="The %s field is required in the config file\n";

	mferr.mfprintf(szCfgErrFmt, str);
	exit(0);
}

char *MConfigFile::GetCritValue(char *str)
{
#ifdef VERY_VERBOSE
	printf("Looking for: %s\n", str);
#endif

	char *t = GetValue(str, NULL);

	if(!t)
		CritHandler(str);
#ifdef VERY_VERBOSE
	else
		printf("Found: %s\n", t);
#endif
	return t;
}

int MConfigFile::GetCritIntValue(char *name)
{
	char * value = GetValue(name, NULL);

	if(!value)
	{
		CritHandler(name);
	}
	else
	{
		if(!strcasecmp(value, "TRUE") || !strcasecmp(value, "YES"))
			return 1;
		else if(!strcasecmp(value, "FALSE")|| !strcasecmp(value, "NO"))
			return 0;
		else
			return (value) ? atoi(value) : -1;
	}
	return 0;
}

int MConfigFile::GetIntValue(char *name, int def)
{
	char *value = GetValue(name, NULL);

	if(value)
	{
		if(!strcasecmp(value, "TRUE") || !strcasecmp(value, "YES"))
			def = 1;
		else if(!strcasecmp(value, "FALSE")|| !strcasecmp(value, "NO"))
			def = 0;
		else
			def = atoi(value);
	}

	return def;
}

char *MConfigFile::GetCfgFileName(void)
{
	return m_filename;
}


static void prusage(char *prog)
{
	printf("Usage: %s -f %s.cfg [options]\n", prog, prog);
	printf("   %s.cfg configures this program\n", prog);
	printf("   all values are in \"name=value\" format\n");
	printf("   if %s.cfg is found (with no options), it will be used.\n", prog);
}
MConfigFile::MConfigFile(int argc, char **argv)
{
	char *optarg;
	int optind=1;
	int 	opt;
	char buffer[MAX_STRING];

	m_filename = NULL;
	char *filename=NULL;
	
	while( (opt = GetOption(argc,argv,"f:ho:", &optarg, &optind)) > 0)
	{
		switch(opt)
		{
			case 'f':
				filename = optarg;
				break;
			case 'o':
				strcpy(buffer, optarg);
				split(buffer);
				break;
		}
	}

	if(filename == NULL)
		sprintf(buffer, "%s.cfg", argv[0]);
	else
		strcpy(buffer, filename);

	FILE *f = fopen(buffer, "r");

	if(f)
	{
		fclose(f);
		if(!LoadTextValues(buffer))
		{
			mferr.mfprintf("Error in config file %s\n", buffer);
			exit(0);
		}
		m_filename = phmstrdup(buffer);
	}
	else
	{
		sprintf(buffer, "/etc/%s.conf", argv[0]);

		f = fopen(buffer, "r");
		if(f)
		{
			if(!LoadTextValues(buffer))
			{
				mferr.mfprintf("Error in config file %s\n", buffer);
				exit(0);
			}
			m_filename = phmstrdup(buffer);
		}
	}
}

MFmt::MFmt()
{
}
MFmt::~MFmt()
{
	int count = m_tbl.count();

	for(int i=0; i < count; i++)
	{
		// Do not delete '.fmt' because it will be
		// deleted by the cfg object which owns it.
		if(m_tbl[i].name)
			phmfree(m_tbl[i].name);		
	}
}
MPrinter::MPrinter(char *fname)
{
	m_fConfigured = FALSE;
	m_cfg.LoadTextValues(fname);
}

MPrinter::MPrinter()
{
	m_fConfigured = FALSE;
}

MPrinter::~MPrinter()
{
}

Boolean MPrinter::ReadConfig(char *fname)
{
	return m_cfg.LoadTextValues(fname);
}

// If this returns FALSE, Mutex is locked!!!!!
Boolean MPrinter::IsSet()
{
	if(!m_fConfigured)
	{
		m_mutex.lock();
		if(m_fConfigured) // Could have changed during mutex
			m_mutex.unlock();
	}
	return m_fConfigured;
}

void MPrinter::Close()
{
	m_mutex.unlock();
}

void MFmt::SetDefault(char *szDefault)
{
	m_szDefault = szDefault;
}
char *MFmt::Get(int ndx)
{
	return m_tbl[ndx].fmt;
}

void MFmt::Set(MConfigFile & cfg, int ndx, char *name)
{
	char *p=NULL;
	char *pname = NULL;
	if(name)
	{
		p = cfg.GetValue(name, m_szDefault);
		pname = phmstrdup(name);
	}

	m_tbl[ndx].name = pname;
	m_tbl[ndx].fmt = p;
}
void MFmt::Pump(MFile *f, int ndx, char *value)
{
	char *p = m_tbl[ndx].fmt;
	if(p)
		f->mfprintf(p, value);
}
