/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to commercially license this library, the author
    can be reached at info@mohawksoft.com
*/ 
#include "Phoenix.h"

Boolean MemHeap::save(MFile *mf)
{
	SHDR shdr;
	memset(&shdr,0,sizeof(SHDR));
	shdr.magic = SHDR_MAGIC;
	shdr.cbarray = sizeof(SNDX) *m_offsets.count();
	shdr.cbheap = m_end;
	if(mf->write(&shdr, sizeof(shdr)) != sizeof(shdr))
		return FALSE;
	if(mf->write(m_offsets.pointer(), shdr.cbarray) != (int)shdr.cbarray)
		return FALSE;
	if(mf->write(m_memheap, m_end) != (int)m_end)
		return FALSE;
	return TRUE;
}

Boolean MemHeap::load(MFile *mf)
{
	SHDR shdr;

	if(mf->read(&shdr, sizeof(shdr)) != sizeof(shdr))
		return FALSE;
	if(shdr.magic != SHDR_MAGIC)
		return FALSE;

	int n = shdr.cbarray / sizeof(SNDX);

	if(!m_offsets.setcount(n))
		return FALSE;
	if(!check_size(shdr.cbheap+1))
		return FALSE;
	if(mf->read(m_offsets.pointer(), shdr.cbarray) != (int) shdr.cbarray)
		return FALSE;
	if(mf->read(m_memheap, shdr.cbheap) != (int) shdr.cbheap)
		return FALSE;
	m_end = shdr.cbheap;
	return TRUE;
}
MemHeap::MemHeap(int start)
{
	m_memheap = NULL;
	m_end=0;
	m_size=0;
	m_step=start;
	m_fCloned = FALSE;
}

MemHeap::MemHeap(MemHeap *that)
{
	m_fCloned = TRUE;
	m_end = that->m_end;
	m_size = that->m_size;
	m_step = that->m_step;
}

MemHeap::~MemHeap()
{
	if(!m_fCloned && m_memheap)
	{
		phmfree(m_memheap);
	}
}
unsigned int MemHeap::getprop(int handle)
{
	return m_offsets[ handle ].prop;
}
void MemHeap::setprop(int handle, unsigned int prop)
{
	m_offsets[ handle ].prop = prop;
}
Boolean MemHeap::grow(int cb)
{
	// Calculate new size
	int newsize = m_size += m_step;

	// Double step size, for next alloc.
	m_step+= m_step;

	if(cb >= newsize)
	{
		m_step = (cb/8)*4;
		newsize = m_step * 3;
	}

	assert(cb < newsize);

	// reallocate heap.
	char *newheap = (char *) phmrealloc(m_memheap, newsize);

	if(!newheap)
	{
		// If here, could not reallocate.
		newheap = (char *)phmalloc(newsize);

		// Yikes!!!!
		if(!newheap)
			return FALSE;

		if(m_memheap)
		{
			// Move it on over
			memcpy(newheap, m_memheap, m_size);

			// Free old one.
			phmfree(m_memheap);
		}
	}
	m_memheap = newheap;
	m_size = newsize;
	return TRUE;
}



int MemHeap::adds(char * str)
{
	return addm(1+strlen(str), (void *) str);
}

int MemHeap::adds(char *str, int ndx)
{
	return addm(1+strlen(str), (void *) str, ndx);
}

char *MemHeap::strdup(char *string)
{
	int handle = addm(1+strlen(string), (void *) string);
	return (char *)&m_memheap[ m_offsets [ handle ].offset ];
}
void *MemHeap::malloc(int cb)
{
	int handle = addm (cb, (void *)NULL);
	return (void *)&m_memheap[ m_offsets [ handle ].offset ];
}
void *MemHeap::memdup(int cb, void *mem)
{
	int handle = addm(cb, mem);
	return (char *)&m_memheap[ m_offsets [ handle ].offset ];
}

void MemHeap::clone(MemHeap &mh)
{
	m_fCloned = TRUE;
	m_end 	= mh.m_end;
	m_size 	= mh.m_size;
	m_step 	= mh.m_step;

	if(m_memheap)
		phmfree(m_memheap);

	m_memheap = mh.m_memheap;
	m_offsets.clone( mh.m_offsets);
}

void MemHeap::clear(void)
{
	m_end = 0;
	m_offsets.trunk(0);
}
