/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to commercially license this library, the author
    can be reached at info@mohawksoft.com
*/ 
#include "Phoenix.h"
#include "mfamort.h"

// Calculate payment for amount.
// Amount, amount loaned
// rate, rate per payment
// payments, total number of payments

#define PSEUDO_1	1

#include <math.h>


MFAmort::MFAmort()
{
	m_periods = 1;
	m_pmts = 12;
	m_rate = .009;
	m_amount = 100;
}

// Financially round a number
double MFAmort::round(double number)
{
	double dfloor = floor(number);
	double dcents;
	double dfract;

	dcents = floor((number - dfloor) * 1000)/10;

	dfract = dcents - floor(dcents);

	if(dfract > .5)
		dcents+=1;

	number = dfloor + (dcents/100);

	return number;
}

double MFAmort::Payment()
{
	return m_payment;
}
double MFAmort::Principal()
{
	return m_principal;
}
double MFAmort::TotalInterest()
{
	return m_TotalInterest;
}
double MFAmort::Interest()
{
	return m_Interest;
}
void MFAmort::SetPosition(int pmt)
{
	m_position = pmt;
}
void MFAmort::SetRate(double rate)
{
	m_rate = rate;
}
void MFAmort::SetPmts(double pmts)
{
	m_pmts = pmts;
}
void MFAmort::SetPeriods(double periods)
{
	m_periods = periods;
}
void MFAmort::SetAmount(double amount)
{
	m_amount = amount;
}

// Calculate loan information
// amount * pow(1+ rate, total_payments) = payment * (1 + rate) * ((pow(1+rate,total_payments)-1)/rate);
int MFAmort::update()
{
	double rate 	= m_rate/m_pmts;		// Adjust rate for payments, not periods
	double payments = m_pmts * m_periods;		// Total payments
	double rate1p 	= pow(rate+1, payments);	

	m_payment 	= round((m_amount * rate1p) /  ((rate1p-1)/rate)*(rate+1));	// Calculate payment

	m_TotalInterest = (m_payment * payments) - m_amount;		// Get total interest

	// TODO, figure how to calculate principal and periodic interest,

	m_principal 	= m_amount;					// Calculate principal.

	return 0;
}

#ifdef __TEST_MORT__
main()
{
	int i;
	MFAmort amort;
	double dT=0;
	double dI=0;

	amort.SetRate(0.1);
	amort.SetPeriods(20);
	amort.SetPmts(12);
	amort.SetAmount(12000);

	
	for(i=1; i < 240; i++)
	{
		amort.SetPosition(i);
		amort.update();

		dI = amort.Interest();
		dT += dI;
		printf("Payment %d, %f -> %f, %f\n", i, amort.Payment(), dI, amort.Principal());
	}

	printf("%f == %f\n", dT, amort.TotalInterest());

}
#endif
