/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to commercially license this library, the author
    can be reached at info@mohawksoft.com
*/ 
// #define TRACE_DEBUG
#include "Phoenix.h"

char MFBLK_SIG[]="MBLK";
#define MFBLK_NEW	(MFBLK_MAGIC | MFBLK_USER | MFBLK_USED | MFBLK_LOCKED | MFBLK_DIRTY | MFBLK_INMEM)
#define MFBLK_FREE_FLAGS (MFBLK_LOCKED | MFBLK_USED | MFBLK_DIRTY | MFBLK_INMEM)

////////////// mlist ("in" memory list)  ////////////////////////
// remove block from memory list
void MFBlkfile::mlistremove(PBLK pblk)
{
	TRACE_FUNCT("mlistremove");
	if(pblk)
	{
		if(m_list == pblk)
			m_list = pblk->pnext;
		if(pblk->pprev)
			pblk->pprev->pnext = pblk->pnext;
		if(pblk->pnext)
			pblk->pnext->pprev = pblk->pprev;
	}
}
// add a block to the memory list
void MFBlkfile::mlistadd(PBLK pblk)
{
	TRACE_FUNCT("mlistadd");
	TRACE_ASSERT(pblk);
	pblk->pnext = m_list;
	pblk->pprev=0;
	if(m_list)
		m_list->pprev = pblk;
	m_list = pblk;
}
// free memory associated with objects in freelist.
void MFBlkfile::mlistdelete(PBLK pblk)
{
	TRACE_FUNCT("mlistdelete");
	blkmemfree(pblk);
}
// clear and free all entries in the memory list
void MFBlkfile::mlistclearall(void)
{
	TRACE_FUNCT("clearmlist");
	PBLK pblk = m_list;
	PBLK pblkT = 0;
	m_list = NULL;

	while(pblk)
	{
		pblkT = pblk->pnext;
		pblk->uflags &= ~(MFBLK_LOCKED | MFBLK_INMEM);

		if(pblk->uflags & MFBLK_DIRTY)
			pblkwrite(pblk);

		mlistdelete(pblk);
		pblk = pblkT;
	}
}
////////// memory alloc and free routines ////////////////////
// allocate a block structure
PBLK MFBlkfile::blkmemalloc(size_t size, HBLK handle)
{
	TRACE_FUNCT("blkmemalloc");
	PBLK pblk = (PBLK) phmalloc(size);
	TRACE_ASSERT(pblk);
	if(pblk)
	{
		memset((void *)pblk, 0, size);
		return blkinit(pblk, size, handle);
	}
	return NULL;
}
// free a block structure
void MFBlkfile::blkmemfree(PBLK pblk)
{
	TRACE_FUNCT("blkmemfree");
	TRACE_ASSERT(pblk);
	phmfree((void *)pblk);
}
//////////////////////// free list functions /////////////////////
void MFBlkfile::flistadd(PBLK pblk)
{
	// free blocks are NEVER in memory!
	TRACE_FUNCT("flistadd");
	TRACE_ASSERT(pblk);
	TRACE_ASSERT(pblk->handle);

	// In free list uprop is previous and utype is next HBLK.
	// Just put this on in the front.
	pblk->uprop = 0;
	pblk->utype = m_header.freelist;

	// Get first free list entry, if there is one.
	if(m_header.freelist)
	{
		BLK blkT;
		if(blkreadheader(m_header.freelist, &blkT))
		{
			blkT.uprop = pblk->handle;
			blkwriteheader(&blkT);
		}
	}
	// Make this block the first.
	m_header.freelist = pblk->handle;

	// if it is memory, remove it.
	if(pblk->uflags & MFBLK_INMEM)
	{
		mlistremove(pblk);
		pblk->uflags = pblk->uflags & ~MFBLK_FREE_FLAGS;
		pblkwrite(pblk);
		blkmemfree(pblk);
	}
	else
	{
		// else, just its update header
		pblk->uflags = pblk->uflags & ~MFBLK_FREE_FLAGS;
		blkwriteheader(pblk);
	}
	TRACE_ASSERT(m_header.freelist);
}
// remove a block from the free list
void MFBlkfile::flistremove(PBLK pblk)
{
	TRACE_FUNCT("flistremove");
	TRACE_ASSERT(pblk);
	TRACE_ASSERT(pblk->handle);
	BLK	blkP;
	BLK	blkN;

	if(pblk->uprop) // previous
	{
		if(blkreadheader(pblk->uprop, &blkP))
		{
			blkP.utype = pblk->utype;
			blkwriteheader(&blkP);
		}
	}
	if(pblk->utype) // next
	{
		if(blkreadheader(pblk->utype, &blkN))
		{
			blkN.uprop = pblk->uprop;
			blkwriteheader(&blkN);
		}
	}
	if(m_header.freelist == pblk->handle)
	{
		m_header.freelist = pblk->utype;
	}
}
// find a a sutable block in the file that was freed.
PBLK  MFBlkfile::flistfindfree(size_t size, enum BLK_REUSE method)
{
	TRACE_FUNCT("blkfindfree");
	unsigned int 	cblargest=0;
	unsigned int	cbclosest=0;
	HBLK		hlargest=0;
	HBLK		hclosest=0;
	BLK		blk;
	unsigned int 	cb;
	HBLK 		handle;
	HBLK		hreturn = 0;
	int		count=0;

	for(handle=m_header.freelist; handle;)
	{
		count ++;
		blkreadheader(handle, &blk);
		cb = blkgetsize(&blk);
		if(cb >= size)
		{
			if(method == BLK_FIRSTFIT)
			{
				hreturn = handle;
				break;
			}
			else if(cb > cblargest);
			{
				cblargest = cb;
				hlargest = handle;
			}
			if((cb < cbclosest) || (cbclosest == 0))
			{
				cbclosest = cb;
				hclosest = handle;
			}
		} 
		handle = blk.utype; // next in list;
		TRACE_ASSERT(handle != m_header.freelist);
		if(handle == 0)
			break;
	}
	if(method == BLK_WORSTFIT)
		hreturn = hlargest = hlargest;
	else if (method == BLK_BESTFIT)
		hreturn =  hclosest;

	return (hreturn) ? flistreuse(hreturn) : NULL;
}

// removes used block from free list, returns block ready to use.
PBLK  MFBlkfile::flistreuse(HBLK handle)
{
	TRACE_FUNCT("flistreuse");
	TRACE_ASSERT(handle);
	BLK	blk;
	PBLK 	pblk;
	blkreadheader(handle, &blk);
	flistremove(&blk);
	pblk=blkget(handle);
	if(pblk)
	{
		pblk->uflags = MFBLK_NEW;
		pblk->uprop = 0;
		pblk->utype = 0;
	}	
	return pblk;
}

// initialize a block structure
PBLK MFBlkfile::blkinit(PBLK pblk, size_t size, HBLK handle)
{
	TRACE_FUNCT("blkinit");
	TRACE_ASSERT(pblk);
	memset((void *)pblk, 0, sizeof(BLK));
	pblk->handle	= handle;
	pblk->usize 	= size;
	pblk->blkf 	= this;
	return pblk;
}
// get a block pointer from a data pointer
PBLK  MFBlkfile::getmfblk(void *data)
{
	TRACE_FUNCT("getmfblk");
	PBLK pblk = (PBLK)  NULL;
	unsigned int offset = (unsigned int) &pblk->data[0];
	pblk = (PBLK)  ((char *)data - offset);

	if(blkvalidate(pblk))
		return pblk;
	else
		return NULL;
}
// test if a block is a block
Boolean MFBlkfile::blkvalidate(PBLK pblk)
{
	TRACE_FUNCT("blkvalidate");
	if((pblk->uflags & MFBLK_MAGICMASK) == MFBLK_MAGIC)
		return TRUE;
	else
	{
		TRACE_HEX(pblk->handle);
		TRACE_HEX(pblk->uflags);
		TRACE_EXIT("block fails validation");
		return FALSE;
	}
}
// find a block from its handle
PBLK  MFBlkfile::blkfind(HBLK handle)
{
	TRACE_FUNCT("blkfind");
	PBLK pblkT;

	if(!handle)
		return NULL;

	for(pblkT=m_list; pblkT; pblkT = pblkT->pnext)
	{
		TRACE_ASSERT(pblkT->handle);
		TRACE_ASSERT(pblkT != pblkT->pnext);
		if(pblkT->handle == handle)
		{
			TRACE_HEX(pblkT->handle);
			break;
		}
	}
	return pblkT;
}
// set lock flags, bump count
void MFBlkfile::blksetlock(PBLK pblk)
{
	TRACE_FUNCT("blksetlock");
	TRACE_ASSERT(pblk);
	pblk->lockcount ++;
	pblk->uflags |= (MFBLK_LOCKED | MFBLK_DIRTY);
}
// reduce count, remove lock flag if count == 0
Boolean MFBlkfile::blksetunlock(PBLK pblk)
{
	TRACE_FUNCT("blksetunlock");
	TRACE_ASSERT(pblk);
	if(pblk->lockcount > 1)
	{
		pblk->lockcount --;
		return TRUE;
	}
	else
	{
		pblk->lockcount = 0;
		pblk->uflags = pblk->uflags & ~MFBLK_LOCKED;
	}
	return pblk->uflags & MFBLK_LOCKED;
}
// test if block is dirty
Boolean MFBlkfile::blkisdirty(PBLK pblk)
{
	TRACE_FUNCT("blkisdirty");
	TRACE_ASSERT(pblk);

	return (pblk->uflags & MFBLK_DIRTY) ? TRUE : FALSE;
}
// Reads a header from a file
Boolean MFBlkfile::blkreadheader(HBLK handle, PBLK pblk)
{
	TRACE_FUNCT("blkreadheader");
	TRACE_ASSERT(pblk);
	TRACE_ASSERT(handle);

	UINT32	offset = m_foffset + handle;
	if(m_file->seek(offset, SEEK_SET) == (long)offset)
	{
		blkinit(pblk,0,handle);
		int len = m_file->read((void *)pblk, sizeof(_mfblk_file));
		if(len == sizeof(_mfblk_file) && blkvalidate(pblk))
			return TRUE;
		TRACE_EXIT("error in block header");
	}
	TRACE_EXIT("can't seek to block header");
	return FALSE;
}

// read a block header
Boolean MFBlkfile::blkgetheader(HBLK handle, PBLK pblk)
{
	TRACE_FUNCT("blkgetheader");
	TRACE_ASSERT(pblk);
	TRACE_ASSERT(handle);

	PBLK pblkT = blkfind(handle);
	if(pblkT)
	{
		*pblk = *pblkT;	
		return TRUE;
	}
	else
		return blkreadheader(handle, pblk);
}
// get a block from the file
PBLK MFBlkfile::blkget(HBLK handle)
{
	TRACE_FUNCT("blkget");
	TRACE_ASSERT(handle);

	BLK	blk;
	PBLK pblk = blkfind(handle);

	if(pblk)
	{
		// do I need something here?
	}
	else if(blkgetheader(handle, &blk))
	{
		pblk = blkmemalloc(blk.usize, handle);
		if(pblk)
		{
			memcpy((void *)pblk, &blk, sizeof(_mfblk_file));
			pblk->uflags |= MFBLK_INMEM;
			m_file->read((void *)pblk->data, blkgetsize(&blk));
			mlistadd(pblk);
			TRACE_ASSERT(pblk->handle == handle);
		}
	}
	return pblk;
}
// write the fileheader to the file
Boolean MFBlkfile::fileheaderwrite(_mfblk_header *pblkh, unsigned long offset)
{
	TRACE_FUNCT("fileheaderwrite");
	TRACE_ASSERT(pblkh);
	m_file->seek(offset, SEEK_SET);
	return 	m_file->write((void *)pblkh, sizeof(_mfblk_header))==sizeof(_mfblk_header);
}
// read the file header from the file
Boolean MFBlkfile::fileheaderread(_mfblk_header *pblkh, unsigned long offset)
{
	TRACE_FUNCT("fileheaderread");
	TRACE_ASSERT(pblkh);

	m_file->seek(offset, SEEK_SET);

	if(m_file->read((void *)pblkh, sizeof(_mfblk_header)) == sizeof(_mfblk_header))
	{
		if( *(int *) pblkh->sig == *(int *)MFBLK_SIG)
		{
			return TRUE;
		} 
	}
	TRACE_EXIT("can't read file header");
	return FALSE;
}
// create a new file header
Boolean MFBlkfile::fileheadercreate(_mfblk_header *pblkh, unsigned long offset)
{
	TRACE_FUNCT("fileheadercreate");
	TRACE_ASSERT(pblkh);

	memset(pblkh, 0, sizeof(_mfblk_header));

	*(int *)pblkh->sig = *(int *)MFBLK_SIG;
	pblkh->offset = offset;
	pblkh->blksize = (sizeof(BLK) & 0xFFFF) | ((MAX_MASTLIST) << 16);
	return fileheaderwrite(pblkh, offset);
}
// return usable size of a block
size_t MFBlkfile::blkgetsize(PBLK pblk)
{
	TRACE_FUNCT("blkgetsize");
	TRACE_ASSERT(pblk);
	return pblk->usize - (m_header.blksize & 0xFFFF);
}

int MFBlkfile::getmaxmaster(void)
{
	TRACE_FUNCT("getmaxmaster");
	return (m_header.blksize >> 16);
}
Boolean MFBlkfile::blkwriteheader(PBLK pblk)
{
	TRACE_FUNCT("blkwriteheader");
	TRACE_ASSERT(pblk);
	m_file->seek(m_foffset + pblk->handle, SEEK_SET);
	pblk->uflags &= ~MFBLK_DIRTY;
	return m_file->write((void *)pblk, sizeof(_mfblk_file))==sizeof(_mfblk_file);
}


// allocate a block
PBLK  MFBlkfile::blkalloc(size_t size)
{
	TRACE_FUNCT("blkalloc");
	unsigned long mlen;
	PBLK pblk=NULL;
	
	if(!m_flocked)
	{	
		m_flocked = TRUE;

		size = (size & ~0x7) + 8;		// Adjust for long long
		mlen = sizeof(BLK) + size;		// memory length

		pblk = (PBLK)  blkmemalloc(mlen, 0);

		if(pblk)
		{
			unsigned long lenWr;
			m_file->seek(0, SEEK_END);

			pblk->uflags = MFBLK_NEW;

			pblk->handle = m_file->tell()-m_foffset;
			TRACE_ASSERT(pblk->handle);

			lenWr = m_file->write((void *)pblk,sizeof(_mfblk_file));
			
			if(lenWr != sizeof(_mfblk_file))
			{
				blkmemfree(pblk);
				pblk=NULL;
				return NULL;
			}
			lenWr = m_file->write((void *)pblk->data, size);
			if(lenWr !=size)
			{
				blkmemfree(pblk);
				pblk=NULL;
			}
			blksetlock(pblk);
			mlistadd(pblk);
		} 
		m_flocked = FALSE;
	}
	return pblk;
}
/********************** public routines ******************************************/
// init class
MFBlkfile::MFBlkfile(void)
{
	TRACE_FUNCT("MFBlkfile::MFBlkfile");
	m_foffset = 0;
	m_flocked = 0;
	m_list = NULL;
	memset(&m_header,0, sizeof(m_header));

	
}
// destroy class
MFBlkfile::~MFBlkfile(void)
{
	TRACE_FUNCT("MFBlkfile::~MFBlkfile");
	close();
}

PBLK MFBlkfile::pblkalloc(size_t size, enum BLK_REUSE method)
{
	TRACE_FUNCT("pblkalloc");
	PBLK  pblk=NULL;

	if(method != BLK_NOREUSE)
		pblk = flistfindfree(size, method);

	if(!pblk)	
		pblk = blkalloc(size);

	return pblk;
}
// allocate a block from the file
void *MFBlkfile::alloc(HBLK *handle, size_t size, enum BLK_REUSE method)
{
	TRACE_FUNCT("alloc");
	PBLK  pblk=pblkalloc(size,method);

	if(pblk)
	{
		if(handle)
			*handle = pblk->handle;
		return (void *) pblk->data;
	}

	if(method != BLK_NOREUSE)
		pblk = flistfindfree(size, method);

	if(!pblk)	
		pblk = blkalloc(size);

	if(pblk)
	{
		if(handle)
			*handle = pblk->handle;
		return pblk->data;
	}
	else
		return NULL;
}

void MFBlkfile::blkfree(PBLK pblk)
{
	TRACE_FUNCT("blkfree");
	TRACE_ASSERT(pblk);

	if((pblk->uflags & MFBLK_USED) == 0)
	{
		TRACE_STR("freeing a freeblock!");
		return ; // just ignore it.
	}
	flistadd(pblk);

}
void MFBlkfile::pblkfree(PBLK pblk)
{
	TRACE_FUNCT("pblkfree");

	if(blkvalidate(pblk))
		blkfree(pblk);
}
void MFBlkfile::free(HBLK handle)
{
	TRACE_FUNCT("free");
	PBLK pblk = blkfind(handle);

	if(pblk)
		blkfree(pblk);
	else
	{
		BLK blk;
		blkreadheader(handle, &blk);
		blkfree(&blk);
	}
}

// unlock a block, referenced by handle, write if dirty
void MFBlkfile::pblkunlock(PBLK pblk, Boolean release)
{
	TRACE_FUNCT("pblkunlock");
	if(blkvalidate(pblk))
	{
		if(!blksetunlock(pblk) && release && blkisdirty(pblk))
		{
			pblk->uflags &= ~MFBLK_INMEM;
			pblkwrite(pblk);
			mlistremove(pblk);
			blkmemfree(pblk);
		}
	}
}
void MFBlkfile::unlock(HBLK handle, Boolean release)
{
	TRACE_FUNCT("unlock");
	PBLK pblk = blkfind(handle);
	if(pblk)
		pblkunlock(pblk, release);
}

// lock a block return a pointer, mark it dirty
PBLK MFBlkfile::pblklock(HBLK handle)
{
	TRACE_FUNCT("lock");
	PBLK pblk = blkget(handle);
	if(pblk)
	{
		blksetlock(pblk);
		return pblk;
	}
	return  NULL;
}
void * MFBlkfile::lock(HBLK handle)
{
	TRACE_FUNCT("lock");
	PBLK pblk = pblklock(handle);
	if(pblk)
		return pblk->data;
	else
		return  NULL;
}
// get a block, do not mark it dirty and don't lock it
PBLK MFBlkfile::pblkget(HBLK handle)
{
	TRACE_FUNCT("pblkget");
	return (handle) ? blkget(handle) : NULL;
}
// write a block to file
void MFBlkfile::pblkwrite(PBLK pblk)
{
	TRACE_FUNCT("pblkwrite");
	TRACE_ASSERT(pblk);
	TRACE_ASSERT(pblk->handle);
	blkwriteheader(pblk);
	TRACE_VAL(blkgetsize(pblk));
	m_file->write((void *)pblk->data, blkgetsize(pblk));
}
// get a block, do not mark it dirty and don't lock it
void * MFBlkfile::get(HBLK handle)
{
	TRACE_FUNCT("get");
	PBLK pblk = pblkget(handle);
	if(pblk)
		return pblk->data;

	return  NULL;
}
// get a handle from a pointer
HBLK MFBlkfile::handle(void *data)
{
	TRACE_FUNCT("handle");
	PBLK pblk = getmfblk(data);
	if(pblk)
		return pblk->handle;
	else
		return 0;
}
size_t MFBlkfile::size(HBLK handle)
{
	TRACE_FUNCT("size");
	PBLK pblk = blkfind(handle);
	BLK blk;

	if(pblk)
		return blkgetsize(pblk);
	else if(blkgetheader(handle, &blk))
		return blkgetsize(&blk);
	else
		return 0;
}
void MFBlkfile::markdirty(HBLK handle, Boolean fdirty)
{
	TRACE_FUNCT("markdirty");
	PBLK pblk = blkfind(handle);

	if(pblk)
	{
		if(fdirty)
			pblk->uflags |= MFBLK_DIRTY;
		else
			pblk->uflags &= ~MFBLK_DIRTY;
	}
}

#ifdef WIN32
#define S_ALL	_S_IREAD | _S_IWRITE
#define O_BINARY	_O_BINARY	
#else
#define S_ALL S_IRWXO | S_IRWXG | S_IRWXU
#define O_BINARY	0
#endif 
// open a block file

Boolean MFBlkfile::use(char *fname)
{
	TRACE_FUNCT("open");

	MIoFile *file = new MIoFile();
	m_file = file;

	if(file->open(fname, O_RDWR | O_BINARY, 0)) // it exists
		return fileheaderread(&m_header, m_foffset);
	else if(file->open(fname, O_RDWR | O_CREAT | O_BINARY, S_ALL))
		return fileheadercreate(&m_header, m_foffset);

	return FALSE;
}
#if 1
// save blocks to disk, free unlocked ones.
void MFBlkfile::flush(void)
{
	TRACE_FUNCT("flush");
	PBLK pblk = m_list;
	PBLK pblkT; 

	for(pblk=m_list; pblk; pblk = pblkT)
	{
		pblkT = pblk->pnext;
		if(pblk->uflags & MFBLK_DIRTY)
			pblkwrite(pblk);

		if(!(pblk->uflags & MFBLK_LOCKED))
		{
			mlistremove(pblk);
			mlistdelete(pblk);
		}

		pblk = pblkT;
	}
}

// close a block file
void MFBlkfile::close(void)
{
	TRACE_FUNCT("close");
	mlistclearall();
	fileheaderwrite(&m_header, m_foffset);
	// m_file->close();
}

#else
typedef Boolean (*PBLKENUM)(MFBlkfile *pblkf, PBLK pblk);

Boolean efclose(MFBlkfile *pblkf, PBLK pblk)
{
	if(pblk->uflags & MFBLK_DIRTY)
	{
		pblk->uflags &= ~MFBLK_DIRTY;	// mark clean;
		pblk->flags |= MFBLK_INTERNAL;	// inform of internal action
		pblkf->pblkwrite(pblk);
		pblk->blkmemfree(pblk);
	}
	return TRUE;
}

Boolean efflush(MFBlkfile *pblkf, PBLK pblk)
{
}

void MFBlkfile::pblkenum(PBLKENUM pblkfn)
{
	PBLK pblk, pblkT;

	pblk = m_list;

	while(pblk)
	{
		pblkT = pblk->pnext;
		if(!(*pblkfn)(this, pblk))
			break;
		pblk = pblkT;
	}
}
#endif
// retrive a handle from the master list
HBLK MFBlkfile::getmaster(int ndx)
{
	TRACE_FUNCT("getmaster");
	int max = MIN(getmaxmaster(),MAX_MASTLIST);

	if((ndx >= 0) && (ndx < max))
		return m_header.master[ndx];
	else
		return 0;
}
// save a handle into the master list
Boolean	MFBlkfile::setmaster(int ndx, HBLK handle)
{
	TRACE_FUNCT("setmaster");
	int max = MIN(getmaxmaster(),MAX_MASTLIST);
	if((ndx >= 0) && (ndx < max))
	{
		m_header.master[ndx] = handle;
		return TRUE;
	}
	else
		return 0;
}

int MFBlkfile::numblocks(void)
{
	TRACE_FUNCT("numblocks");
	int i=0;
	PBLK pblk = m_list;
	
	while(pblk)
	{
		i++;
		pblk = pblk->pnext;
	}
	return i;	
}
UINT32 MFBlkfile::gettype(HBLK handle)
{
	TRACE_FUNCT("gettype");
	UINT32 uret=0;

	PBLK pblk = blkfind(handle);
	BLK blk;

	if(pblk)
	{
		uret =	pblk->utype;
	}
	else if(blkreadheader(handle, &blk))	// it ain't in mem
	{
		uret = blk.utype;
	}
	return uret;
}
UINT32 MFBlkfile::settype(HBLK handle, UINT32 type)
{
	TRACE_FUNCT("settype");
	UINT32 uret=0;

	PBLK pblk = blkfind(handle);
	BLK blk;

	if(pblk)
	{
		uret =	pblk->utype;
		pblk->uprop = type;
	}
	else if(blkreadheader(handle, &blk))	// it ain't in mem
	{
		uret = blk.utype;
		blk.uprop = type;
		blkwriteheader(&blk); 
	}
	return uret;
}
UINT32 MFBlkfile::getprop(HBLK handle)
{
	TRACE_FUNCT("getprop");
	UINT32 uret=0;

	PBLK pblk = blkfind(handle);
	BLK blk;

	if(pblk)
	{
		uret =	pblk->uprop;
	}
	else if(blkreadheader(handle, &blk))	// it ain't in mem
	{
		uret = blk.uprop;
	}
	return uret;
}
UINT32 MFBlkfile::setprop(HBLK handle, UINT32 prop)
{
	TRACE_FUNCT("setprop");
	UINT32 uret=0;

	PBLK pblk = blkfind(handle);
	BLK blk;

	if(pblk)
	{
		TRACE_FUNCT("in memory list");
		uret =	pblk->uprop;
		pblk->uprop = prop;
	}
	else if(blkreadheader(handle, &blk))	// it ain't in mem
	{
		TRACE_FUNCT("not in memory list");
		uret = blk.uprop;
		blk.uprop = prop;
		blkwriteheader(&blk); 
	}
	return uret;
}
