/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to commercially license this library, the author
    can be reached at info@mohawksoft.com
*/ 
#include "Phoenix.h"
#include "stdarg.h"

#ifdef WIN32
FILE * winpopen ( char *cmd, char *mode);
void winpclose (FILE *file);
#endif

MStdioFile mferr(stderr);

#define DREAD	1
#define DWRITE	2

#ifdef NEEDS_RW_SYNC
#define SET_DIR( TYPE )	if(m_dir != TYPE ){fseek(m_file,0,SEEK_CUR); m_dir = TYPE;}
#else
#define SET_DIR( TYPE )
#endif

// Core MFile routines 
void *MFile::getptr(int cb)
{
	void *p = phmalloc(cb);
	if(p)
	{
		ssize_t cbT = read(p, cb);
		
		if(cbT != cb)
		{
			phmfree(p);
			p = NULL;
		}
	}
	return p;
}

void *MFile::getptr(FOFFSET off, int cb)
{
	seek(off, SEEK_SET);
	return getptr(cb);
}

void MFile::freeptr(void *p)
{
	phmfree(p);
}

ssize_t MFile::writes(char *str)
{
	ssize_t cb = strlen(str);
	cb =  write((void *)str, cb);
	flush();
	return cb;
}
void MFile::flush(void)
{
	// Place holder!!!
}
Boolean MFile::stat(struct stat *buf)
{
	return FALSE;
}


FOFFSET MFile::seek(FOFFSET offset, int whence)
{
	return (FOFFSET)-1;
}
FOFFSET MFile::tell()
{
	return (FOFFSET)-1;
}

void MFile::rewind()
{
}


ssize_t MFile::mfprintf(char *fmt, ...)
{
	va_list arglist;
	char buffer[4096];
	va_start(arglist, fmt);
	VSNPRINTF(buffer, sizeof(buffer), fmt, arglist);
	va_end(arglist);
	return writes(buffer);
}

MFile::~MFile()
{
}

MRandomFile::MRandomFile()
{
	m_fCloned = FALSE;
}

// MIoFile routines 

MIoFile::MIoFile(FILE *file)
{
	m_fname = NULL;
	m_file = -1;
}

MIoFile::MIoFile(char *fname, int flags, mode_t mode)
{
	m_file = ::OPEN(fname, flags, mode);

	if(m_file != -1)
		m_fname = phmstrdup(fname);
}

MIoFile::~MIoFile()
{
	close();

	if(m_fname)
		phmfree(m_fname);
}

Boolean MIoFile::open(char *filename, int flags, mode_t mode)
{
	m_file = ::OPEN(filename, flags, mode);

	if(m_file != -1)
	{
		m_fname = phmstrdup(filename);
		return TRUE;
	}
	return FALSE;
}
void MIoFile::close(Boolean fremove)
{
	if(m_file != -1)
	{
		::CLOSE(m_file);
		m_file = -1;
	}

	if(m_fname)
	{
		if(fremove)
			unlink(m_fname);

		phmfree(m_fname);
	}
}
ssize_t	MIoFile::read(void *buf, size_t count)
{
	if(m_file != -1)
		return ::READ(m_file, buf, count);
	else
		return 0;
}
ssize_t	MIoFile::write(void *buf, size_t count)
{
	if(m_file != -1)
		return ::WRITE( m_file, buf, count);
	else
		return 0;
}

FOFFSET MIoFile::seek(FOFFSET offset, int whence)
{
	if(m_file != -1)
		return LSEEK(m_file, offset, whence);
	else
		return 0;
}
FOFFSET MIoFile::tell()
{
	if(m_file != -1)
		return LSEEK(m_file, 0, SEEK_CUR);
	else
		return 0;
}
void MIoFile::rewind()
{
	if(m_file != -1)
		LSEEK(m_file, 0, SEEK_SET);
}
int MIoFile::getbyte()
{
	unsigned char c;
	if(m_file != -1)
	{
		if(::READ(m_file, &c, 1) == 1)
			return (int) (c & 0xFF);
	}
	return -1;
}

Boolean MIoFile::eof()
{
	return FALSE; // BUG BUG fix this
}
char *MIoFile::gets(char *buf, int size)
{
	char *s = buf;
	
	while(size)
	{
		int c = getbyte();

		if(c == -1)
			break;
		size--;
		*s++ = c;

		if(c == 0xa || c == 0xd)
			break;
	}
	*s = 0;
	return buf;
}
void MIoFile::flush(void)
{
#ifdef POSIX
	if(m_file != -1)
		fsync(m_file);
#endif
}

MRandomFile *MIoFile::clone(void)
{
	return NULL;
}
Boolean MIoFile::stat(struct stat *buf)
{
	if(buf && m_file != -1)
	{
		if(!fstat(m_file, buf))
			return TRUE;
	}
	return FALSE;
}
void MStdioFile::d()
{
	printf("File Debug:%X\n", (unsigned int) m_file);
}

void MStdioFile::SetFILE(FILE *f, Boolean fCloseOld)
{
	if(fCloseOld)
		close();
	m_file = f;
}
char *MStdioFile::gets(char *buf, int size)
{
	SET_DIR(DREAD)

	if(m_file)
	{
		if(feof(m_file))
			return NULL;
		else	
			return fgets(buf,size,m_file);
	}
	else
		return NULL;
}

Boolean MStdioFile::eof(void)
{
	if(m_file)
		return (feof(m_file));
	else
		return 0;
}
MStdioFile::MStdioFile(FILE *file)
{
	m_file 	= file;
	m_fname = NULL;
	m_exec = FALSE;
}


#define MFBUF_SIZE	65536
MStdioFile::MStdioFile(char *fname, char *how)
{
#ifdef NEEDS_RW_SYNC
	m_dir = -1;
#endif
	m_fname = NULL;
	m_file = NULL;
	m_exec = FALSE;
	if(how)
	{
		m_file = fopen(fname, how);
		if(m_file)
			m_fname = phmstrdup(fname);
	}
	else
	{
		m_file = fopen(fname, "r+");

		if(!m_file)
			m_file = fopen(fname, "w+");

		if(m_file)
			m_fname = phmstrdup(fname);
	}
}

MStdioFile::~MStdioFile()
{
	close();
}

void MStdioFile::close(Boolean fremove)
{
	if(m_file)
	{
		if(m_exec)
			fclose(m_file);
		else
			PCLOSE(m_file);
	}

	if(m_fname)
	{
		if(fremove)
			unlink(m_fname);
		phmfree(m_fname);
	}

	m_file = NULL;
	m_fname = NULL;
}
Boolean MStdioFile::open(char *fname, char *how)
{
	m_exec = FALSE;
	if(how)
	{
		m_file = fopen(fname, how);
		if(m_file)
			m_fname = phmstrdup(fname);
	}
	else
	{
		m_file = fopen(fname, "r+");

		if(!m_file)
			m_file = fopen(fname, "w+");

		if(m_file)
			m_fname = phmstrdup(fname);
	}
	return (m_file != NULL);
}
Boolean MStdioFile::exec(char *command, char *how)
{
	TRACE_FUNCT("MStdioFile::exec");
	m_exec = TRUE;
	if(how)
	{
		m_file = POPEN(command, how);
		if(m_file)
			m_fname = phmstrdup(command);
#ifdef TRACE_DEBUG
		mferr.mfprintf("[%s] command failed\n", command);
#endif
	}
	else
	{
		m_file = POPEN(command, "r");

		if(m_file)
			m_fname = phmstrdup(command);
	}
	return (m_file != NULL);
}

ssize_t MStdioFile::read(void *buf, size_t count)
{
	SET_DIR(DREAD)
	ssize_t ret = 0;	

	if(m_file)
		ret = 	fread(buf, 1, (size_t)count, m_file);
	return ret;
}
ssize_t MStdioFile::write(void *buf, size_t count)
{
	SET_DIR(DWRITE)

	ssize_t ret = 0;	

	if(m_file)
		ret = 	fwrite(buf, 1, (size_t)count, m_file);

	return ret;
}

FOFFSET MStdioFile::seek(FOFFSET offset, int whence)
{
	if(m_file && (fseek(m_file, (long)offset, (long)whence)!=-1))
			return ftell(m_file);

	return (FOFFSET)-1;
}
FOFFSET MStdioFile::tell()
{
	if(m_file)
		return (FOFFSET) ftell(m_file);
	else
		return (FOFFSET)-1;
}

void MStdioFile::rewind()
{
	if(m_file)
		::rewind(m_file);
}

int MStdioFile::getbyte()
{
	SET_DIR(DREAD)
	return fgetc(m_file);
}
void MStdioFile::flush(void)
{
	if(m_file)
		fflush(m_file);
}
MRandomFile *MStdioFile::clone(void)
{
	return NULL;
}

#ifdef WIN32


static FILE *winh2FILE(HANDLE hfile, int mode, char *szmode)
{
	int file = _open_osfhandle((unsigned long)hfile, mode);
	
	if(file == -1)
		return NULL;

	return _fdopen(file, szmode);
}
	
FILE * winpopen ( char *cmd, char *mode)
{
	Boolean fSuccess;
	Boolean fRead = strcmp(mode, "w");
	HANDLE	hRead;
	HANDLE	hWrite;
	HANDLE	hOldRead;
	HANDLE	hOldWrite;
	SECURITY_ATTRIBUTES	security = { sizeof(SECURITY_ATTRIBUTES), NULL, TRUE };
	PROCESS_INFORMATION	procinfo;
	STARTUPINFO startup = { sizeof(STARTUPINFO), 0, NULL, "popen", 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL };

	if (! CreatePipe (&hRead, &hWrite, &security, 0)) 
		return NULL;

	hOldRead = GetStdHandle(STD_INPUT_HANDLE);
	hOldWrite = GetStdHandle(STD_OUTPUT_HANDLE);


	if (fRead)
		SetStdHandle (STD_OUTPUT_HANDLE, hWrite);
	else 
		SetStdHandle (STD_INPUT_HANDLE, hRead);

	fSuccess = CreateProcess (NULL, cmd, NULL, NULL, TRUE, 0, NULL, NULL, &startup, &procinfo);

	if(fRead)
		SetStdHandle(STD_OUTPUT_HANDLE, hOldWrite);
	else
		SetStdHandle(STD_INPUT_HANDLE, hOldRead);

	if(!fSuccess)
		return NULL;


    	if (fRead) 
	{
		CloseHandle(hWrite);
		return winh2FILE(hRead, _O_RDONLY, "r");
	}
	else
	{
		CloseHandle(hWrite);
		return winh2FILE(hWrite, _O_WRONLY, "w");
	}
}
void winpclose (FILE *file)
{
	fclose(file);
}

int pipe ( int *handles)
{
	HANDLE hRead;
	HANDLE hWrite;

	if (! CreatePipe (&hRead, &hWrite, NULL, 0)) 
		return -1;

	*handles++ = (int) _open_osfhandle((unsigned long)hRead, _O_RDONLY);
	*handles   = (int) _open_osfhandle((unsigned long)hWrite, _O_WRONLY);

	return 0;
}
#endif

