/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to commercially license this library, the author
    can be reached at info@mohawksoft.com
*/ 
#include "Phoenix.h"

#ifdef Linux
#include <termio.h>
#include <termios.h>
#include <curses.h>
#include <sys/file.h>
#include <sys/ioctl.h>

Boolean 	MFKbd::s_open=FALSE;
struct	termios MFKbd::s_iosave;

#endif
#ifdef WIN32
#include <io.h>
#include <conio.h>
#endif

MFKbd::MFKbd(void)
{
	TRACE_FUNCT("MFKbd::MFKbd");
	#ifdef Linux
	if(s_open == FALSE)
	{
		tcgetattr(0, &s_iosave);
		m_ionew = s_iosave;

		m_ionew.c_lflag = m_ionew.c_lflag & ~ ICANON;
		m_ionew.c_lflag = m_ionew.c_lflag & ~ ECHO;
		m_ionew.c_lflag = m_ionew.c_lflag & ~ ISIG;
		m_ionew.c_lflag = m_ionew.c_cc[VMIN] = 1;
		m_ionew.c_lflag = m_ionew.c_cc[VTIME] = 0;

		tcsetattr(0, TCSANOW, &m_ionew);
		s_open = TRUE;
		m_waiting = EOF;
	}
	#endif
}


MFKbd::~MFKbd()
{
	TRACE_FUNCT("MFKbd::~MFKbd");
	#ifdef Linux
	tcflush(0, TCIOFLUSH);
	s_iosave.c_lflag = s_iosave.c_lflag | ICANON | ECHO;
	tcsetattr(0, TCSANOW, &s_iosave);
	#endif
}

// is a key waiting?
Boolean MFKbd::iskey()
{
	#ifdef Linux

	int n;
	char ch;
	int result = FALSE;

	TRACE_FUNCT("iskey");
	
	if(m_waiting != EOF)
		result = TRUE;
	else
	{
		m_ionew.c_cc[VMIN]=0;
		tcsetattr(0,TCSANOW,&m_ionew);
		n = read(0,&ch,1);
		m_ionew.c_cc[VMIN]=1; 
		tcsetattr(0,TCSANOW,&m_ionew);

		if(n == 1)
		{
			m_waiting = ch;
			result = TRUE;
		}
	}	
		
	return result;

	#elif defined( WIN32 )
	return _kbhit();
	#else
	return 0;
	#endif
}

int MFKbd::gete(void)
{
	TRACE_FUNCT("gete");
	int c = get();
	put(c);
	return c;
}
int MFKbd::get(void)
{
	TRACE_FUNCT("get");
#ifdef Linux
	char ch;
	if(m_waiting != EOF)
	{
		ch = m_waiting;
		m_waiting = EOF;
		
	}
	else
	{
		read(0,&ch,1);
	}
	return ch;

#elif defined (WIN32)
	return _getch();
#else
	return 0;
#endif
}

void MFKbd::put(int c)
{
	TRACE_FUNCT("put");
	putchar(c);
}

char *MFKbd::getst(char *s, int cbmax)
{
	TRACE_FUNCT("getst");
	char *psave = s;
	while(iskey())	
		*s++ = get();
	*s=0;
	return psave;
}

char *MFKbd::getste(char *s, int cbmax)
{
	TRACE_FUNCT("getste");
	char *psave=s;
	while(iskey())
		*s++ = gete();
	*s = 0;
	return psave;
}
