/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to commercially license this library, the author
    can be reached at info@mohawksoft.com
*/ 
#define TRACE_DEBUG

#include "Phoenix.h"

MFPersist::MFPersist()
{
	TRACE_FUNCT("MFPersist::MFPersist");
}

MFPersist::~MFPersist()
{
	TRACE_FUNCT("MFPersist::~MFPersist");
}

void MFPersist::operator delete(void *pclass)
{
	TRACE_FUNCT("delete");
	PBLK pblk;
	MFPersist *p = (MFPersist *)pclass;
	TRACE_ASSERT(p);	
	pblk = p->m_pblk;
	p->m_blkf->pblkunlock(pblk, TRUE);
}
void * MFPersist::newop(size_t size, MFPersistfile *pblkf, HBLK *phblk)
{
	TRACE_FUNCT("newop");
	TRACE_ASSERT(phblk);
	PBLK pblk;

	if(*phblk)
	{
		TRACE_FUNCT("newop:handle");
		pblk = pblkf->pblklock(*phblk);
		TRACE_ASSERT(pblk);
	}
	else
	{
		TRACE_FUNCT("newop:alloc");
		pblk = pblkf->pblkalloc(size, BLK_BESTFIT);
		TRACE_ASSERT(pblk);
	}
	if(pblk)
	{
		MFPersist *p = (MFPersist *)pblk->data;
		p->m_pblk = pblk;
		p->m_blkf = pblkf;
		p->m_frestored = (*phblk) ? TRUE : FALSE;
		*phblk = pblk->handle;
		return (void *) pblk->data;
	}
	return NULL;
}

void * MFPersist::operator new(size_t size, MFPersistfile *pblkf, HBLK *phblk)
{
	return newop(size,pblkf,phblk);
}
void * MFPersist::operator new(size_t size, MFPersistfile *pblkf, int ndx)
{
	void *	vpret;
	HBLK	hblkT;
	HBLK	hblk = pblkf->getmaster(ndx);

	hblkT = hblk;

	vpret = newop(size,pblkf, &hblk);
	if(!hblkT && vpret)
		pblkf->setmaster(ndx, hblk);
	
	return vpret;
}

Boolean MFPersist::isrestored(void)
{
	TRACE_FUNCT("isresored");
	return m_frestored;
}
UINT32 MFPersist::getprop(void)
{
	TRACE_FUNCT("getprop");
	return m_pblk->uprop;
}
void MFPersist::setprop(UINT32 prop)
{
	TRACE_FUNCT("setprop");
	m_pblk->uprop = prop;
}
UINT32 MFPersist::gettype(void)
{
	TRACE_FUNCT("gettype");
	return m_pblk->utype;
}
void MFPersist::settype(UINT32 type)
{
	TRACE_FUNCT("settype");
	m_pblk->utype = type;
}
void MFPersist::setdirty(Boolean fdirty)
{
	TRACE_FUNCT("setdirty");
	if(fdirty)
		m_pblk->uflags |= MFBLK_DIRTY;
	else
		m_pblk->uflags &= MFBLK_DIRTY; 
}
Boolean MFPersist::isdirty(void)
{
	TRACE_FUNCT("isdirty");
	return m_pblk->uflags & MFBLK_DIRTY;
}
void MFPersist::setmaster(int ndx)
{
	TRACE_FUNCT("setmaster");
	m_blkf->setmaster(ndx,m_pblk->handle);
}
HBLK MFPersist::gethandle(void)
{
	TRACE_FUNCT("sethandle");
	return m_pblk->handle;
}
