/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to commercially license this library, the author
    can be reached at info@mohawksoft.com
*/ 
#include "Phoenix.h"

// Default timing loop.
void MFTimer::TimeLoop(unsigned count)
{
	while(count--)
		;
}

// Fast timer, used to time things much faster than the granularity of
// the clock.
unsigned MFTimer::Calibrate(Boolean fQuiet)
{
	unsigned long		ticks=0;
	unsigned	count=0;
	unsigned	reg;
	unsigned	prevreg;

	count	= 0;
	reg 	= 1;
	prevreg = 0;

	while(1)
	{
		ticks = TickTime(count + reg);

		if(ticks == 0)
		{
			if(reg == 0)
			{
				// Wow! an overflow condition, can't beleive it
				if(count & 0x80000000)
				{
					count = 0xFFFFFFFF;  // Total Wow! fast cpu!
					break;
				}
				else
				{
					reg = 1;
					count = 0x80000000;
				}
			}
		}
		else
		{
			if(reg == 1) // Done!
				break;

			count	= count + prevreg;
			prevreg = 0;
			reg		= 1;

			continue;
		}
		prevreg = reg;
		reg = reg << 1;
	}

	if(!fQuiet) printf("Calibrate at %d cycles per %ld ticks, ", count, ticks);

	if(count != 0xFFFFFFFF)
		count = count * (int) ((ONE_SECOND / ticks));
	else
		if(!fQuiet) printf("[overflow occured] ");


	return count;
}

double MFTimer::CountToMicro(unsigned count)
{
	return (double) 1000000 / (double) count;
}
double MFTimer::MeasureFast(Boolean fQuiet)
{
	unsigned count = Calibrate(fQuiet);
	double result = CountToMicro(count);
	return result;

}
double MFTimer::MeasureSlow(Boolean fQuiet)
{
	unsigned count;
	unsigned adder;
	unsigned long dwt;
	double result;
	unsigned start;
	char szmsg[]="Timing %c\r";


	start = Calibrate(fQuiet);
	adder = start / 16;
	count = start;

	if(!fQuiet) printf(szmsg, '@');

	while(1)
	{
		dwt = TickTime(count);

		if(ONE_SECOND > dwt)
		{
			count += adder;
		}
		else if(dwt > ONE_SECOND)
		{
			count -= adder;
			adder  = adder >> 1;

			if( (adder <= 1) || (count / adder) > 100000)
			   break;
		}

	}

	result = CountToMicro(count);

	if(!fQuiet) printf("Timing: %d cps, %f Microseconds, %f%% of calibration\n", count, result, ((double)start/(double)count) * (double)100);

	return result;
}

//
// This function returns waits for the timer to tick. On a PC this is
// about a .01 seconds (it used to be 1/19 seconds, but I guess NT
// is using real quanta these days, or was that Windows?)
//
unsigned long MFTimer::TickWait()
{
	unsigned long dwStart;
	unsigned long dwt = clock();

	do
		dwStart = clock();
	while(dwt == dwStart);

	return dwStart;
}

//
//	This function sets a high priorty for the thread and then calls
//	TimeLoop with a count. It returns the number of ticks that had
//	elapsed.
//	Note: It waits for a tick to change, then starts. The max error in
//	this routine is less than one tick instead of less than two.
//
unsigned long MFTimer::TickTime(unsigned count)
{
	unsigned long dwt;
	unsigned long dwStart;

	BoostPriority();

	dwStart = TickWait();
	TimeLoop(count);
	dwt = clock();

	DropPriority();

	return dwt - dwStart;
}

