/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to commercially license this library, the author
    can be reached at info@mohawksoft.com
*/ 
#include "Phoenix.h"

MHash::MHash(int tsize)
{
	m_tsize = tsize;
	m_array.setcount(tsize);
#ifdef MNODE_HAS_TYPE
	m_type |= MNODE_DATA_MHASH;
#endif
}
MHash::~MHash()
{
}

// Calculate the hash value
HASH MHash::chash(NKey key)
{
	unsigned char *str = (unsigned char *)key;

	register HASH hash;
	for(hash=0; *str; str++)
	{
		register HASH val =g_tolower_tbl[*str];
		hash = (hash*val)+val; 
	}
	return (int)( hash % m_tsize);
}

/* prev/next used for list, use meta for linking */
MNode *MHash::Find(HASH hash, NKey key)
{
	MNode * node = m_array[hash];

	if(node)
	{
		while(node)
		{
			if(node->cmp(key)==0)
				return node;
			else
				node = (MNode *)node->MetaGet();
		}
	}
	return NULL;
}
/* prev/next used for list, use meta for linking */
MNode *MHash::FindStr(HASH hash, char *key)
{
	MNode * node = m_array[hash];

	if(node)
	{
		while(node)
		{
			if(node->scmp(key)==0)
				return node;
			else
				node = (MNode *)node->MetaGet();
		}
	}
	return NULL;
}

Boolean MHash::Insert(HASH hash, MNode *node)
{
	MNode	*nodeT = m_array[hash];

	m_list.push(node);

	// Make sure there is a link to any previous
	// node.
	node->MetaSet((NMeta)nodeT);

	m_array[hash]=node;

	return TRUE;
}
MNode *MHash::Find(NKey key) 
{ 
	return Find(chash(key),key); 
}
Boolean MHash::Insert(MNode *node)
{
	return Insert(chash(node->KeyGet()), node);
}
Boolean MHash::Remove(MNode *node)
{
	return Remove(chash(node->KeyGet()), node);
}

// Remove a node
Boolean MHash::Remove(HASH hash, MNode *node)
{
	MNode	*nodeP = m_array[hash];

	m_list.pull(node);

	if(nodeP == node)
	{
		m_array[hash] = (MNode *)node->MetaGet();
	}
	else	// We must find correct node
	{
		MNode *nodeT = (MNode *)nodeP->MetaGet();
		while((nodeT != node) && (nodeT))
		{
			nodeP = nodeT;
			nodeT = (MNode *)nodeP->MetaGet();
		}

		if(nodeT == node)
			nodeP->MetaSet(nodeT->MetaGet());
	}
	return TRUE;
}

void MHash::Rehash(int tsize)
{
	m_tsize = tsize;
	m_array.clear();
	m_array.setcount(tsize);

	MNode *node = m_list.FirstGet();

	while(node)
	{
		HASH hash = chash(node->KeyGet());
		node->MetaSet((NMeta)m_array[hash]);
		m_array[hash] = node;
		node = node->NextGet();
	}
}
void MHash::Enum(EnumFn fn, void *param1, void *param2, Boolean ascending)
{
	m_list.Enum(fn,param1,param2,ascending);
}
