/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to commercially license this library, the author
    can be reached at info@mohawksoft.com
*/ 
#include "Phoenix.h"

int nncmp(const void *p1, const void *p2)
{
	register MNode *mn1 = ((MNode **)p1)[0];
	register MNode *mn2 = ((MNode **)p2)[0];
	register int result = mn2->CountGet() - mn1->CountGet();
	return (result) ? result : mn2->cmp(mn1);
}
MNode **MHistogramBase::calc(int *ce)
{
	MTSort <MNode *>sort;
	if(ce)
		*ce = m_entries;
	sort.shell(m_table,m_entries, nncmp); 
	return m_table;
}

Boolean MHistogramBase::tableGrow(int n)
{
	MNode * *phist;	
	unsigned int cb;
	int newmax;

	if(m_maxent > n)
		return TRUE;

	newmax = m_maxent * 2;

	if(newmax < n)
		newmax = n+1;

	cb = newmax * sizeof(MNode *);

	// will do initial malloc if m_table==NULL
	phist = (MNode **)phmrealloc((void *)m_table, cb);

	if(phist)
	{
		m_table = (MNode **)phist;
		m_maxent = newmax;
		return TRUE;
	}
	return FALSE;
}

MNode *MHistogramBase::indexAdd(MNode *mn)
{
	if(!tableGrow(m_entries))
		return NULL;

	m_index.Insert(mn);
	mn->CountSet(1);
	m_table[m_entries]= mn;
	m_entries++;
	return mn;
}
MNode *MHistogramBase::indexFind(NKey key)
{
	return m_index.Find(key);
}

MHistogramBase::MHistogramBase(int init)
{
	m_table = NULL;
	m_maxent=0;
	m_entries=0;
	tableGrow(init);	
}
MHistogramBase::~MHistogramBase()
{
	for(int i=0; i < m_entries; i++)
		delete m_table[i];

	m_entries=0;

	m_index.clear();

	if(m_table)
		phmfree(m_table);
}
void MHistogramBase::clear(void)
{
	for(int i=0; i < m_entries; i++)
		delete m_table[i];

	m_entries=0;
	m_index.clear();
}
