/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to commercially license this library, the author
    can be reached at info@mohawksoft.com
*/ 
#include "Phoenix.h"
//
// List goes like this:
//			    first          last
//		  ---------m_First<-MList->m_Last -----------------------
//                |                                                     |
//                v                                                     v
//	m_prev<-MNode->m_next <--> m_prev<-MNode->m_next <--> m_prev<-MNode->m_next 
//
MList::MList()
{
	m_count = 0;
	m_First = NULL;
	m_Last = NULL;
	m_prev = NULL;
	m_next = NULL;
#ifdef MNODE_HAS_TYPE
	m_type |= MNODE_DATA_MLIST;
#endif
}

MList::~MList()
{
}
void MList::DeleteAll(EnumFn fn, void *p1, void *p2)
{
	MNode *mn = m_First;

	while(mn)
	{
		MNode *mnT = mn->NextGet();
		if(fn)
			(*fn) (mn,p1, p2);
		delete mn;
		mn = mnT;
	}	
}

int MList::count(void)
{
	return m_count;
}
void MList::FirstAdd(MNode *mn)
{
	MNode *mnT = m_First;
	mn->NextSet(m_First);
	mn->PrevSet((MNode *)NULL);
	m_First = mn;
	if(mnT)
		mnT->PrevSet(m_First);
	if(m_Last == NULL)
		m_Last = m_First;
	m_count++;	
}
void MList::LastAdd(MNode *mn)
{
	MNode *mnT = m_Last;
	mn->PrevSet(m_Last);
	mn->NextSet((MNode *)NULL);
	m_Last = mn;

	if(mnT)
		mnT->NextSet(m_Last);

	if(m_First == NULL)
		m_First = m_Last;
	m_count++;
}

MNode *MList::FirstGet(Boolean fRemove)
{
	MNode *mnR = m_First;
	if(mnR && fRemove)
	{
		m_First = mnR->NextGet();
		if(mnR == m_Last)
		{
			TRACE_ASSERT(m_First == NULL);
			m_Last = m_First;
		}
		m_count--;
	}
	return mnR;
}

MNode *MList::LastGet(Boolean fRemove)
{
	MNode *mnR = m_Last;

	if(mnR && fRemove)
	{
		m_Last = mnR->PrevGet();
		if(mnR == m_First)
		{	// Removing one entry!
			TRACE_ASSERT(m_Last == NULL);
			m_First = m_Last;
		}
		m_count--;
	}
	return mnR;
}

void MList::pull(MNode *mn)
{
	MNode *mnPrev;
	MNode *mnNext;

	if(mn)
	{
		mnPrev = mn->PrevGet();
		mnNext = mn->NextGet();

		if(mnNext)
			mnNext->PrevSet(mnPrev);
		if(mnPrev)
			mnPrev->NextSet(mnNext);

		mn->PrevSet(NULL);
		mn->NextSet(NULL);

		m_count--;
		
		if(m_First == mn)
		{
			m_First = mnNext;

			if(m_First)
				m_First->PrevSet(NULL);
			
		}
		if(m_Last == mn)
		{
			m_Last = mnPrev;
			if(m_Last)
				m_Last->NextSet(NULL);
		}

		if(!m_count)
			m_First = m_Last = 0;
	}
}

Boolean MList::Insert(MNode *mn)
{
	LastAdd(mn);
	return TRUE;
}

MNode *MList::Find(NKey key)
{
	MNode *mn = FirstGet();

	while(mn)
	{
		if(!mn->cmp(key))
			return mn;
		mn = mn->NextGet();
	}
	return NULL;
}
void MList::Enum(EnumFn fn, void *p1, void *p2, Boolean ascending)
{
	MNode *nodeFirst = FirstGet();
	MNode *node = nodeFirst;

	if(!fn)
		return;

	while(node)
	{
		if(!(*fn)(node,p1,p2))
			break;
		node = node->NextGet();
		if(node == nodeFirst)
			break;
	}
}
