/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to commercially license this library, the author
    can be reached at info@mohawksoft.com
*/ 
// #define TRACE_DEBUG
#include "Phoenix.h"
#include "stdarg.h"

MRandomFile *MMemoryFile::clone(void)
{
	TRACE_FUNCT("clone");
	return new MMemoryFile(this);
}

MMemoryFile::MMemoryFile(unsigned  growsize)
{
	TRACE_FUNCT("MMemoryFile::MMemoryFile");
	m_offset 	= 0;
	m_bufsize 	= 0;
	m_end 		= 0;
	m_buffer 	= NULL;
	grow(growsize);
}
MMemoryFile::MMemoryFile(MMemoryFile *mmf)
{
	TRACE_FUNCT("MMemoryFile::MMemoryFile");
	m_fCloned = TRUE;
	m_bufsize = mmf->m_bufsize;
	m_buffer  = mmf->m_buffer;
	m_end	  = mmf->m_end;
}
MMemoryFile::~MMemoryFile()
{
	TRACE_FUNCT("MMemoryFile::~MMemoryFile");

	if(m_fCloned == FALSE)
	{
		AutoMutex amtx(&m_mutex);

		if(m_buffer)
			phmfree(m_buffer);
	}

}
ssize_t MMemoryFile::read(void *buf, size_t count)
{
	TRACE_FUNCT("read");
	AutoMutex amtx(&m_mutex);
	size_t cb = 0;
	if(m_end > m_offset)
	{
		cb = m_end - m_offset;
		if(cb > count)
			cb = count;
		memcpy(buf, &m_buffer[m_offset], cb);
		m_offset += cb;
	}
	return cb;
}

ssize_t MMemoryFile::write(void *buf, size_t count)
{
	TRACE_FUNCT("write");
	AutoMutex amtx(&m_mutex);
	unsigned cb = m_offset + count;

	if(grow(cb))
	{
		memcpy(&m_buffer[m_offset], buf, count);
		m_offset += count;

		if(m_offset > m_end)
		{
			m_end = m_offset;
			m_buffer[m_offset]=0;
		}

		return count;
	}
	return 0;
}

FOFFSET MMemoryFile::seek(FOFFSET offset, int whence)
{
	TRACE_FUNCT("seek");
	AutoMutex amtx(&m_mutex);
	FOFFSET noffset;

	switch(whence)
	{
		case SEEK_SET:
			noffset = offset;
			break;
		case SEEK_CUR:
			noffset = m_offset+offset;
			break;
		case SEEK_END:
			noffset = m_bufsize+offset;
			break;
		default:
			return -1;
			break;
	}
	if(grow(noffset))
	{
		m_offset = noffset;
		return m_offset;
	}
	return -1;
}

FOFFSET MMemoryFile::tell()
{
	TRACE_FUNCT("tell");
	AutoMutex amtx(&m_mutex);
	return (FOFFSET) m_offset;
}

void MMemoryFile::rewind()
{
	TRACE_FUNCT("rewind");
	AutoMutex amtx(&m_mutex);
	m_offset = 0;
}

void MMemoryFile::trunc(void)
{
	TRACE_FUNCT("trunc");
	AutoMutex amtx(&m_mutex);
	m_end = m_offset;
	m_buffer[m_offset]=0;
}

int MMemoryFile::getbyte()
{
	TRACE_FUNCT("getbyte");
	AutoMutex amtx(&m_mutex);
	return (m_offset < m_bufsize) ? (int) (m_buffer[m_offset++]&0xFF) : -1;
}

unsigned MMemoryFile::grow(unsigned size)
{
	TRACE_FUNCT("grow");
	if(!m_buffer)
	{
		m_growsize = ((size/MEM_ALLOC)*MEM_ALLOC)+MEM_ALLOC;
		m_buffer = (char *)phmalloc(m_growsize);
		return (m_buffer) ? m_growsize : 0;
	}
	else if(size > m_bufsize)
	{
		char *p;
		unsigned newsize = ((size/m_growsize)*m_growsize)+m_growsize;
		
		p = (char *)phmrealloc(m_buffer, newsize);

		if(p)
		{
			m_buffer = p;
			m_bufsize = newsize;
			return m_bufsize;
		}
		else
		{
			p = (char *)phmalloc(newsize);

			if(p)
			{
				char *t = m_buffer;
				memcpy(p, m_buffer, m_bufsize);
				m_buffer = p;
				phmfree(t);
				m_bufsize = newsize;
				return m_bufsize;
			}
		}
	}
	else
		return m_bufsize;

	return 0;
}


void *MMemoryFile::getptr(int cb)
{
	TRACE_FUNCT("getptr");
	AutoMutex amtx(&m_mutex);
	void *p = (void *)&m_buffer[m_offset];
	if((m_end - m_offset) >= (unsigned int) cb)
	{
		m_offset += cb;
	}
	return p;
}

void *MMemoryFile::getptr(FOFFSET off, int cb)
{
	TRACE_FUNCT("getptr");
	AutoMutex amtx(&m_mutex);
	
	if(grow(off))
		m_offset = off;
	else
		return NULL;

	void *p = (void *)&m_buffer[m_offset];

	if((m_end - m_offset) >= (unsigned int) cb)
		m_offset += cb;

	return p;

}
void MMemoryFile::freeptr(void *)
{
	TRACE_FUNCT("freeptr");
	// No need to do anything!
}

char MMemoryFile::operator [] (unsigned offset)
{
	TRACE_FUNCT("operator []");
	AutoMutex amtx(&m_mutex);
	return (offset < m_bufsize) ? m_buffer[offset] : 0;
}

Boolean MMemoryFile::eof(void)
{
	TRACE_FUNCT("eof");
	AutoMutex amtx(&m_mutex);
	return (m_offset >= m_end);
}

ssize_t MMemoryFile::dump(MFile *mmf)
{
	TRACE_FUNCT("dump");
	AutoMutex amtx(&m_mutex);
	ssize_t cb = m_end;
	if(m_end)
	{
		ssize_t cb = m_end;
		cb = mmf->write((void *)m_buffer, cb);
		m_end = 0;
		m_offset = 0;
		m_buffer[m_offset]=0;
	}
	return cb;
}
char *MMemoryFile::gets(char *string, int size)
{
	TRACE_FUNCT("gets");
	AutoMutex amtx(&m_mutex);
	char *s = string;
	
	if(m_offset < m_end)
	{
		while(--size)
		{
			char c = m_buffer[m_offset++];

			*s++ = c;

			if((c == 0xa || c == 0xd) || (m_offset >= m_end))
			{
				*string=0;
				return string;
			}
		}
	}
	*s = 0;
	return string;
}


#include <sys/types.h>
#ifdef POSIX
#include <sys/mman.h>
#include <unistd.h>
#endif
#include <fcntl.h>

unsigned MMapFile::grow(unsigned cb)
{
	TRACE_FUNCT("grow");
	if(cb <= m_end)
		return TRUE;
	else
		return FALSE;
}

MMapFile::~MMapFile()
{
	TRACE_FUNCT("MMapFile::~MMapFile");
	if(m_fCloned == FALSE)
	{
#ifdef POSIX
		if(m_buffer)
		{
			munmap(m_buffer, m_bufsize);
			m_buffer = NULL;
		}

		if(m_fd != -1)
		{
			close(m_fd);
		}
#endif
#ifdef WIN32
		if(m_hmap)
		{
			CloseHandle(m_hmap);
			m_buffer = NULL;
		}

		if(m_hfile)
		{
			CloseHandle(m_hfile);
		}
#endif
	}
}

MMapFile::MMapFile(char *fname)
{
	TRACE_FUNCT("MMapFile::MMapFile");
#ifdef POSIX
	FOFFSET off;
	int fd = open(fname, O_RDONLY);

	if(fd != -1)
	{
		m_fd = fd;
		off = lseek(fd, 0, SEEK_END);
		lseek(fd,0,SEEK_SET);
		m_bufsize = off;
		m_end = off;
		m_buffer = (char *)mmap(NULL, off, PROT_READ, MAP_PRIVATE, fd, 0);
	}
	else
	{
		m_fd = -1;
		m_buffer = NULL;
		m_end = 0;
	}
#endif

#ifdef WIN32
	HANDLE hfile = CreateFile(fname, GENERIC_READ, FILE_SHARE_READ,
			NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

	if(hfile != INVALID_HANDLE_VALUE)
	{
		SECURITY_ATTRIBUTES	sa;
		sa.nLength = sizeof(sa);
		sa.lpSecurityDescriptor = NULL;
		sa.bInheritHandle = TRUE;
		m_hfile = hfile;

		m_hmap = CreateFileMapping(m_hfile, &sa, PAGE_READONLY, 0,0,NULL); 
		if(m_hmap)
		{
			m_buffer = (char *) MapViewOfFile(m_hmap, FILE_MAP_READ, 0,0,0);
			m_bufsize = GetFileSize(hfile, NULL);
			m_end = m_bufsize;
		}
	}
	else
	{
		m_hfile = NULL;
		m_hmap = NULL;
	}
#endif
}

MMapFile::MMapFile(MMapFile *mmf)
{
	TRACE_FUNCT("MMapFile::MMapFile");
#ifdef POSIX
	m_fd = -1;
#endif
#ifdef WIN32
	m_hfile = NULL;
	m_hmap = NULL;
#endif
	m_fCloned = TRUE;
	m_end = mmf->m_end;
	m_buffer = mmf->m_buffer;
	m_bufsize = mmf->m_bufsize;
}

MRandomFile *MMapFile::clone(void)
{
	TRACE_FUNCT("clone");
	return new MMapFile(this);
}
