/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to commercially license this library, the author
    can be reached at info@mohawksoft.com
*/ 
#include "Phoenix.h"

MNode::MNode(NKey key, NData data)
{
	if(key)
		m_key = key;
	if(data)
		m_data = data;
	m_next = NULL;
	m_prev = NULL;
	m_count=0;
	m_meta=0;
}
MNode::~MNode()
{
}
#ifdef MNODE_HAS_TYPE
NType 	MNode::TypeGet()
{
	return m_type;
}

void	MNode::TypeSet(NType type)
{
	m_type = type;
}
#endif
void MNode::DataSet(NData data)
{
	m_data = data;
}
NData MNode::DataGet()
{
	return m_data;
}
void MNode::KeySet(NKey key)
{
	m_key = key;
}
NKey MNode::KeyGet()
{
	return m_key;
}
void MNode::NextSet(MNode *next)
{
	m_next = next;
}
MNode *MNode::NextGet()
{
	return m_next;
}
void MNode::PrevSet(MNode *prev)
{
	m_prev = prev;
}
MNode *MNode::PrevGet()
{
	return m_prev;
}

NMeta MNode::MetaGet(void)
{
	return m_meta;
}
void MNode::MetaSet(NMeta m)
{
	m_meta = m;
}
unsigned int MNode::PropGet(void)
{
	return m_prop;
}
void MNode::PropSet(unsigned int prop)
{
	m_prop = prop;
}
int MNode::cmp(NKey key)
{
	int c  = ((char *)key)[0] - ((char *)m_key)[0];
	return (c) ? c : strcmp((char *)key, (char *)m_key);
}

int MIntNode::cmp(NKey key)
{
	return ((int)key) - ((int)m_key);
}


MNode *MNode::Max()
{
	return (m_next) ? m_next->Max() : this;
}
MNode *MNode::Min()
{
	return (m_prev) ? m_prev->Min() : this;
}

void * MNode::operator new (size_t size)
{
	MNode *mn = (MNode *)phmalloc(size);
	if(mn)
	{
		mn->m_key = NULL;
		mn->m_data = NULL;
#ifdef MNODE_HAS_TYPE
		mn->m_type = MNODE_KEY_UNKNOWN;
#endif
	}
	return (void *) mn;
}
void * MNode::operator new (size_t size, void *mem)
{
	register MNode *mn = (MNode *)mem;
#ifdef MNODE_HAS_TYPE
	mn->m_type = MNODE_KEY_UNKNOWN;
#endif
	mn->m_key=0;
	mn->m_data = 0;
	return mem;
}
void * MNode::operator new (size_t size, void *key, int cbkey, void * data, int cbdata)
{
	if(key && (cbkey == 0))
		cbkey = strlen((char *)key)+1;

	if(data && (cbdata == 0))
		cbdata = strlen((char *)data)+1;

	char *ptr = (char *)phmalloc(size + cbkey + cbdata);

	if(ptr)
	{
		if(cbkey)
		{
			((MNode *)ptr)->m_key = &ptr[size];
			if(key)
				memcpy(((MNode *)ptr)->m_key, key, cbkey);
#ifdef MNODE_HAS_TYPE
				((MNode *)ptr)->m_type = MNODE_KEY_STRING;
#endif
		}
		else
		{
			((MNode *)ptr)->m_key = NULL;
#ifdef MNODE_HAS_TYPE
			((MNode *)ptr)->m_type = MNODE_KEY_UNKNOWN;
#endif
		}

		if(cbdata)
		{
			((MNode *)ptr)->m_data = &ptr[size+cbkey];
			if(data)
				memcpy(((MNode *)ptr)->m_data, data, cbdata);
		}
		else
			((MNode *)ptr)->m_data = NULL;
	}
	return (void *)ptr;
}
void * MNode::operator new (size_t size, char *key)
{
	int cbkey = strlen(key)+1;

	char *ptr = (char *)phmalloc(size + cbkey);

	if(ptr)
	{
		((MNode *)ptr)->m_key = &ptr[size];
		memcpy(((MNode *)ptr)->m_key, key, cbkey);
		((MNode *)ptr)->m_data = NULL;
#ifdef MNODE_HAS_TYPE
		((MNode *)ptr)->m_type = MNODE_KEY_STRING;
#endif
	}
	return (void *)ptr;
}

Boolean MNodeDataDel(MNode *mn, void *p1, void *p2)
{
	if(mn)
	{
		void *p = mn->DataGet();
		if(p)
			phmfree(p);
	}
	return TRUE;
}

MNode * MNode2D::NextYGet()
{
	return m_Y_Next;
}
MNode * MNode2D::PrevYGet()
{
	return m_Y_Prev;
}
void MNode2D::NextYSet(MNode2D *mn2d)
{
	m_Y_Next = mn2d;
}
void MNode2D::PrevYSet(MNode2D *mn2d)
{
	m_Y_Prev = mn2d;
}
MNode * MNode3D::NextZGet()
{
	return m_Z_Next;
}
MNode * MNode3D::PrevZGet()
{
	return m_Z_Prev;
}
void MNode3D::NextZSet(MNode3D *mn3D)
{
	m_Z_Next = mn3D;
}
void MNode3D::PrevZSet(MNode3D *mn3D)
{
	m_Z_Prev = mn3D;
}


MNode *MNodeMulti::MNodeGet(int n)
{
	if(n < m_nNodes)
	{
		char *p = (char *)m_data;
		return (MNode *) &p[n * sizeof(MNode)];
	}
	return (MNode *)NULL;
}

MNodeMulti::MNodeMulti()
{
	if(m_nNodes)
	{
		MNode *mn;
		char *p = (char *)m_data;
		for(int i=0; i < m_nNodes; i++)
		{
			mn = new ((void *)p) MNode();

			mn->KeySet((NKey) this);
			p+=sizeof(MNode);
		}
	}
}

MNodeMulti::~MNodeMulti()
{
	if(m_nNodes)
	{
		MNode *mn;
		char *p = (char *)m_data;
		for(int i=0; i < m_nNodes; i++)
		{
			mn = (MNode *) p;
			// Don't call delete
			mn->~MNode();	
			p+=sizeof(MNode);
		}
	}

	if(m_strings)
		phmfree(m_strings);
}

void MNodeMulti::StringsSet(char *strings)
{
	m_strings = strings;
}
char *MNodeMulti::StringsGet(void)
{
	return m_strings;
}
void * MNodeMulti::operator new(size_t size, void *key, int cbkey, int nNodes)
{
	size_t cb;
	// A key with zero length is a string!
	if(key && !cbkey)
		cbkey = strlen((char *)key)+1;

	size_t cbnodes = nNodes * sizeof(MNode);

	cb = size + cbkey + cbnodes;

	char *ptr = (char *)phmalloc(cb);

	if(ptr)
	{
		MNodeMulti *mnm = (MNodeMulti *)ptr;
		mnm->m_nNodes = nNodes;
		mnm->m_key = &ptr[size];
		mnm->m_data = &ptr[size + cbkey];
	}
	return ptr;
}

MNodeHolder::~MNodeHolder()
{
	// Do nothing.
}
MNode *MNodeHolder::Add(char *key, char *data)
{
	int cbdata = (data) ? strlen(data)+1 : 0;
	MNode * mn = new ((void *)key, strlen(key)+1, (void *)data, cbdata) MNode();
	if(mn)
		Insert(mn);
	return mn;
}
MNode *MNodeHolder::Add(char *key, void *data, int cbdata)
{
	MNode * mn = new ((void *)key, strlen(key)+1, data, cbdata) MNode();
	if(mn)
		Insert(mn);
	return mn;
}
MNode *MNodeHolder::Add(void *key, int cbkey, char *data)
{
	MNode * mn = new (key, cbkey, (void *)data, strlen(data)+1) MNode();
	if(mn)
		Insert(mn);
	return mn;
}
MNode *MNodeHolder::Add(void *key, int cbkey, void *data, int cbdata)
{
	MNode * mn = new (key, cbkey, data, cbdata) MNode();
	if(mn)
		Insert(mn);
	return mn;
}

MNode *MNodeHolder::Find(MNode *mn)
{
	return Find(mn->KeyGet());
}

char *MNodeHolder::fixvalue(char *buffer, char *str)
{
 
        if(strlen(str) > (MAX_STRING/2))
                return NULL;
 
        char *dest = buffer;
 
	*dest=0;

        while(*str)
        {
                char c = *str++;
                if(c == '\\')
                {
                        char t = *str++;
                        switch(t)
                        {
                                case 'n': *dest++ = '\n'; break;
                                case 'r': *dest++ = '\r'; break;
                                case 't': *dest++ = '\t'; break;
                                default : *dest++ = t;
                        }
                }
                else
                        *dest++ = c;
        }
        *dest=0;

	return (strlen(buffer)) ? buffer : 0;
}
         

Boolean MNodeHolder::split(char *buffer)
{
	char *name = trim(buffer);

	if(*name == '#')
		return FALSE;

	char *value = strchr(buffer, '=');
		

	if(value)
	{
		*value++=0;
		value = trim(value);
		name = trim(name);
	}
	else
	{
		name = trim(name);
		value = name;
	}

	// printf("%s=%s\n", name, value);

	if(strlen(name))
	{
        	char buffer[MAX_STRING];
		value = fixvalue(buffer, value);

		if(value)
		{
			Add(name, value);
			return TRUE;
		}
	}
	return FALSE;
}

Boolean MNodeHolder::LoadTextValues(char *fname)
{
	char buffer[MAX_STRING];
	MStdioFile mf;

	if(!mf.open(fname, "r"))
	{
		mferr.mfprintf("Could not open config file %s\n", fname);
		return FALSE;
	}

	while(mf.gets(buffer, sizeof(buffer)))
		split(buffer);

	return TRUE;
}
typedef struct NodeDelete
{
	void *param1;
	void *param2;
	int nstack;
	int nstackmax;
	MTArray <MNode *> *stack;
}NODEDEL;
Boolean MNodeHolder::deletewalker(MNode *mn, void *param1, void *param2)
{
	NODEDEL *ndel = (NODEDEL *)param2;
	EnumFn fn = (EnumFn) param1;

	Boolean ret = TRUE;
	if(fn)
		ret = (*fn) (mn,ndel->param1, ndel->param2);

	ndel->stack->Add(mn);
	return ret;
}

void MNodeHolder::DeleteAll(EnumFn fn, void *p1, void *p2)
{
	NODEDEL ndel;
	MTArray <MNode *> stack(100);

	ndel.stack = &stack;
	ndel.param1 = p1;
	ndel.param2 = p2;

	Enum(deletewalker,(void *)fn,&ndel,FALSE);

	int n = stack.count();

	for(int i=0; i < n; i++)
		delete stack[i];

	ndel.stack=0;
	ndel.nstack=0;
	ndel.nstackmax=0;
}

#ifdef DO_MHNODE
void MHNode::operator delete(void *n)
{
	((MHNode *)n)->m_blkf->phmfree(((MHNode *)n)->m_hblk);
}
void * MHNode::operator new (size_t size, MFBlkfile *blkf, void *key,int cbkey,void *data=0,int cbdata=0)
{
	HBLK hblk;
	if(key && (cbkey == 0))
		cbkey = strlen((char *)key)+1;

	if(data && (cbdata == 0))
		cbdata = strlen((char *)data)+1;

	char *ptr = (char *) blkf->alloc(&hblk, size+cbkey+cbdata);

	if(ptr)
	{
		((MHNode *)ptr)->m_hblk = hblk;
		((MHNode *)ptr)->m_blkf = blkf;
		if(cbkey)
		{
			((MHNode *)ptr)->m_key = &ptr[size];
			memcpy(((MHNode *)ptr)->m_key, key, cbkey);
		}
		else
			((MHNode *)ptr)->m_key = NULL;

		if(cbdata)
		{
			((MHNode *)ptr)->m_data = &ptr[size+cbkey];
			memcpy(((MHNode *)ptr)->m_data, data, cbdata);
		}
		else
			((MHNode *)ptr)->m_data = NULL;
	}
	return (void *)ptr;
}
void * MHNode::operator new (size_t size, MFBlkfile *blkf, HBLK hblk)
{
	return blkf->get(hblk);
}
void * MHNode::operator new (size_t size)
{
	assert(0);
}

MHNode *MHNodeHolder::Add(char *key, char *data)
{
	int cbdata = (data) ? strlen(data)+1 : 0;
	MHNode * mn = new (m_blkf, (void *)key, strlen(key)+1, (void *)data, cbdata) MHNode();
	if(mn)
		Insert(mn);
	return mn;
}
MHNode *MHNodeHolder::Add(char *key, void *data, int cbdata)
{
	MHNode * mn = new (m_blkf, (void *)key, strlen(key)+1, data, cbdata) MHNode();
	if(mn)
		Insert(mn);
	return mn;
}
MHNode *MHNodeHolder::Add(void *key, int cbkey, char *data)
{
	MHNode * mn = new (m_blkf, key, cbkey, (void *)data, strlen(data)+1) MHNode();
	if(mn)
		Insert(mn);
	return mn;
}
MHNode *MHNodeHolder::Add(void *key, int cbkey, void *data, int cbdata)
{
	MHNode * mn = new (m_blkf, key, cbkey, data, cbdata) MHNode();
	if(mn)
		Insert(mn);
	return mn;
}

MHNode *MHNodeHolder::Find(MNode *mn)
{
	return Find(mn->KeyGet());
}
#endif
