/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to commercially license this library, the author
    can be reached at info@mohawksoft.com
*/ 
// #define TRACE_DEBUG

#include "Phoenix.h"
#include <sql.h>
#include <sqlext.h>
#include <sqltypes.h>

class MSqlODBC;

void 	odbc_error(MSqlODBC * sql, SQLHENV h);

#define ODBCSQL_ERR(err) ((err != SQL_SUCCESS) && (err != SQL_SUCCESS_WITH_INFO))


typedef struct _bindTemplate
{
	int 		type;
	int 		cb;
}BINDTMPL;

typedef struct _sqlBindParm
{
	SQLINTEGER	type;	// Holds current row
	SQLINTEGER	cb;	// Holds column count on [0] element
	SQLPOINTER	val;	// Holds hstmt on [0] element
}BINDVAL;

typedef struct _ODBCRESULT
{
	HSTMT		hstmt;	// Statement
	int		ncols;	// Number of colums
	int 		row;	// Current row
	int		offset;	// Offset from start
	BINDVAL		bval[1];
} ODBCRESULT;

class MSqlODBC : public MSql
{
	friend MSql *	NewMODBCSql(char *dbname, char *name, char *auth);
	protected:
	SQLHENV		m_Env;     	// Handle ODBC environment
	SQLHDBC		m_hdbc;    	// Handle connection
	ODBCRESULT *	m_cursor;	// The result set for the cursor

	Boolean		bind(ODBCRESULT *bval);
	ODBCRESULT *	build_result(HSTMT hstmt);
	ODBCRESULT *	set_template(int num, BINDTMPL *tmpl);
	void 		odbcerr(HSTMT hstmn, char *where=NULL);

	public:
			MSqlODBC(Boolean *fErr);
	virtual 	~MSqlODBC();
	SQLRESULT	ExecResult(char *sql);
	int		Exec(char *szsql);
	SQLRESULT	CursorFetch(int num, char *curs);
	void 		CursorDestroy(SQLRESULT sqlres);
	int		ColumnCount(SQLRESULT res);
	int		RowsCount(SQLRESULT res);
	int		ColumnIndex(SQLRESULT res, char *key);
	char *		ColumnValue(SQLRESULT res, int row, int ndx);
	char *		ColumnName(SQLRESULT res, int ndx);
	void		FreeResult(SQLRESULT sqlres);
	int 		ColumnType(SQLRESULT sqlres, int ndx);
	Boolean		CursorCreate(char *name, char *sql, Boolean fBinary);	
	Boolean		Begin(void);
	Boolean 	connect(char *dbname, char *name, char *auth);
};


// Default result translation
// Translates everything to strings
static void odbc_ascii_translate(int col, int *type, unsigned int *cb)
{
	TRACE_FUNCT("odbc_translate");
	if(*type == SQL_VARCHAR)
	{
		*type = SQL_C_CHAR;
		*cb = 1024;
	}
	else if(*type == SQL_NUMERIC)
	{
		*type = SQL_C_CHAR;
		*cb = 64;
	}
	else if(*type == SQL_DATETIME)
	{
		*type = SQL_C_CHAR;
		*cb = 64;
	}
	else if(*type != SQL_C_CHAR)
	{
		*type = SQL_C_CHAR;
		*cb = 128;
	}
}
// Binary result translation
// Performs minimal translation
static void odbc_binary_translate(int col, int *type, unsigned int *cb)
{
	TRACE_FUNCT("odbc_translate");
	if(*type == SQL_VARCHAR)
	{
		*type = SQL_C_CHAR;
		*cb = 1024;
	}
	else if(*type == SQL_NUMERIC)
	{
		*type = SQL_C_DOUBLE;
		*cb = 64;
	}
}
// Report ODBC errors.
void MSqlODBC::odbcerr(HSTMT hstmt, char * where)
{
	TRACE_FUNCT("odbcerr");
	SWORD cb;
	SDWORD sdw;
	char buffer[1024];
	SQLError(m_Env, m_hdbc, hstmt, NULL, &sdw, 
			(SQLCHAR *)buffer, sizeof(buffer), &cb);
	if(!where)
		where = "unknown";

	mferr.mfprintf("%s [%ld] (%s)\n", buffer, sdw, where);
}

// From a result template, translate and create BINDVAL.
ODBCRESULT *MSqlODBC::set_template(int n, BINDTMPL *tmpl)
{
	TRACE_FUNCT("set_template");
	// ODBC uses 1 to start, use 0 for book keeping.
	BINDVAL *bval=NULL;
	ODBCRESULT *podbcr;
	int cb=sizeof(ODBCRESULT)+sizeof(BINDVAL)*n;
	unsigned int *cba = (unsigned int *)alloca(sizeof(int) * n);
	int i;

	// Go through template and get info about
	// the nature and size of the template
	for(i=0; i < n; i++)
	{
		switch(tmpl[i].type)
		{
			case  SQL_C_TYPE_DATE:
			case  SQL_C_DATE:
				cba[i] = sizeof(SQL_DATE_STRUCT);
				break;
			case  SQL_C_TYPE_TIME:
			case  SQL_C_TIME:
				cba[i] = sizeof(SQL_TIME_STRUCT);
				break;
			case  SQL_C_TYPE_TIMESTAMP:
			case  SQL_C_TIMESTAMP:
				cba[i] = sizeof(SQL_TIMESTAMP_STRUCT);
				break;

			case  SQL_INTEGER:
				cba[i] = sizeof(int);
				break;
			case  SQL_NUMERIC:
			case  SQL_VARCHAR:
			case  SQL_C_CHAR:
			case  SQL_C_BINARY:
				cba[i]  = tmpl[i].cb;
				break;
			case  SQL_C_SLONG:
				cba[i]  = sizeof(signed long);
				break;
			case  SQL_C_SSHORT:
				cba[i]  = sizeof(signed short);
				break;
			case  SQL_C_ULONG:
				cba[i]  = sizeof(unsigned long);
				break;
			case  SQL_C_USHORT: 
				cba[i]  = sizeof(unsigned short);
				break;
			case  SQL_C_DOUBLE:
				cba[i]  = sizeof(double);
				break;
			case  SQL_C_FLOAT:
				cba[i]  = sizeof(float);
				break;
			default:
				mferr.mfprintf("Invalid SQL type col %d = %d\n", i, tmpl[i].type);
				return FALSE;
		}

		(m_translate)(i, &tmpl[i].type, &cba[i]);

		// Even shorts end up being the size of ints.
		if(cba[i] < sizeof(int))
			cb += sizeof(int);
		else
			cb += cba[i];
	}

	// Allocate space for everything.
	podbcr = (ODBCRESULT *) phmalloc(cb);
	bval = &podbcr->bval[0];

	// Find beginning of freestore
	char *p = (char *)&bval[n]; 

	podbcr->ncols = n;

	for(i=0; i < n; i++)
	{
		bval[i].type = tmpl[i].type;
		bval[i].cb = cba[i];
		bval[i].val = (SQLPOINTER) p;
		p+= (cba[i] < sizeof(int)) ? sizeof(int) : cba[i];
	}
	return podbcr;
}
// Binds a statement to a set of results in BINDVAL.
Boolean MSqlODBC::bind(ODBCRESULT *podbcr)
{
	TRACE_FUNCT("bind");
	int num = podbcr->ncols;

	for(int i=0; i < num; i++)
	{
		if(SQLBindCol(podbcr->hstmt, i+1, podbcr->bval[i].type, podbcr->bval[i].val, 
					podbcr->bval[i].cb, &podbcr->bval[i].cb) != SQL_SUCCESS)
		{
			odbcerr(podbcr->hstmt, "trying to bind results");
			return FALSE;
		}
	}
	return TRUE;
}

// Builds and binds a result structure.
ODBCRESULT *MSqlODBC::build_result(HSTMT hstmt)
{
	TRACE_FUNCT("build_result");
	SWORD sw;
	ODBCRESULT *podbcr;

	if(SQLNumResultCols(hstmt, &sw)!=SQL_SUCCESS)
	{
		odbcerr(hstmt, "Getting number of columns");
		return NULL;
	}

	BINDTMPL *ptmpl = (BINDTMPL *)alloca(sw *sizeof(BINDTMPL));

	for(int i=0; i < sw; i++)
	{
		SWORD n;
		SDWORD type;
		SDWORD len;
		int col = i+1;
		if(SQLColAttributes(hstmt,col, SQL_COLUMN_TYPE,NULL,(SWORD)0, &n, &type) !=SQL_SUCCESS)
		{
			odbcerr(hstmt, "using SQL_C_CHAR");
			type = SQL_C_CHAR;
		}
		if(SQLColAttributes(hstmt,col, SQL_COLUMN_LENGTH,NULL,(SWORD)0, &n, &len) !=SQL_SUCCESS)
		{
			odbcerr(hstmt, "using 256");
			len = 256;
		}
		ptmpl[i].cb = len;
		ptmpl[i].type = type;
	}

	podbcr = set_template(sw, ptmpl);

	if(podbcr)
	{
		podbcr->offset = 0;
		podbcr->row = -1;
		podbcr->hstmt = hstmt;
	}

	if(!bind(podbcr))
	{
		phmfree(podbcr);
		podbcr = NULL;
	}
	return podbcr;
}

// Executes SQL, does not collect any returned results.
// An update which does not effect any rows :e #
//
int MSqlODBC::Exec(char *sql)
{
	TRACE_FUNCT("Exec");
	int retval;
	unsigned long err;
	HSTMT hstmt;

	TRACE_STR(sql);

	err = SQLAllocStmt(m_hdbc, &hstmt);

	if(err != SQL_SUCCESS)
		return SQLERR_FAIL;

	err = SQLExecDirect(hstmt, (SQLCHAR *)sql, strlen(sql));

	if(err == SQL_SUCCESS)
	{
		SDWORD rows=0;
		SQLRowCount(hstmt, &rows);
		retval = (rows > 0) ? SQLERR_OK_AFFECTED : SQLERR_OK_NONE;
	}
	else
		retval = SQLERR_FAIL;

	SQLFreeStmt(hstmt, SQL_DROP);

	return retval;
}
SQLRESULT MSqlODBC::ExecResult(char *sql)
{
	TRACE_FUNCT("ExecResult");
	unsigned long err;
	HSTMT hstmt;

	TRACE_STR(sql);

	err = SQLAllocStmt(m_hdbc, &hstmt);

	if(err != SQL_SUCCESS)
	{
		mferr.mfprintf("Could not get Stmt\n");
		return NULL;
	}

	err = SQLPrepare(hstmt, (SQLCHAR *)sql, SQL_NTS);

	if(err != SQL_SUCCESS)
	{
		mferr.mfprintf("Could not Prepare Statement [%s]\n", sql);
		return NULL;
	}

	err = SQLExecute(hstmt);

	if(err != SQL_SUCCESS && err != SQL_SUCCESS_WITH_INFO)
	{
		mferr.mfprintf("Could not ExecDirect\n");
		SQLFreeStmt(hstmt, SQL_DROP);
		hstmt = NULL;
	}
	return (SQLRESULT)build_result(hstmt);
}
void MSqlODBC::FreeResult(SQLRESULT sqlres)
{
	TRACE_FUNCT("FreeResult");
	ODBCRESULT *podbcr = (ODBCRESULT *)sqlres;
	SQLFreeStmt(podbcr->hstmt, SQL_DROP);
	phmfree(podbcr);
}
// Fake a cursor
Boolean MSqlODBC::CursorCreate(char *name, char *sql, Boolean fBinary)
{
	if(fBinary)
		m_translate = odbc_binary_translate;


	m_cursor = (ODBCRESULT *)ExecResult(sql);

	return (m_cursor) ? TRUE : FALSE;
}

// Does ODBC use cursors?
SQLRESULT MSqlODBC::CursorFetch(int num, char *curs)
{
	TRACE_FUNCT("Fetch");

	// Assume an offset
	m_cursor->offset = m_cursor->row+1;
	
	return m_cursor;
}
void MSqlODBC::CursorDestroy(SQLRESULT sqlres)
{
	this->FreeResult(sqlres);
}

Boolean MSqlODBC::Begin()
{ // Do we need begin in ODBC
	TRACE_FUNCT("Begin");
	return TRUE;
}
int MSqlODBC::ColumnCount(SQLRESULT sqlres)
{
	TRACE_FUNCT("ColumnCount");
	return ((ODBCRESULT *)sqlres)->ncols;;
}
int MSqlODBC::RowsCount(SQLRESULT sqlres)
{		
	TRACE_FUNCT("RowsCount");
	ODBCRESULT *podbcr = (ODBCRESULT *)sqlres;
	return podbcr->ncols;
}

int MSqlODBC::ColumnIndex(SQLRESULT sqlres, char *key)
{
	TRACE_FUNCT("ColumnIndex");
	UWORD	icol;
	for(icol=0; 1; icol++)
	{
		char *colname = ColumnName(sqlres, icol);

		if(!colname)
			break;
		else if(!strcasecmp(colname, key))
			return icol;
	}
	return -1;
}
char *	MSqlODBC::ColumnValue(SQLRESULT sqlres, int rawrow, int ndx)
{
	TRACE_FUNCT("ColumnValue");
	ODBCRESULT *podbcr = (ODBCRESULT *)sqlres;
	int row = rawrow + podbcr->offset;

	if(ndx >= podbcr->ncols)
	{
		TRACE_STR("ndx > ncols");
		return NULL;
	}
	// ODBC does not go backward.
	if( podbcr->row > row)
	{
		TRACE_STR("current row > row");
		return NULL;
	}

	while(podbcr->row < row )
	{
		unsigned long err = SQLFetch(podbcr->hstmt);

		if(err == SQL_NO_DATA_FOUND)
		{
			TRACE_STR("End of data");
			return NULL; // Not an error
		}
		else if(err != SQL_SUCCESS)
		{	
			odbcerr(podbcr->hstmt, "Fetching next data set");
			return NULL;
		}
		podbcr->row++;
	}
	return (char *) podbcr->bval[ndx].val;
}

int MSqlODBC::ColumnType(SQLRESULT sqlres, int ndx)
{
	TRACE_FUNCT("ColumnType");
	ODBCRESULT *podbcr = (ODBCRESULT *)sqlres;

	if(ndx < podbcr->ncols)
		return podbcr->bval[ndx].type;
	else
		return SQLTYPE_UNDEFINED;
}
char *	MSqlODBC::ColumnName(SQLRESULT sqlres, int ndx)
{
	TRACE_FUNCT("ColumnName");
	static char buffer[1024];
	ODBCRESULT *podbcr = (ODBCRESULT *)sqlres;
	SWORD cbDesc;
	SDWORD fdesc;
	unsigned long err;

	err = SQLColAttribute(podbcr->hstmt, ndx+1, 
		SQL_COLUMN_NAME, 
		buffer, (SWORD)sizeof(buffer), 
		(SWORD *)&cbDesc, (SDWORD *)&fdesc);

	if(err != SQL_SUCCESS)
	{
		err = SQLColAttribute(podbcr->hstmt, ndx+1, 
			SQL_DESC_BASE_COLUMN_NAME, 
			buffer, (SWORD)sizeof(buffer), 
			(SWORD *)&cbDesc, (SDWORD *)&fdesc);
	}
	if(err != SQL_SUCCESS)
		odbcerr(podbcr->hstmt, "Getting Column Name");

	return err == SQL_SUCCESS ? buffer : NULL;
}

MSqlODBC::~MSqlODBC()
{
	SQLFreeHandle(SQL_HANDLE_DBC, m_hdbc);
	SQLFreeHandle(SQL_HANDLE_ENV, m_Env);
}

MSqlODBC::MSqlODBC(Boolean *fError)
{
	long err;

	// m_lifetime = 1000;

	err=SQLAllocHandle(SQL_HANDLE_ENV,SQL_NULL_HANDLE,&m_Env);

	if (ODBCSQL_ERR(err))
		odbc_error(this, m_Env);

	err = SQLSetEnvAttr(m_Env, SQL_ATTR_ODBC_VERSION, (void*)SQL_OV_ODBC3, 0); 

	if (ODBCSQL_ERR(err))
		odbc_error(this, m_Env);

	err = SQLAllocHandle(SQL_HANDLE_DBC, m_Env, &m_hdbc); 

	if (ODBCSQL_ERR(err))
		odbc_error(this, m_Env);

	SQLSetConnectAttr(m_hdbc, SQL_LOGIN_TIMEOUT, (SQLPOINTER *)5, 0);
	
	m_translate = odbc_ascii_translate;
}

Boolean MSqlODBC::connect(char *dbname, char *name, char *auth = "")
{
	long err = SQLConnect(m_hdbc, (SQLCHAR*) dbname, SQL_NTS,
                                       (SQLCHAR*) name, SQL_NTS,
                                       (SQLCHAR*) auth, SQL_NTS);
	return (ODBCSQL_ERR(err)) ? FALSE : TRUE;
}

void odbc_error(MSqlODBC * sql, SQLHENV h)
{
	printf("Error in ODBC class");
	exit(0);
}

MSql *NewMODBCSql(char *dbname, char *name, char *auth)
{
	Boolean fErr=FALSE;
	MSqlODBC *sql = new MSqlODBC(&fErr);

	if(!fErr)
	{
		if(sql->connect(dbname, name, auth))
			return sql;
	}
	delete sql;
	return NULL;
}
