/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to commercially license this library, the author
    can be reached at info@mohawksoft.com
*/ 
#include "Phoenix.h"
#include "libpq-fe.h"


class MPGSql : public MSql
{
	protected:
	friend MSql *NewMPGSql(char *host, char *db, char *port);
	PGconn	* m_conn;
	MPGSql();
	Boolean	   Connect(char *host, char *db, char *sys=NULL, char *opt=NULL, char * console=NULL);
	void	   Disconnect();

	public:
	~MPGSql();
	SQLRESULT  ExecResult(char *sql);
	int 	   Exec(char *szsql);
	SQLRESULT  CursorFetch(int num, char *curs);
	int 	   ColumnCount(SQLRESULT res);
	int 	   RowsCount(SQLRESULT res);
	int 	   ColumnIndex(SQLRESULT res, char *key);
	char *	   ColumnValue(SQLRESULT res, int row, int ndx);
	char *	   ColumnName(SQLRESULT res, int ndx);
	void	   FreeResult(SQLRESULT sqlres);
	void 	   CursorFreeResult(SQLRESULT sqlres);
	void 	   CursorDestroy(char *curs);

	Boolean	   Begin(void);
};


MPGSql::MPGSql()
{
	m_conn = NULL;
}

MSql *NewMPGSql(char *host, char *db, char *port)
{
	MPGSql *pgsql = new MPGSql();

	if(pgsql)
	{
		if(!pgsql->Connect(host,db,port))
		{
			delete pgsql;
			return NULL;
		}
	}
	return pgsql;
	
}
MPGSql::~MPGSql()
{
	Disconnect();
}

Boolean MPGSql::Connect(char *host, char *db, char *sys, char *opt, char * console)
{
	m_conn = PQsetdb(host,sys,opt,console, db);

	if (PQstatus(m_conn) == CONNECTION_BAD)
	{
       		mferr.mfprintf("Connection to database '%s' failed.\n", db);
		mferr.mfprintf("%s", PQerrorMessage(m_conn));
		return FALSE;
	}
	return TRUE; 
}

void MPGSql::Disconnect()
{
	PQfinish(m_conn);
}

SQLRESULT MPGSql::ExecResult(char *sql)
{
	PGresult *res = PQexec(m_conn, sql);

	if(res)
	{
		int n = PQresultStatus(res);
		if( (n ==PGRES_BAD_RESPONSE) || (n ==PGRES_NONFATAL_ERROR) || (n ==PGRES_FATAL_ERROR) )
		{
			mferr.mfprintf("could not execute \n%s\n", sql);
			mferr.mfprintf("%s\n", PQresultErrorMessage(res));
			PQclear(res);
			res = NULL;
		}
	}
	return res;
}
int MPGSql::Exec(char *szsql)
{
	int n=PGRES_BAD_RESPONSE;
	PGresult *res = PQexec(m_conn, szsql);

	if(res)
	{
		n = PQresultStatus(res);
		if( (n ==PGRES_BAD_RESPONSE) || (n ==PGRES_NONFATAL_ERROR) || (n ==PGRES_FATAL_ERROR) )
		{
			mferr.mfprintf("could not execute \n%s\n", szsql);
			mferr.mfprintf("%s\n", PQresultErrorMessage(res));
			n = SQLERR_FAIL;
		}
		else if(n == PGRES_TUPLES_OK)
			n = SQLERR_OK_DATA;
		else if (n == PGRES_COMMAND_OK)
		{
			int cmd = atoi(PQcmdTuples(res));
			if(cmd > 0)
				n = SQLERR_OK_AFFECTED;
			else
				n = SQLERR_OK_NONE;
		}
		PQclear(res);
	}
	return n;
}
SQLRESULT MPGSql::CursorFetch(int num, char *curs)
{
	char buffer[MAX_STRING];
	sprintf(buffer, "fetch forward %d in %s", num, curs);

	PGresult *res = PQexec(m_conn, buffer);

	if(res)
	{
		if(PQresultStatus(res) != PGRES_TUPLES_OK)
		{
			mferr.mfprintf("could not execute \n%s\n", buffer);
			mferr.mfprintf("%s\n", PQresultErrorMessage(res));
			PQclear(res);
			res = NULL;
		}
	}
	return res;
}
void MPGSql::CursorFreeResult(SQLRESULT sqlres)
{
	PQclear((PGresult *)sqlres);
}
void MPGSql::CursorDestroy(char *curs)
{
	Execf("close %s", curs);
}

int MPGSql::ColumnCount(SQLRESULT res)
{
	return PQnfields((PGresult *)res);
}
int MPGSql::RowsCount(SQLRESULT res)
{
	return PQntuples((PGresult *)res);
}
int MPGSql::ColumnIndex(SQLRESULT res, char *key)
{
	return PQfnumber((PGresult *)res, key);
}

char *MPGSql::ColumnValue(SQLRESULT res, int row, int ndx)
{
	return PQgetvalue((PGresult *)res, row, ndx);
}

char *MPGSql::ColumnName(SQLRESULT res, int ndx)
{
	return PQfname((PGresult *)res, ndx);
}
void MPGSql::FreeResult(SQLRESULT sqlres)
{
	PQclear((PGresult *)sqlres);
}

Boolean MPGSql::Begin(void)
{
	return ErrorToBool( Exec("BEGIN") );
}
