/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to commercially license this library, the author
    can be reached at info@mohawksoft.com
*/ 

/*
NOTE: This is a BAD algorithm, while it performs OK for smallish
sets of data, it is horrible with larger sets. If you wish to 
do anthing with this, you should seriously rethink your position. 

In an application where you need to maintain a very large number
of fairly small sorted lists this may do the trick. The overhead 
for this is a fixed size, plus a theoretical 1:1 relationship 
between data and memory use.
*/

#include "Phoenix.h"

// #define BRUTE_FORCE

#define GROW(X)	if(X>=m_elementMax) {if(!grow(X)) throw((void *)NULL);}
#define MOVE(X) memmove(&m_array[X+1], &m_array[X],(m_Count-X)*sizeof(IPAIR))

unsigned int MIPairList::s_totalMem=0;

Boolean MIPairList::setmax(int newmax)
{
	IPAIR *p;

	if(newmax < m_elementMax)
		return TRUE;

	unsigned int cb = m_elementMax * sizeof(IPAIR);
	
	// Remove old count of bytes
	s_totalMem -= cb;

	cb = newmax * sizeof(IPAIR);

	// Add new count of bytes
	s_totalMem += cb;
	
	// will do initial malloc if m_table==NULL
	p = (IPAIR *)phmrealloc((void *)m_array, cb);

	if(p)
	{
		m_array = p;
		m_elementMax = newmax;
		return TRUE;
	}
	else
	{

		p = (IPAIR *) phmalloc(cb);
		assert(p);
		memcpy(p, m_array, sizeof(IPAIR) * m_Count);
		phmfree((void*)m_array);
		m_array = p;
		m_elementMax = newmax;
		return TRUE;
	}
	return FALSE;
}
	
/** Grow to allow n items in array */
Boolean MIPairList::grow(int n)
{
	if(m_elementMax > n)
		return TRUE;
	int newmax = m_elementMax * 2;

	if(newmax < n)
		newmax = n+1;
	return setmax(newmax);
}

MIPairList::MIPairList(int start)
{
	m_elementMax=0;
	m_Count=0;
	m_array = NULL;
	setmax(start);
}
MIPairList::~MIPairList()
{
	s_totalMem -= (m_elementMax * sizeof(IPAIR));
	phmfree((void *)m_array);
}
int MIPairList::FindKey(MPLKEY key, int *pinsert)
{
#ifndef BRUTE_FORCE
	register int start=0;
	register int range=m_Count;
	// The bsearch break point is tuned by a profiler
	// it is a MAGIC number, you need not change it
	// The insert function performs better with a 
	// larger scan
	while(range > 16)
	{
		int nrange = range/2;
		int ndx=start+nrange+(range&1);
		MPLKEY pkey = m_array[ndx].key;
		if(pkey == key)
			return ndx;
		if(pkey > key)
			range = nrange;
		else
		{
			range = nrange;
			start = ndx;
		}
	}
	int end = start+range+2;
	end = (end > m_Count) ? m_Count : end;
	start = (start > 0) ? start-1: 0;

	for(int i=start; i < end; i++)
	{
		if(m_array[i].key == key)
			return i;
		else if(m_array[i].key > key)
		{
			if(pinsert)
				*pinsert=i;
			return -1;
		}
	}
	if(pinsert)
		*pinsert = m_Count;
	return -1;
#else
	for(int i=0; i < m_Count; i++)
	{
		if(m_array[i].key == key)
			return i;
		else if(m_array[i].key > key)
		{
			if(pinsert)
				*pinsert=i;
			return -1;
		}
	}
	if(pinsert)
		*pinsert = m_Count;
	return -1;
#endif
}

Boolean MIPairList::AddIndex(MPLKEY key, MPLVAL value, int ndx)
{
	m_Count++;
	GROW(m_Count);
	MOVE(ndx);
	m_array[ndx].key=key;
	m_array[ndx].value=value;
	return TRUE;
}

MPLVAL & MIPairList::operator [] (MPLKEY key)
{
	int ndxinsert;
	int ndx = FindKey(key, &ndxinsert);

	if(ndx == -1)	// New value
	{
		AddIndex(key, 0, ndxinsert);
		ndx = ndxinsert;
	}
	return m_array[ndx].value;
}

/** return number of objects in array */
int MIPairList::count()
{
	return m_Count;
}
void MIPairList::trunk(int ndx)
{
	m_Count = MIN(m_Count, ndx);
}
