/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to professionally license this library, the author
    can be reached at info@mohawksoft.com
*/ 
#include "Phoenix.h"
#ifdef GCC
#ifndef DARWIN
#include <dlfcn.h>
#else
#include <mach-o/dyld.h>
#endif

#endif

#ifdef DARWIN

#define RTLD_NOW	0
char dyn_buf[1024];
char noerror[]="noerror";
static char * dyn_error = noerror;
void * dlopen(char *filename, int mode)
{
        NSObjectFileImage image;

	dyn_error = noerror;
        if (NSCreateObjectFileImageFromFile(filename, &image) != NSObjectFileImageSuccess)
	{
		dyn_error = "Could not open file";
                return NULL;
	}
        return NSLinkModule(image, filename, NSLINKMODULE_OPTION_PRIVATE);
}

void dlclose(void *handle)
{
	dyn_error = noerror;
        NSUnLinkModule(handle, FALSE);
        return;
}

void *dlsym(void *handle, char *funcname)
{
        NSSymbol symbol;
	int cb = strlen(funcname)+2;
	char *symname = (char *)ALLOCA(cb);
        sprintf(symname, "_%s", funcname);

	dyn_error = noerror;

	symbol = NSLookupSymbolInModule(handle, symname);
        if (symbol)
        {
                return (void *) NSAddressOfSymbol(symbol);
        }
        else
	{
		sprintf(dyn_buf, "Symbol [%s] not found", symname);
		dyn_error = dyn_buf;
                return (void *) NULL;
	}
}

char *dlerror(void)
{
	return dyn_error;
}
#endif
void * 	MPlugin::def_init(int ver, int mod, PLUGFN **imports, PLUGFN *exports)
{
	return NULL;
}
int 	MPlugin::def_handler(void *pdata, int msg, void *p1, void *p2)
{
	return NO_PLUGIN_FUNCTION;
}
void	MPlugin::def_shutdown(void *pdata)
{
}

MPlugin::MPlugin(PLUGFN *exports)
{
	m_exports = exports;
	m_imports = NULL;
	m_dll = NULL;
	m_pdata = NULL;
	m_ver = 0;
	m_init = def_init;
	m_handler = def_handler;
	m_shutdown = def_shutdown;
}

MPlugin::~MPlugin()
{
	if(m_dll && m_shutdown)
		(*m_shutdown)(m_pdata);
}

#ifdef GCC
void *MPlugin::getpfn(char *name)
{
	void * pfn = dlsym(m_dll, name);
	if(!pfn)
		mferr.mfprintf("MPlugin dlsym Warning: %s\n", dlerror());
	return pfn;
}

#define MPLUGINDIR "/usr/lib/msession/"

Boolean MPlugin::openlib(char *name)
{
	m_dll = dlopen(name, RTLD_NOW);

	if(!m_dll)
	{
		char buffer[PATH_MAX];
		if(strlen(MPLUGINDIR) + strlen(name) + 1) {
			strcpy(buffer, MPLUGINDIR);
			strcat(buffer, name);
			m_dll = dlopen(buffer, RTLD_NOW);
			if(m_dll)
				return TRUE;
		}
		if(!getcwd(buffer, sizeof(buffer) - strlen(name) - 1))
			return FALSE;
		sprintf(&buffer[strlen(buffer)], "/%s", name);

		m_dll = dlopen(buffer, RTLD_NOW);
	}

	if(!m_dll)
		mferr.mfprintf("MPlugin dlopen Error: %s %s\n", name, dlerror());

	return m_dll ? TRUE : FALSE;

}

void MPlugin::closelib()
{
	dlclose(m_dll);
}
#endif
#ifdef WIN32
void *MPlugin::getpfn(char *name)
{
	void * pfn = GetProcAddress(m_dll, name);
	if(!pfn)
		mferr.mfprintf("MPlugin Warning: %d\n", GetLastError());
	return pfn;
}

Boolean MPlugin::openlib(char *name)
{
	m_dll = LoadLibrary(name);

	if(!m_dll)
		mferr.mfprintf("MPlugin Error: %s\n", name, GetLastError());
	return m_dll ? TRUE : FALSE;
}

void MPlugin::closelib()
{
	FreeLibrary(m_dll);
}
#endif

void MPlugin::Close()
{
	if(m_dll)
	{
		if(m_shutdown)
			(*m_shutdown)(m_pdata);
		closelib();
	}
	m_imports = NULL;
	m_dll = NULL;
	m_pdata = NULL;
	m_ver = 0;
	m_init = def_init;
	m_handler = def_handler;

}
int MPlugin::Open(char *module, char *name, int mod)
{
	m_ver = 0;

	if(openlib(module))
	{	
		char initfn[128];
		char handfn[128];
		char shutfn[128];

		snprintf(initfn, sizeof(initfn), "%s_INIT_", name);
		snprintf(handfn, sizeof(handfn), "%s_HANDLER_", name);
		snprintf(shutfn, sizeof(shutfn), "%s_SHUTDOWN_", name);

		
		m_init = (PlugInit)getpfn(initfn);
		m_handler = (PlugHandler)getpfn(handfn);
		m_shutdown = (PlugShutdown)getpfn(shutfn);

		if(m_init)
			m_pdata = (*m_init)(MPLUGIN_VER,mod, &m_imports, m_exports);
		else
			m_init = def_init;

		if(m_handler)
			m_ver = (*m_handler)(m_pdata, PLUGMSG_VER,NULL,NULL);
		else
			m_handler = def_handler;
	}
	return m_ver;
}
PLUGFN *MPlugin::getimports(void)
{
	return m_imports;
}
PLUGFN *MPlugin::getexports(void)
{
	return m_exports;
}
PlugHandler MPlugin::SetHandler(PlugHandler plh)
{
	PlugHandler plhT = m_handler;
	m_handler = plh;
	return plhT;
}
void MPlugin::SetExports(PLUGFN *exports)
{
	m_exports = exports;
}

PlugShutdown MPlugin::SetShutdown(PlugShutdown pls)
{
	PlugShutdown plsT = m_shutdown;
	m_shutdown = m_shutdown;
	return plsT;
}


MPlugMgr::MPlugMgr(PLUGFN *exports)
{
	m_exports = exports;
	// Do nothing?
}
MPlugMgr::~MPlugMgr()
{
	CloseAll();
}
void MPlugMgr::CloseAll(void)
{
	int n= m_plugins.count();

	m_functions.DeleteAll(NULL,NULL,NULL);
	m_functions.clear();
	
	for(int i=0; i < n; i++)
	{
		delete m_plugins[i];
		m_plugins[i] = 0;
	}
	m_plugins.trunk(0);

}
int MPlugMgr::MapFunctions(PLUGFN *plgf)
{
	int i=0;
	if(plgf)
	{
		for(i=0; plgf[i].funct && plgf[i].name; i++)
		{
#ifdef TRACE_DEBUG
			mferr.mfprintf("Loading:%s\n", plgf[i].name);
#endif
			MNode *mn = m_functions.FindStr(plgf[i].name);
			if(mn != NULL)	// We have name overlaps!!
			{
				mferr.mfprintf("Could not add function \"%s\"\n", plgf[i].name);
				mferr.mfprintf("Reason: Name already registered\n");
				return 0;
			}
		}

		// OK, lets map them!
		for(i=0; plgf[i].funct && plgf[i].name; i++)
		{
			MNode * mnfn = new (plgf[i].name) MNode();
			assert(mnfn);
			mnfn->DataSet((NData)plgf[i].funct);
			if(!m_functions.Insert(mnfn))
			{
				mferr.mfprintf("Could not add function \"%s\"\n", plgf[i].name);
				mferr.mfprintf("Reason: Failure in tree manager\n");
				delete mnfn;
				assert(0);
			}
		}
		return i;
	}
	return 0;
}
MPlugin *MPlugMgr::LoadPlugin(char *name, char *path)
{
	MPlugin *pplugin = new MPlugin(m_exports);

	if(pplugin->Open(path, name))
	{
		MapFunctions(pplugin->getimports());
		m_plugins.Add(pplugin);
		return pplugin;
	}
	else
		mferr.mfprintf("Could not load plugin \"%s\"\n", path);

	delete pplugin;
	return NULL;
}

PlugFn MPlugMgr::FindFn(char *name)
{
	MNode *mn = m_functions.FindStr(name);
	return (mn) ? (PlugFn) mn->DataGet() : (PlugFn) NULL;
}
int MPlugMgr::CallPluginFn(char *name, void *p1, void *p2)
{
	MNode *mn = m_functions.FindStr(name);
	if(mn)
	{
		// fn should always be valid. If it isn't
		// then there is memory corruption elsewhere!
		PlugFn fn = (PlugFn) mn->DataGet();
		return (*fn)(p1,p2);
	}
	return 0;
}

void MPlugMgr::SetExports(PLUGFN *exports)
{
	m_exports = exports;
}
// Splits up a plugin string to a name and path.
// name:subname
void MPlugMgr::GetPluginInfo(char *optarg, char *name, char *path)
{
	char *p = strchr(optarg, ':');

	if(!p)
	{
		strcpy(path, optarg);
		strcpy(name, "default");
	}
	else
	{
		int cbname = (p-optarg);
		memcpy(name, optarg, cbname);
		name[cbname]=0;
		p++;
		strcpy(path, p);
	}
}
// Load a function pluingin.
Boolean MPlugMgr::LoadFunctionMod(char *optarg)
{
	char name[PATH_MAX];
	char path[PATH_MAX];

	if(!optarg)
		return FALSE;

	GetPluginInfo(optarg, name, path);
	
	if(!LoadPlugin(name, path))
		return FALSE;
	else
		return TRUE;
}

// Get a plugin function pointer.
PlugFn MPlugMgr::GetFn(char *fnname, PlugFn fnDefault)
{
	PlugFn fn = FindFn(fnname);
	return (fn) ? fn : fnDefault;
}



#if 0 // test
main()
{
	MPlugin mf;
	mf.Open("/home/markw/dl/dll.so", "demo");
	mf.pcall(10);
	mf.pcall(11);
	mf.pcall(12);
}
#endif

