/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to commercially license this library, the author
    can be reached at info@mohawksoft.com
*/ 
// #define TRACE_DEBUG
#include "Phoenix.h"

// #define FAST_LOCK

MRWMutex::MRWMutex()
{
	m_readers = 0;
	m_writers = 0;
	m_wrwait = 0;
}
void MRWMutex::LockRead()
{
	TRACE_FUNCT("LockRead");
	TRACE_VAL(m_readers);
	TRACE_VAL(m_writers);

#ifdef USE_GATEWAY
	// Gateway mutex regulates waiters on write mutex.
	// It is locked when a writer wishes to write.
	m_mtxGateway.lock();
	m_mtxGateway.unlock();
#endif
#ifdef FAST_LOCK
// This should only to be used on a uniprocessor system
// that has an atomic increment!!!
	if(m_readers)
		ATOMIC_INC(m_readers);
	else
	{
		m_mtxWrite.lock();
		m_readers++;
		m_mtxWrite.unlock();
	}
#else
	// Lock the global lock mutex
	m_mtxWrite.lock();

	TRACE_STR("mtxWrite locked");

	// Increment readers
	m_readers++;
	
	// Unlock write mutex
	m_mtxWrite.unlock();
#endif

	TRACE_FUNCT("LockRead:Exit");

}
void MRWMutex::UnlockRead()
{
	TRACE_FUNCT("UnlockRead");
	TRACE_VAL(m_readers);
	TRACE_VAL(m_writers);

#ifdef FAST_LOCK
	if(m_readers > 1)
		ATOMIC_DEC(m_readers);
	else
	{
		m_mtxWrite.lock();
		m_readers--;
		m_mtxWrite.unlock();
	}
#else
	// Lock the global lock mutex
	m_mtxWrite.lock();
	TRACE_STR("mtxWrite locked");
	// Dec readers
	m_readers--;

	// Debugging, bug in a program that call unlock
	// on an unlocked mutex.
	assert(m_readers >= 0);

	// Unlock mutex.
	m_mtxWrite.unlock();
#endif
	// Signal that a reader has left.
	m_mtxWrite.signal();

	TRACE_FUNCT("UnlockRead:Exit");
}
Boolean MRWMutex::LockWrite(int wait)
{
	TRACE_FUNCT("LockWrite");
	TRACE_VAL(m_readers);
	TRACE_VAL(m_writers);

	if(m_writers > 0 && wait == 0)
	{
		TRACE_STR("LockWrite: Fail, no wait");
		return FALSE;
	}

	time_t start = time(0);

	m_mtxWrite.lock();

	TRACE_STR("mtxWrite locked");

#ifdef USE_GATEWAY
	// Keep readers from starting.
	m_mtxGateway.lock();
	TRACE_STR("mtxGateway locked");
#endif
	// Unlocks write mutex and waits
	// reacquires mutex on signal.
	while(m_readers && ((time(0) - start) < wait))
	{
		TRACE_STR("RWLock Waiting");
		m_mtxWrite.wait();
	}

	TRACE_STR("RWLock done Waiting");
#ifdef USE_GATEWAY
	// re-allow readers to wait on write mutex
	m_mtxGateway.unlock();
	TRACE_STR("mtxGateway unlocked");
#endif
	if(m_readers)
	{
		// Failed to get write in time
		TRACE_STR("Timeout on write lock");
		m_mtxWrite.unlock();
		return FALSE;
	}
	TRACE_STR("Write lock");

	m_writers ++;
	
	TRACE_FUNCT("LockWrite:Exit");
	return TRUE;
}
void MRWMutex::UnlockWrite()
{
	TRACE_FUNCT("UnlockWrite");
	TRACE_VAL(m_readers);
	TRACE_VAL(m_writers);
	assert(m_readers == 0);
	assert(m_writers == 1);

	m_writers--;
	m_mtxWrite.unlock();
	m_mtxWrite.signal();
	TRACE_FUNCT("UnlockWrite:Exit");
}

void MRWMutex::UnlockWriteLockRead()
{
	TRACE_FUNCT("UnlockWriteLockRead");
	TRACE_VAL(m_readers);
	TRACE_VAL(m_writers);
	TRACE_VAL(m_mtxWrite.stat());
	assert(m_readers == 0);
	assert(m_writers == 1);
	m_readers=1;
	m_writers = 0;
	TRACE_STR("mtxRead lock");
	m_mtxWrite.unlock();
	m_mtxWrite.signal();
	TRACE_STR("mtxWrite unlocked");
	TRACE_FUNCT("UnlockWriteLockRead:Exit");
}
