/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to commercially license this library, the author
    can be reached at info@mohawksoft.com
*/ 

// #define TRACE_DEBUG
#include "Phoenix.h"
#include <errno.h>

#define LSTRIP(str)     while(isspace(*str) && *str) str++ 

static IPV4NM	s_allow_hba[MAX_HBA];
static int	s_count_hba = 0;

#ifdef WIN32
#define SOCKBUF	(char *)
#define OPTCAST	(char *)
#define SOCKCLOSE(s)	closesocket(s)
#define RECV(s,b,l,f)	::recv(s, (char FAR *)b,l,f)
#define SEND(s,b,l,f)	::send(s, (char FAR *)b,l,f)
#endif
#ifdef POSIX
#define SOCKBUF	(void *)
#define OPTCAST	(void *)
#define SOCKCLOSE(s)	close(s)
#define RECV(s,b,l,f)	::recv(s,b,l,f)
#define SEND(s,b,l,f)	::send(s,b,l,f)
#endif

#ifndef MSG_NOSIGNAL
#define MSG_NOSIGNAL 0
#endif
// "000.000.000.000
int MSock::FillIPV4(IPV4 *ipv4, char *string)
{
        ipv4->p[0]=ipv4->p[1]=ipv4->p[2]=ipv4->p[3]=0;

        for(int i=0; i < 4; i++)
        {
                if(!string)
                        return 0;

                LSTRIP(string);
 
                if(!*string)
                        return FALSE;
 
                ipv4->p[i] = atoi(string);
 
                string = strchr(string, '.');
 
                if(string)
                        string++;
		else
			break;

		if(!isdigit(*string))
			break;
        }
        return (ipv4->p[0]+ipv4->p[1]+ipv4->p[2]+ipv4->p[3]) ? TRUE : FALSE;
}

// "255.255.255.255/000.000.000.000
int MSock::FillIPV4NM(IPV4NM *ipv4nm, char *string)
{
	int len = strlen(string)+1;
	char *buf = (char *)ALLOCA(len);

	strcpy(buf, string);
        char *net = strchr(buf,'/');
        if(!net) 
		return FALSE;
	else
		*net++=0;

        return (MSock::FillIPV4(&ipv4nm->mask, buf) && MSock::FillIPV4(&ipv4nm->net, net));
}
 
int MSock::checkip(IPV4NM *ip, char *string)
{
        IPV4 ipv4;
        if(MSock::FillIPV4(&ipv4, string))
        {
                unsigned int t = ip->mask.raw & ipv4.raw;
                return (t == ip->net.raw);
        }
        return FALSE;
}
int MSock::checkip(IPV4NM *ip, unsigned int bits)
{
	unsigned int t = ip->mask.raw & bits;
	return (t == ip->net.raw);
}

Boolean MSock::checkhba(unsigned int uparts)
{
	for(int i=0; i < s_count_hba; i++)
	{
		if((s_allow_hba[i].mask.raw & uparts) == s_allow_hba[i].net.raw)
			return TRUE;
	}
#ifdef TRACE_DEBUG
	unsigned char *parts = (unsigned char *)&uparts;
	mferr.mfprintf("Reject IP: %d.%d.%d.%d\n", parts[0],parts[1],parts[2],parts[3]);
#endif
	return FALSE;
}

int MSock::allowhba(char *string)
{
	IPV4NM ip;

	if(s_count_hba < MAX_HBA)
	{
		if(FillIPV4NM(&ip, string))
		{
			s_allow_hba[s_count_hba++] = ip;
			return TRUE;
		}
	}
	return FALSE;
}


MFile * MSock::clone()
{
	return NULL;
}
ssize_t MSock::read(void *buf, size_t count)
{
	TRACE_FUNCT("read");
	int ret = ::recv(m_sock, SOCKBUF buf, count, MSG_NOSIGNAL);
	TRACE_ASSERT( ret != SOCK_ERR);
	return ret;
}
ssize_t MSock::write(void *buf, size_t count)
{
	TRACE_FUNCT("write");
	int ret = ::send(m_sock, SOCKBUF buf, count, 0);
	TRACE_ASSERT( ret != SOCK_ERR);
	return ret;
}
ssize_t MSock::gets(char *buffer, size_t end)
{
	int cb=0;
	while(cb < (int) end && m_sock != EMPTY_SOCK)
	{
		char c;
		if(recv(m_sock, SOCKBUF  &c, 1, 0)==1)
		{
			*buffer++=c;
			cb++;
			if(c == 0x0A || c == 0x0D)
				break;
		}
		else
			break;

	}
	*buffer = 0;
	return (ssize_t) cb;
}
ssize_t MSock::retryread(void *buf, size_t count, int retry)
{
	TRACE_FUNCT("retryread");
	int i;
	size_t cbT;
	char *pct = (char *)buf;
	int retries = 0;

#if 0
	for(cbT=0, i=0; retries < retry ; i++)
	{
		if(ReadSelect(1))
		{
			int cb = read(pct, count);

			if(cb < 0) // Error!
				break;
			else if (cb == 0)
				retries++;
			else
				retries=0;

			cbT += cb;

			if(cbT == count)
				break;
			pct+=cb;
		}
		else
			retries++;
	}
#else
	for(cbT=0, i=0; retries < retry ; i++)
	{
		int cb = read(pct, count);

		if(cb < 0) // Error!
			break;
		else if (cb == 0)
			retries++;
		else
			retries=0;

		cbT += cb;

		if(cbT == count)
			break;
		pct+=cb;

		while((ReadSelect(1)==0) && (retries < retry))
			retries++;
	}
#endif
	return cbT;
}
ssize_t MSock::retrywrite(void *buf, size_t count, int retry)
{
	TRACE_FUNCT("retrywrite");
	int i,cbT;
	char *p = (char *)buf;
	int countT = count;
	size_t cb;

	for(countT=count,cb=0,cbT=0, i=0; (i < retry) && (cb != count); i++)
	{
		cbT = write(p, countT);

		if(cbT < 0)
		{
			// an error, get out.
			cb = (size_t)-1;
			break;
		}
		else 
		{	// positive result.

			cb += cbT;		// add it to cb;
			countT -= cbT;
			p += cbT;
			if(countT)
			{
				TRACE_STR("Waiting on rwrite");
				if(WriteSelect(1) > 0)
					i=0; // Keep retrying data is being read.
			}
		}
	}
	return cb;
}

int MSock::FdSet(fd_set *fdset)
{
	TRACE_FUNCT("FdSet");
	FD_SET(m_sock, fdset);
	return m_sock;
}

int MSock::FdClr(fd_set *fdset)
{
	TRACE_FUNCT("FdCler");
	FD_CLR(m_sock, fdset);
	return m_sock;
}

Boolean MSock::FdIsSet(fd_set *fdset)
{
	TRACE_FUNCT("FdIsSet");
	return FD_ISSET(m_sock, fdset);
}

void MSock::Close(void)
{
	if(m_sock != EMPTY_SOCK)
	{	
		SOCKCLOSE(m_sock);
		m_sock = EMPTY_SOCK;
	}
}
void MSock::Renew(int fd, struct sockaddr *psaddr)
{
	TRACE_FUNCT("Renew");
	
	m_flags = 0; // Not a template

	assert(m_sock == EMPTY_SOCK);

	if(psaddr)
		m_addr.saddr = *psaddr;
	else
		memset(&m_addr,0,sizeof(m_addr));

	m_sock 	= fd;
}

void MSock::Reset(void)
{
	// Close and recreate, reusing a socket not
	// supported on all platforms.
	if(m_sock != EMPTY_SOCK)
	{
		Linger(0);
		SOCKCLOSE(m_sock);
		m_sock = socket(m_fam, m_type, m_proto);
	}
}
MSock::MSock(int port, int fam, int type, int protocol)
{
	TRACE_FUNCT("MSock");
	memset(&m_addr, 0, sizeof(m_addr));

	m_addr.saddrin.sin_family = fam;
	m_addr.saddrin.sin_port = htons(port);

	m_sock = socket(fam, type, protocol);

	m_flags = SOCKET_TEMPLATE;
	m_type = type;
	m_proto = protocol;
	m_fam = fam;
	m_port = port;

	// Something seriously wrong if
	// can't create a socket here.
	assert(m_sock != EMPTY_SOCK);

#ifdef NON_REENTRENT_SOCKET	
	MutexInit(&m_mutex);
#endif
}

MSock::MSock(void)
{
	m_sock = EMPTY_SOCK; 	// An empty socket
	m_flags = 0; 		// Not a template
	m_type = 0;
	m_proto = SOCK_STREAM;
	m_fam = AF_INET;
	m_port=0;
	memset(&m_addr,0,sizeof(m_addr));
#ifdef NON_REENTRENT_SOCKET	
	MutexInit(&m_mutex);
#endif
}

MSock::MSock(int fd, struct sockaddr *psaddr)
{
	TRACE_FUNCT("MSock(private)");

	m_type = 0;
	m_proto = SOCK_STREAM;
	m_fam = AF_INET;

	m_flags = 0; // Not a template

	if(fd != EMPTY_SOCK) // An empty socket
	{

		if(psaddr)
			m_addr.saddr = *psaddr;
		else
			memset(&m_addr,0,sizeof(m_addr));
	}
	m_sock 	= fd;
#ifdef NON_REENTRENT_SOCKET	
	MutexInit(&m_mutex);
#endif
}
int MSock::Shutdown(int type)
{
	TRACE_FUNCT("ShutDown");
	int ret = shutdown(m_sock,type);
	TRACE_ASSERT( ret != SOCK_ERR );
	return ret;
}

MSock::~MSock()
{
	TRACE_FUNCT("~MSock");
	Close();
#ifdef NON_REENTRENT_SOCKET	
	MutexDestroy(&m_mutex);
#endif
}

Boolean MSock::Reconnect()
{
	TRACE_FUNCT("Reconnect");

	int ret = SOCK_ERR;
	if(m_sock != EMPTY_SOCK)
		SOCKCLOSE(m_sock);

	m_addr.saddrin.sin_family = m_fam;
	m_addr.saddrin.sin_port = htons(m_port);

	m_sock = socket(m_fam, m_type, m_proto);
#ifdef TRACE_DEBUG
	if(m_sock == SOCK_ERR)
		TRACE_STR("Could not create socket");
#endif

	if(m_sock != SOCK_ERR)
	{
		ret = connect(m_sock, (sockaddr *)&m_addr.saddrin, sizeof(m_addr.saddrin));
#ifdef TRACE_DEBUG
		if(ret == SOCK_ERR)
			TRACE_STR("Could not reconnect socket");
#endif
	}

	return (ret == SOCK_ERR) ? FALSE : TRUE;
}

Boolean MSock::Connect()
{
	TRACE_FUNCT("Connect");
	int ret = connect(m_sock, (sockaddr *)&m_addr.saddrin, sizeof(m_addr.saddrin));
	TRACE_ASSERT( ret != SOCK_ERR);
	return (ret == SOCK_ERR) ? FALSE : TRUE;
}

Boolean MSock::Connect(char *hostname)
{
	int ret = FALSE;
	TRACE_FUNCT("Connect(host)");
	if(SetHostByName(hostname))
	{
		int val = connect(m_sock, (sockaddr *)&m_addr.saddrin, sizeof(m_addr.saddrin));
		ret = (val == SOCK_ERR) ? FALSE : TRUE;
	}
	return ret;
}
Boolean MSock::Connect(struct hostent *hent)
{
	int ret = FALSE;
	TRACE_FUNCT("Connect(hent)");
	if(hent)
	{
		memcpy((char *)&m_addr.saddrin.sin_addr, (char *)hent->h_addr, hent->h_length);
		int val = connect(m_sock, (sockaddr *)&m_addr.saddrin, sizeof(m_addr.saddrin));
		ret = (val == SOCK_ERR) ? FALSE : TRUE;
	}
	return ret;
}

Boolean MSock::Bind()
{
	int one=1;
	TRACE_FUNCT("Bind");
	TRACE_ASSERT(m_flags & SOCKET_TEMPLATE);
	setsockopt(m_sock, SOL_SOCKET, SO_REUSEADDR, OPTCAST &one, sizeof(one));
	int ret = bind(m_sock, (sockaddr *)&m_addr.saddrin, sizeof(m_addr.saddrin));
	return (ret == SOCK_ERR) ? FALSE : TRUE;
}

Boolean MSock::Listen(int queue)
{
	TRACE_FUNCT("Listen");
	TRACE_ASSERT(m_flags & SOCKET_TEMPLATE);
	int ret = SOCK_ERR;

	if(Bind())
		ret = listen(m_sock, queue);
	
	return (ret == 0);
}

Boolean MSock::Linger(Boolean fLinger, int seconds)
{
	TRACE_FUNCT("Linger");
	int ret;

	struct linger sling;
	socklen_t len= sizeof(sling);

	sling.l_onoff = fLinger;
	sling.l_linger= (seconds == SOCKET_LINGER_ABORT) ? 0 :seconds;

	ret = setsockopt(m_sock, SOL_SOCKET, SO_LINGER, OPTCAST &sling,len);

	TRACE_ASSERT( ret != SOCK_ERR);
	return ret != SOCK_ERR;
}

Boolean MSock::SetSockOpt(int level, int opt, void *val, socklen_t cb)
{
	TRACE_FUNCT("SetSockOpts");
	int ret = setsockopt(m_sock,level,opt,OPTCAST val,cb);
	TRACE_ASSERT( ret != SOCK_ERR );
	return (ret == 0) ? TRUE : FALSE;
}

Boolean MSock::GetSockOpt(int level, int opt, void *val, socklen_t *cb)
{
	TRACE_FUNCT("GetSockOpts");
	int ret = getsockopt(m_sock,level,opt,OPTCAST val,cb);
	TRACE_ASSERT( ret != SOCK_ERR );
	return (ret == 0) ? TRUE : FALSE;
}


Boolean MSock::Poll(int events, int *revents, int to)
{
#ifdef POSIX
#ifndef DARWIN
	TRACE_FUNCT("Poll");
	pollfd pofd;

	pofd.fd 	= m_sock;
	pofd.events 	= events;
	pofd.revents 	= 0;

	int ret = poll(&pofd, 1, to);
	
	TRACE_ASSERT( ret != SOCK_ERR );

	if(ret > 0)
	{
		if(revents)
			*revents = pofd.revents;
		return TRUE;
	}
#endif
#endif
	return FALSE;
}
Boolean MSock::TcpNoDelay(Boolean fOn)
{
	TRACE_FUNCT("TcpNoDelay");
	int var= fOn;

#ifdef SOL_TCP
	int ret = setsockopt(m_sock,SOL_TCP,TCP_NODELAY, OPTCAST &var, sizeof(var));
#else
	struct protoent *pr_ent;
	pr_ent = getprotobyname("TCP");

	if(!pr_ent)
		return FALSE;
	int ret = setsockopt(m_sock,pr_ent->p_proto,TCP_NODELAY, OPTCAST &var, sizeof(var));
#endif
	TRACE_ASSERT( ret != SOCK_ERR );
	 
	return (ret == 0) ? TRUE : FALSE;
}
char *MSock::IpAddr(void)
{
	return inet_ntoa(((struct sockaddr_in *)&m_addr.saddr)->sin_addr);
}

MSock *MSock::Accept(void)
{
	TRACE_FUNCT("Accept");
	int fd;
	struct  sockaddr_in  addr;
	socklen_t addrlen = sizeof(addr);
#ifdef NON_REENTRENT_SOCKET	
	MutexLock(&m_mutex);
#endif
	fd = accept(m_sock, (sockaddr *)&addr, &addrlen);

#ifdef NON_REENTRENT_SOCKET	
	MutexUnlock(&m_mutex);
#endif

	if(fd != EMPTY_SOCK)
	{
		if(s_count_hba && !checkhba(addr.sin_addr.s_addr))
			SOCKCLOSE(fd);
		else
			return new MSock(fd, (sockaddr *)&addr);
	}
	mferr.mfprintf("Accept failed\n");
	return NULL;
}
Boolean MSock::Accept(MSock *ms)
{
	TRACE_FUNCT("Accept ms");
	int fd;
	struct  sockaddr_in  addr;
	socklen_t addrlen = sizeof(addr);

	ms->m_sock = 0;

#ifdef NON_REENTRENT_SOCKET	
	MutexLock(&m_mutex);
#endif
	fd = accept(m_sock, (sockaddr *)&addr, &addrlen);
#ifdef NON_REENTRENT_SOCKET	
	MutexUnlock(&m_mutex);
#endif
	if(fd != EMPTY_SOCK)
	{ 
                if(s_count_hba && !checkhba(addr.sin_addr.s_addr))
                        SOCKCLOSE(fd);
                else
		{
			m_flags = 0; // Not a template
			memcpy((void *)&ms->m_addr.saddr, (void *)&addr, addrlen);
			ms->m_sock = fd;
			return TRUE;
		}
	}
	return FALSE;
}

Boolean	MSock::SetHostEnt(struct hostent *hent)
{
	TRACE_FUNCT("SetHostEnt");
	
	if(hent)
	{
		memcpy((char *)&m_addr.saddrin.sin_addr, (char *)hent->h_addr, hent->h_length);
		return TRUE;
	}
	else 
		return FALSE;
	
}
Boolean MSock::SetHostByName(char *hname, struct hostent *hentRet)
{

	TRACE_FUNCT("SetHostByName");
	struct hostent *hentT;

	if(hname)
	{
#ifdef WIN32
		IPV4	ipv4;
		if(FillIPV4(&ipv4, hname))
		{
			unsigned long ipaddr = inet_addr(hname);
			memcpy((char *)&m_addr.saddrin.sin_addr, (char *)&ipaddr, 4);
			return TRUE;
		}
		else
		{
			hentT = gethostbyname(hname);
			return SetHostEnt(hentT);
		}
#else
		hentT = gethostbyname(hname);
		if(hentT)
		{
			if(hentRet)
				*hentRet = *hentT;
			return SetHostEnt(hentT);
		}
#endif
	}
	return FALSE;
}

#ifndef WIN32
#include <sys/utsname.h>

char **GetLocalNetAddr(int *type)
{
	struct hostent *hptr;
	struct utsname name;

	if(uname(&name) < 0)
		return NULL;

	hptr = gethostbyname(name.nodename);

	if(!hptr)
		return NULL;

	if(type)
		*type = hptr->h_addrtype;
	return hptr->h_addr_list;
}

unsigned int GetLocalIPV4(void)
{
	IPV4 ip;

	ip.raw = 0;

	unsigned char ** host_list = (unsigned char **)GetLocalNetAddr(NULL);

	if(host_list)
	{
		unsigned char *raw = host_list[0];
		for(int i=0; i < 4; i++)
			ip.p[i] = raw[i];
	}
	return ip.raw;
}
#endif
int MSock::ReadSelect(int seconds, int useconds)
{	if(m_flags & SOCKET_SELECT_READ)
	{ 	m_flags &= ~SOCKET_SELECT_READ;
		return 1;
	}
	else if(m_sock >= 0)
	{
		timeval	to; to.tv_sec = seconds; to.tv_usec = useconds;
		fd_set 	fdset;
		FD_ZERO(&fdset);
		FD_SET(m_sock,&fdset);
		return select(m_sock+1, &fdset, NULL, &fdset, &to);
	}
	else 
		return 0;
}	
int MSock::WriteSelect(int seconds, int useconds)
{	
	if(m_flags & SOCKET_SELECT_WRITE)
	{ 	m_flags &= ~SOCKET_SELECT_WRITE;
		return 1;
	}
	else if(m_sock >= 0)
	{
		timeval	to; to.tv_sec = seconds; to.tv_usec = useconds;
		fd_set 	fdset;
		FD_ZERO(&fdset);
		FD_SET(m_sock,&fdset);
		return select(m_sock+1, NULL, &fdset, &fdset, &to);
	}
	else 
		return 0;
}
