/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to commercially license this library, the author
    can be reached at info@mohawksoft.com
*/ 
// #define TRACE_DEBUG
#include "Phoenix.h"

SQL_TRANSLATOR MSql::SetTranslator(SQL_TRANSLATOR fn)
{
	SQL_TRANSLATOR fnOld = m_translate;
	m_translate = fn;
	return fnOld;
}

Boolean MSql::ErrorToBool(int err)
{
	return ((err >= SQLERR_OK) && (err <= SQLERR_OK_NONE));
}

int MSql::ColumnType(SQLRESULT sqlres, int ndx)
{
	return SQLTYPE_UNDEFINED;
}
void MSql::SetQid(int qid)
{
	m_qid = qid;
}
int MSql::GetQid(void)
{
	return m_qid;
}
SQLRESULT MSql::ExecfResult(char *fmt, ...)
{
	va_list arglist;
	char buffer[MAX_STRING];
	va_start(arglist, fmt);
	VSNPRINTF(buffer, sizeof(buffer), fmt, arglist);
	va_end(arglist);
	return ExecResult(buffer);
}

int MSql::Execf(char *fmt, ...)
{
	va_list arglist;
	char buffer[MAX_STRING];
	va_start(arglist, fmt);
	VSNPRINTF(buffer, sizeof(buffer), fmt, arglist);
	va_end(arglist);
	return Exec(buffer);
}

Boolean MSql::ExecInt(char *sql, int *ret)
{
	Boolean fResult = FALSE;

	SQLRESULT res = ExecResult(sql);

	if(res)
	{
		char *val = ColumnValue(res,0,0);

		if(val)
		{
			*ret = atoi(val);
			fResult = TRUE;
		}
		FreeResult(res);
	}
	return fResult;
}

// Most SQL do not need an explicit begin
Boolean MSql::Begin(void)
{
	return TRUE;
}

Boolean MSql::Commit()
{
	return ErrorToBool( Exec("COMMIT") );
}

Boolean MSql::Abort()
{
	return ErrorToBool( Exec("ABORT") );
}
void MSql::CursorFreeResult(SQLRESULT sqlres)
{
	// Default, do nothing
}
void MSql::CursorDestroy(char *curs)
{
	// Default, do nothing
}
Boolean	MSql::CursorCreate(char *name, char *sql, Boolean fBinary)
{
	char szBCurs[]="declare %s binary cursor for %s";
	char szCurs[]= "declare %s cursor for %s";

	char *sqls = (fBinary) ? szBCurs : szCurs;

	return ErrorToBool( Execf(sqls, name, sql) );
}
MSql::MSql()
{
	m_lifetime=-1; // Live for ever
}
MSql::~MSql()
{
}
int MSql::Lifetime(void)
{
	return m_lifetime;
}

MSql *MSql::Create(char *conninfo)
{
	char datasource[64];
	char host[64];
	char name[64];
	char auth[64];
	char db[64];
	char port[64];
	MSql *sql=NULL;
	
	if(!conninfo)
		return NULL;

	if(!getparam(datasource, "DS:", conninfo))
	{
		mferr.mfprintf("SQL Connect: No datasource provided! (%s)\n", conninfo);
		return NULL;
	}

	getparam(name, "NAME:", conninfo);

	getparam(auth, "AUTH:", conninfo);

	if(!getparam(db, "DB:", conninfo))
		strcpy(db, "ODBC");

	if(!getparam(host, "HOST:", conninfo))
		strcpy(host, "localhost");

	if(!getparam(port, "PORT:", conninfo))
		strcpy(port, "5432");

	// This is sort of bogus, it should be completely dynamic
	// but, right now, it works as well as I need it.

#ifdef POSTGRESQL
	if(!sql && (strcasecmp(db, "PGSQL")==0))
	{
#ifdef TRACE_DEBUG
		mferr.mfprintf("PGSQL Connect: %s, %s, %s: ", datasource, host, port);
#endif
		sql = NewMPGSql(host, datasource, port);
	}
#endif
#ifdef ODBC
	if(!sql && (strcasecmp(db, "ODBC")==0))
	{
#ifdef TRACE_DEBUG
		mferr.mfprintf("ODBC Connect: %s, %s, %s: ", datasource, name, auth);
#endif
		sql = NewMODBCSql(datasource, name, auth);
	}
#endif
#ifdef TRACE_DEBUG
	if(!sql)
		mferr.mfprintf("Failed\n");
	else
		mferr.mfprintf("Success\n");
#endif

	return sql;
}


MSqlQueue::MSqlQueue()
{
	m_conninfo=0;
	m_num=0;
	m_count=0;
	m_entries=NULL;
}
/* Destroys a queue of SQL connections */
MSqlQueue::~MSqlQueue(void)
{
	for(int i=0; i < m_num; i++)
	{
		if(m_entries[i].sql)
			delete m_entries[i].sql;
	}
	phmfree(m_conninfo);
}
/* initializes a queue of SQL connections */
Boolean MSqlQueue::Init(char *coninfo, int nentries)
{
	int count=0;
	int cb = sizeof(QENTRY) * nentries;

	m_entries = (QENTRY *) phmalloc(cb);

	assert(m_entries);

	m_num = nentries;
	m_count = 0;
	m_conninfo = (char *)phmstrdup(coninfo);

	for(int i=0; i < nentries; i++)
	{
		MSql *conn = MSql::Create(coninfo);

		if(conn)
		{
			conn->SetQid(i);
			m_entries[i].sql = conn;
			m_entries[i].fInuse = FALSE;
			m_entries[i].lifetime = conn->Lifetime();
			count++;
		}
		else
		{
			m_entries[i].sql = NULL;
			m_entries[i].fInuse = TRUE; // Don't use it!!!
		}
	}
	m_count = count;
	return (count) ? TRUE : FALSE;
}


/* Gets first free entry in a queue list */
MSql *MSqlQueue::Get(void)
{
	int qid=-1;
	MSql *conn=NULL;

	if(!m_count)
		return NULL;

	while(!conn)
	{
		m_mutex.lock();

		for(int i=0; i < m_num; i++)
		{
			if(m_entries[i].fInuse == FALSE)
			{
				conn = m_entries[i].sql;
				m_entries[i].fInuse = TRUE;
				qid = i;
				break;
			}
		}
		m_mutex.unlock();

		if(!conn)
			YieldTimeSlice();
	}

	if(m_entries[qid].lifetime != -1)
	{
		m_entries[qid].lifetime--;

		if(m_entries[qid].lifetime <= 0)
		{
			delete conn;
			conn = MSql::Create(m_conninfo);
			assert(conn);
			m_entries[qid].sql = conn;
			m_entries[qid].lifetime = conn->Lifetime();;
		}
	}
	return conn;
}

/* releases a queue entry from use */
void MSqlQueue::Release(MSql *conn)
{
	int qid = conn->GetQid();

	assert(m_entries[qid].sql == conn);
	m_entries[qid].fInuse = FALSE;
}

int MSqlQueue::GetCount(void)
{
	return m_count;
}
