/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to commercially license this library, the author
    can be reached at info@mohawksoft.com
*/ 
// #define TRACE_DEBUG

#include "Phoenix.h"
#define THREAD(x)	((MThread*)(x))
#include <signal.h>
#include <limits.h>

// All threads use this thunk.
THUNK_LANG THREAD_RET THUNK_DECL MThread::thunk(void *th)
{
	THREAD_RETVAL;

	THREAD(th)->m_status = THSTAT_STARTING;
	
	if(THREAD(th)->threadStart())
	{
		try
		{
			THREAD(th)->m_status = THSTAT_RUNNING;
        		THREAD(th)->threadLoop();
			THREAD(th)->m_status = THSTAT_EXITING;
			THREAD_SET_RETVAL THREAD(th)->threadEnd(FALSE);
			THREAD(th)->m_exit = THEXIT_NORMAL;
		}
		catch (MThread *mth)
		{
			THREAD(th)->m_status = THSTAT_EXITING;
			THREAD_SET_RETVAL THREAD(th)->threadEnd(TRUE);
			THREAD(th)->m_exit = THEXIT_EXCEPT;
		}
		catch (...)	// Catch everything?
		{
			THREAD(th)->m_status = THSTAT_EXITING;
			THREAD_SET_RETVAL THREAD(th)->threadEnd(TRUE);
			THREAD(th)->m_exit = THEXIT_ERROR;
		}
	}

	THREAD(th)->m_status = THSTAT_DEFUNCT;

	if(THREAD(th)->m_mgr)
		THREAD(th)->m_mgr->death(THREAD(th)->m_fDelOnExit, THREAD(th)->m_mgrid);

	if(THREAD(th)->m_fDelOnExit)
	{
		delete THREAD(th);
	}
	THREAD_RETURN;
}

Boolean MThread::threadStart(void)
{
	return TRUE;
}

THREAD_RET MThread::threadEnd(Boolean fException)
{
	if(fException)
		mferr.mfprintf("Exeption in thread %ld\n", m_pthread_t);

	THREAD_RETURN_VAL(NULL);
}

MThread::MThread(size_t stacksize)
{
	m_fQuit = FALSE;
	m_status = THSTAT_PRESTART;
	m_mgr = NULL;
#ifdef POSIX
	pthread_attr_init(&m_thread_attr);
#endif
	SetStackSize(stacksize);
}

void MThread::SetStackSize(int stacksize)
{
#ifdef POSIX
#ifndef CYGWIN
// This don't work on Cygwin!
	if(PTHREAD_STACK_MIN < stacksize)
		pthread_attr_setstacksize(&m_thread_attr, stacksize);
#endif
#endif
#ifdef WIN32
	m_stacksize = stacksize;
#endif
}

MThread::~MThread()
{
#ifdef POSIX
	pthread_attr_destroy(&m_thread_attr);
#endif
}

int MThread::run(Boolean fdetach)
{
#ifdef POSIX
	int err;
	m_status = THSTAT_PRESTART;
	m_fDelOnExit = fdetach;

        err = pthread_create(&m_pthread_t, &m_thread_attr,&thunk,(void *)this);

	assert( err == 0);

	if(fdetach)
		pthread_detach(m_pthread_t);
        return err; 
#endif
#ifdef WIN32
	m_pthread_t = _beginthread(&thunk, m_stacksize, (void *) this); 
	return (m_pthread_t > 0) ? 0 : errno;
#endif
}
void MThread::SetMgr(MThreadMgrCore *mgr, int mgrid)
{
	m_mgr = mgr;
	m_mgrid = mgrid;
}

void MThread::detach(void)
{
	m_fDelOnExit = TRUE;
#ifdef POSIX
	pthread_detach(m_pthread_t);
#endif
}
void MThread::cancel(void)
{
	m_fQuit = TRUE;
#ifdef POSIX
	pthread_cancel(m_pthread_t);
#endif
}
void MThread::exit(void *value)
{
#ifdef POSIX
	pthread_exit(value);
#endif
#ifdef WIN32
	_endthread();
#endif
}

MThreadMgrCore::MThreadMgrCore()
{
}
MThreadMgrCore::~MThreadMgrCore()
{
}
void MThreadMgrCore::death(Boolean fdeleteme, int mgrid)
{
	m_death.Inc();
	m_numThreads.Dec();

	if(fdeleteme) // Thread deletes itself
		m_threads[mgrid] = NULL;
}

int MThreadMgrCore::Interp(MThread *mth)
{
	return TINTERP_NEW;
}

Boolean MThreadMgrCore::Pulse(void)
{
	return TRUE;
}

Boolean MThreadMgrCore::Poll(void)
{
	Boolean f = Pulse();
	if(f)
		sleep(1);
	return f;
}

int MThreadMgrCore::RunThread(MThread *mth, int mgrid)
{
	m_numThreads.Inc();
	mth->SetMgr(this, mgrid);
	mth->run(FALSE);
	mth->detach();
	return mgrid;
}

Boolean MThreadMgrCore::Prepare(void)
{
	return TRUE;
}

MThreadMgr::MThreadMgr()
{
}

int MThreadMgrCore::AddThread(MThread *mth)
{
	int count = m_threads.count(); // see how many slots
	for(int n=0; n < count; n++)
	{
		if(m_threads[n] == NULL)
		{
			m_threads[n] = mth;
			return RunThread(mth, n);
		}
	}
	return RunThread(mth, m_threads.AddNdx(mth));
}
void MThreadMgr::RunThreads(int n)
{
	// Run Threads
	for(int i=0; i < n ; i++)
	{
		MThread * mth = NewThread();
		assert(mth);
		AddThread(mth);
	}
}
int MThreadMgr::count()
{
	return m_numThreads.Get();
}
void MThreadMgr::Run(int n)
{
	TRACE_FUNCT("MThreadMgr::Run");
	int i;
	int count;
	m_fContinue=TRUE;

	m_died = 0;

	if(!Prepare())
		return;

	m_threads.trunk(0);

	// Seed server
	RunThreads(n);

	while(m_fContinue)
	{
		while(m_death.Get() < 1 && m_fContinue)
		{
			if(!Poll())
				m_fContinue = FALSE;
		}

		if(!m_fContinue)
			break;

		count = m_threads.count();

		for(i=0; i < count; i++)
		{
			if(m_threads[i] == NULL)
				m_death.Dec();
			
			else if(m_threads[i]->m_status == THSTAT_DEFUNCT)
			{
				m_death.Dec();

				int interp = Interp(m_threads[i]);

				m_threads[i]->m_status = THSTAT_PRESTART;

				if(interp == TINTERP_QUIT)
				{
					count = 0;
					m_fContinue = FALSE;
					break;
				}
				else if(interp == TINTERP_RESTART)
				{
					RunThread(m_threads[i], i);
				}
				else if(interp == TINTERP_NEW)
				{
					delete m_threads[i];
					m_threads[i] = NewThread();
					if(m_threads[i])
						RunThread(m_threads[i], i);
				}
				else
				{
					count =0;
					m_fContinue = FALSE;
				}
			}
		}
	}

	Poll(); // Do one last poll!!!

	TRACE_STR("Out of Run loop");
	// Time to shut down the loop
	count = m_threads.count();

	for(i=0; i < count; i++)
	{
		if(m_threads[i])
		{
			m_threads[i]->m_fQuit=TRUE;
			m_threads[i]->m_fDelOnExit = TRUE;
		}
	}
	TRACE_STR("Out of Thread kill loop");

	// Retry for up to 5 seconds, wait for threads to exit
	for(int retry=0; retry < 20; retry++) 
	{
		int n = 0;
		for(i=0; i < count; i++)
		{
			if(m_threads[i])
				n++;
		}

		if(n)
			usleep(250);
		else
			break;

	};
	TRACE_STR("Out of Thread wait loop");

	for(i=0; i < count; i++)
	{
		if(m_threads[i])
			m_threads[i]->cancel();
	}
	TRACE_STR("Out of Thread cancel loop");
}

#include <signal.h>
#include <sys/types.h>
#ifdef POSIX
#include <sys/wait.h>  
#endif
int     	MThreadMgrCore::m_sigval = 0;
MThreadMgrCore *	MThreadMgrCore::m_root = NULL;
 
#ifdef POSIX
void MThreadMgrCore::do_sighup(int num)
{
        m_sigval = SIGHUP;
}
 
void MThreadMgrCore::do_sigchld(int num)
{
        m_sigval = SIGCHLD;
 
}
#endif
void MThreadMgrCore::do_sigterm(int num)
{
        m_sigval = SIGTERM;
}   

#ifdef HAS_FORK
Boolean MThreadMgrCore::FTPrepare(void)
{
	return TRUE;
}

void MThreadMgr::FTRun(int n)
{
	time_t start;
        Boolean fStartServers = TRUE;
	m_fContinue = TRUE;
 
	start = time(0);

        for(m_sigval=0, m_fContinue=TRUE; m_fContinue;)
        {
		int status;
		pid_t pidServer=0;

                // Create a child process to serve stuff because dead threads
                // kill parents.
                m_sigval = 0;

#ifdef POSIX
        	signal(SIGHUP,  SIG_DFL);
        	signal(SIGCHLD, SIG_DFL);
#endif
        	signal(SIGTERM, SIG_DFL);
 
                if(fStartServers)
                {
			start = time(0);

                        fStartServers = FALSE;

                        pidServer = fork();

                        if(pidServer == 0)
                        {
				if(FTPrepare())
                                	Run(n);

                                exit(0);
                        }
			else
				FTStart(pidServer);
                }

#ifdef POSIX
        	signal(SIGHUP,  MThreadMgr::do_sighup);
        	signal(SIGCHLD, MThreadMgr::do_sigchld);
#endif
        	signal(SIGTERM, MThreadMgr::do_sigterm);
 
                while(!m_sigval && m_fContinue)
			m_fContinue = Poll();
 
                switch(m_sigval)
                {
#ifdef POSIX
                        case SIGHUP:
                        {
                                mferr.mfprintf("Warning:SIGHUP in Master\n");
                                kill(pidServer, SIGTERM);
				/* Wait for SIGCHILD to restart servers */
                                fStartServers = FALSE;
                                break;
                        }
 
                        case SIGCHLD:
                        {
                                mferr.mfprintf("Warning:SIGCHLD in Master\n");
				waitpid(pidServer, &status, WNOHANG);
                                fStartServers = TRUE;
                                break;
                        }
#endif
                        case SIGTERM:
                        {
                                mferr.mfprintf("Warning:SIGTERM in Master\n");
                                kill(pidServer, SIGTERM);
                                m_fContinue = FALSE;
				break;
                        }
 
                        default:
                                mferr.mfprintf("Unknown signal (%d) in signal handler\n", m_sigval);
                                break;
                }
		if((time(0) - start)< 2)
			break;
        }                                                                                                   
}
void MThreadMgr::FTStart(pid_t pid)
{
}
#endif
