/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to commercially license this library, the author
    can be reached at info@mohawksoft.com
*/ 

#include "Phoenix.h"

#define height(n)	((n) ? m_nodes[n].height : 0)
#define maxheight(r, n1,n2)\
	{ 	int h1 = height(n1);\
		int h2 = height(n2);\
		r = MAX(h1,h2);	\
	} r

int MIntTree::SingleRotateLeft( int ndx )
{
	MINT * cn2 	= m_nodes.GetPtr(ndx);
	MINT * cn1 	= m_nodes.GetPtr(cn2->prev);
	int	ret	= cn2->prev;
	cn2->prev 	= cn1->next;	
	cn1->next 	= ndx;
	maxheight(cn2->height,cn2->prev, cn2->next)+=1;
	maxheight(cn1->height,cn1->prev, ret)+=1;
	return ret; 
}
int MIntTree::SingleRotateRight( int ndx )
{
	MINT * cn1	= m_nodes.GetPtr(ndx);
	MINT * cn2 	= m_nodes.GetPtr(cn1->next);
	int	ret	= cn1->next;
	cn1->next 	= cn2->prev;
	cn2->prev 	= ndx;
	maxheight(cn1->height,cn1->prev, cn1->next)+=1;
	maxheight(cn2->height,cn2->next, ret)+=1;
	return ret; 
}
int MIntTree::DoubleRotateLeft( int ndx )
{
	MINT *cn3 = m_nodes.GetPtr(ndx);
	cn3->prev = SingleRotateRight(cn3->prev);
	return SingleRotateLeft(ndx);
}
int MIntTree::DoubleRotateRight( int ndx )
{
	MINT *cn1 = m_nodes.GetPtr(ndx);
	cn1->next = SingleRotateLeft(cn1->next);
	return SingleRotateRight(ndx);
}
int MIntTree::Insert(int ndx, MINT *insert)
{
	if( ndx == -1 )
	{
		ndx = insert->height; // height carries index on insert!
		insert->height=0;
		maxheight(insert->height, insert->prev, insert->next)+=1;
		return ndx;
	}
	else
	{
		MINT *cnT = m_nodes.GetPtr(ndx);

		int cmpresult = insert->key - cnT->key;

		if(cmpresult < 0)
		{
			cnT->prev = Insert(cnT->prev, insert);
			if(height(cnT->prev) - height(cnT->next) == 2)
			{
				if((insert->key - m_nodes[cnT->prev].key) < 0)
                        		ndx = SingleRotateLeft( ndx );
                    		else
                        		ndx = DoubleRotateLeft( ndx );
			}
		}
		else if(cmpresult > 0)
		{
			cnT->next = Insert(cnT->next, insert);
			if(height(cnT->next) - height(cnT->prev) == 2)
			{
                    		if((insert->key - m_nodes[cnT->next].key) > 0 )
                        		ndx = SingleRotateRight( ndx );
                    		else
                        		ndx = DoubleRotateRight( ndx );
			}
		}
		else 
			m_state = FALSE;
		maxheight(cnT->height, cnT->prev, cnT->next)+=1;
		return ndx;
	}

}


Boolean MIntTree::TreeWalkUp(int ndx, MINTWalk fn, void *p1)
{
	Boolean retval=TRUE;
	MINT *mn = m_nodes.GetPtr(ndx);
	if(mn->prev)
		retval = TreeWalkUp(mn->prev,fn,p1);
	if(retval)
		retval = (*fn)(mn,p1);
	if(retval && mn->next)
		retval = TreeWalkUp(mn->next,fn,p1);
	return retval;
}

Boolean MIntTree::TreeWalkDn(int ndx, MINTWalk fn, void *p1)
{
	Boolean retval=TRUE;
	MINT *mn = m_nodes.GetPtr(ndx);
	if(mn->next)
		retval = TreeWalkDn(mn->next,fn,p1);
	if(retval)
		retval = (*fn)(mn,p1);
	if(retval && mn->prev)
		retval = TreeWalkDn(mn->prev,fn,p1);
	return retval;
}

void MIntTree::Enum(MINTWalk fn, void *p1, Boolean ascending)
{
	if(m_top && fn)
	{
		if(ascending)
			TreeWalkUp(m_top,fn,p1);
		else
			TreeWalkDn(m_top,fn,p1);
	}
}
int & MIntTree::operator [] (int key)
{
	int res;
	MINT *node = m_nodes.GetPtr(m_top);
	while(node)
	{
		res = key - (int)node->key;
		if(!res)
			return node->value;

		node = (res < 0) ?  
			(MINT *) m_nodes.GetPtr(node->prev) : 
			(MINT *) m_nodes.GetPtr(node->next);
	}

	MINT mnt = {-1,-1,0,key,0};
	int ndx;
	node = m_nodes.AddNdxPtr(mnt, &ndx);
	node->height = ndx;
	m_top = Insert(m_top, node);
	return node->value;
}
