/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to commercially license this library, the author
    can be reached at info@mohawksoft.com
*/ 
#include "Phoenix.h"

#include "strtbl.h"
#include "misalpha.h"
#include "metatbl.h"

#ifndef INLINE_STROPS
#define STRDECL
#include "mstrops.h"
#endif

unsigned char _modec_array[]=
{	// Mohawk Decimal! Base 36
		'0', // 0
		'1', // 1
		'2', // 2
		'3', // 3
		'4', // 4
		'5', // 5
		'6', // 6
		'7', // 7
		'8', // 8
		'9', // 9
		'a', // 10
		'b', // 11
		'c', // 12
		'd', // 13
		'e', // 14
		'f', // 15
		'g', // 16
		'h', // 17
		'i', // 18
		'j', // 19
		'k', // 20
		'l', // 21
		'm', // 22
		'n', // 23
		'o', // 24
		'p', // 25
		'q', // 26
		'r', // 27
		's', // 28
		't', // 29
		'u', // 30
		'v', // 31
		'w', // 32
		'x', //	33
		'y', // 34
		'z' // 35
};

unsigned char _modec_array_u[]=
{	// Mohawk Decimal! Base 36
		'0', // 0
		'1', // 1
		'2', // 2
		'3', // 3
		'4', // 4
		'5', // 5
		'6', // 6
		'7', // 7
		'8', // 8
		'9', // 9
		'A', // 10
		'B', // 11
		'C', // 12
		'D', // 13
		'E', // 14
		'F', // 15
		'G', // 16
		'H', // 17
		'I', // 18
		'J', // 19
		'K', // 20
		'L', // 21
		'M', // 22
		'N', // 23
		'O', // 24
		'P', // 25
		'Q', // 26
		'R', // 27
		'S', // 28
		'T', // 29
		'U', // 30
		'V', // 31
		'W', // 32
		'X', //	33
		'Y', // 34
		'Z' // 35
};


// Takes hostname:port and returns port.
// copies hostname porton to dest buffer
int HostPort(char *dest, char *src, int defport)
{
	int cb=0;
	char *port = strchr(src, ':');

	if(port)
	{
		cb = port++ - src;
		defport = atoi(port);
	}
	else
		cb = strlen(src);

	if(dest)
	{
		memcpy(dest, src, cb);
		dest[cb]=0;
	}
	return defport;
}

void TimeRandString(char *buffer, int size)
{
	int b=0;
	int r=(int)time(0);
	int chunk=0;

	// Encode time for uniqueness over time
	while(size && r)
	{
		chunk = r & 0x1F;
		*buffer++ = (char) _modec_array[chunk];
		r >>= 5; 
		size--;
	}
	while(size)
	{
		// I don't like this, numerically.
		if(!r)
		{
			r = rand(); // get more bits
			b = r & 3;
			r >>= 7; // Skip lowest bits, not too random
		}

		chunk = r & 0x1F;

		*buffer++ = (char) _modec_array[chunk+b];
		r >>= 5; 
		size--;
	}
	*buffer = 0;
}
void RandString(char *buffer, int size)
{
	int b=0;
	int r=0;
	int chunk=0;

	while(size)
	{
		// I don't like this, numerically.
		if(!r)
		{
			r = rand(); // get more bits
			b = r & 3;
			r >>= 7; // Skip lowest bits, not too random
		}

		chunk = r & 0x1F;

		*buffer++ = (char) _modec_array[chunk+b];
		r >>= 5; 
		size--;
	}
	*buffer = 0;
}

typedef Boolean (*TOKENFN)(char *token, void *param1, void *param2);


static Boolean fmtoken(char *pstart, char *pend, TOKENFN fn, void *p1, void *p2)
{
	unsigned int n;
	char word[MAX_STRING];

	n = pend - pstart;
	if(n && (n < sizeof(word)))
	{
		memcpy(word, pstart, n);
		word[n]=0;
		return (*fn)(word,p1,p2);
	}
	return TRUE;
}

void Tokenize(char *buffer, TOKENFN fn, void *p1, void *p2, Boolean quote)
{
	char *pstart;
        char *pend;
 
	pstart = buffer;
	while(*pstart)
	{

		if(quote)
			while(*pstart && !isalnum(*pstart) && *pstart != '\"')
				pstart++;
		else
			while(*pstart && !isalnum(*pstart))
				pstart++;
		if(!*pstart)
			break;

		if((*pstart != '\"') || !quote)
		{
			pend = pstart;
			while(isalnum(*pend))
				pend++;

			if(!fmtoken(pstart,pend,fn,p1,p2))
				break;

			pstart = pend;
		}
		else
		{
			pstart++;
			pend = strchr(pstart, '\"');

			if(pend)
			{
				if(!fmtoken(pstart,pend,fn,p1,p2))
					break;
				pstart = ++pend;
			}
		}
	}                                                                                     	
}

void trace_exit(char *file, int d, char * msg)
{
	mferr.mfprintf("%s,%d (%s)\n", file,d,msg);
	mferr.flush();
	exit(-1);
}

char *snprintf(char *buffer, size_t cb, char *fmt, ...)
{
	va_list arglist;
	va_start(arglist, fmt);
	VSNPRINTF (buffer, cb, fmt, arglist);
	va_end(arglist);
	return buffer;
}

unsigned int hextoi(char *s)
{
	unsigned int t = 0;
	unsigned int u = 0;

	while(isxdigit(*s))
	{
		u = u << 4;

		if((*s >= '0') && (*s <= '9'))
		{
			t = *s - '0';
		}
		else if((*s >= 'A') && (*s <= 'F'))
		{
			t = (*s - 'A') + 10;
		}
		else if((*s >= 'a') && (*s <= 'f'))
		{
			t = (*s - 'a') + 10;
		}
		
		u += t;
		s++;
	}	
	return u;
}
unsigned int hextoin(char *s, int n)
{
	unsigned int t = 0;
	unsigned int u = 0;

	for(int i=0; i < n && isxdigit(*s); i++)
	{
		u = u << 4;

		if((*s >= '0') && (*s <= '9'))
		{
			t = *s - '0';
		}
		else if((*s >= 'A') && (*s <= 'F'))
		{
			t = (*s - 'A') + 10;
		}
		else if((*s >= 'a') && (*s <= 'f'))
		{
			t = (*s - 'a') + 10;
		}
		
		u += t;
		s++;
	}	
	return u;
}

unsigned int modectoi(char *s)
{
	unsigned int t = 0;
	unsigned int u = 0;

	while(isalnum(*s))
	{
		u = u * sizeof(_modec_array);

		if((*s >= '0') && (*s <= '9'))
		{
			t = *s - '0';
		}
		else if((*s >= 'A') && (*s <= 'Z'))
		{
			t = (*s - 'A') + 10;
		}
		else if((*s >= 'a') && (*s <= 'z'))
		{
			t = (*s - 'a') + 10;
		}
		
		u += t;
		s++;
	}	
	return u;
}
unsigned int modectoin(char *s, int n)
{
	unsigned int t = 0;
	unsigned int u = 0;

	while(isalnum(*s) && n--)
	{
		u = u * sizeof(_modec_array);

		if((*s >= '0') && (*s <= '9'))
		{
			t = *s - '0';
		}
		else if((*s >= 'A') && (*s <= 'Z'))
		{
			t = (*s - 'A') + 10;
		}
		else if((*s >= 'a') && (*s <= 'z'))
		{
			t = (*s - 'a') + 10;
		}
		
		u += t;
		s++;
	}	
	return u;
}

char modec(int n)
{
	return ((n>=0) && (n<(int)sizeof(_modec_array))) ? _modec_array[n] : 0;
}

char *itomodec(int val, char *p, Boolean is_signed)
{
	char stackbuf[64];
	char *psave = p;
	register char *stack = stackbuf;

	if (is_signed && (val < 0))
	{
		*p++ = '-';
		val = (unsigned long)(-(long)val);
	}

	do
	{
		register unsigned n = (unsigned) (val % sizeof(_modec_array));
		val /= sizeof(_modec_array);
		*stack++ = _modec_array[n];

	} while (val);

		/* Unwind stack into ascii string. */

	while(stack > stackbuf)
		*p++ = *--stack;

	*p=0;
	return psave;
}

// Returns milliseconds.
unsigned long GetMilliTicks(void)
{
	#ifdef WIN32
	return (unsigned long) GetTickCount();
	#else
	unsigned long tick_sec =
#ifdef _SC_CLK_TCK
	sysconf(_SC_CLK_TCK);
#else
#warning Using CLK_TCK for GetMilliTicks
	CLK_TCK;
#endif
	struct tms t;
	clock_t ct = times(&t);

	unsigned long ret = (((unsigned long) ct)*1000)/tick_sec;
	return (unsigned long) ret;
	#endif
}
#ifndef YIELD_INLINE
void YieldTimeSlice(void)
{
#ifdef POSIX
#ifdef _POSIX_PRIORITY_SCHEDULING
	sched_yield();
#else
#warning using usleep for yield
	usleep(1000);
#endif
#endif
#ifdef WIN32
	Yield();
#endif
}
#endif

unsigned long GetDiffTicks(unsigned long start_ticks)
{
	unsigned long now = GetMilliTicks();

	return now - start_ticks;
}

#ifdef WIN32
void Win32_printerr(char *title, Boolean fPopup)
{
	LPVOID lpMsgBuf;

	FormatMessage(
		FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
		NULL,
		GetLastError(),
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
		(LPTSTR) &lpMsgBuf,
		0,
		NULL);

	if(fPopup)
	{
		MessageBox( NULL, (char *) lpMsgBuf, title, MB_OK|MB_ICONINFORMATION );
	}
	else
	{
		printf("%s : %s\n", title, lpMsgBuf);
	}
	// Free the buffer.
	LocalFree( lpMsgBuf );

}

#endif

AutoMalloc::AutoMalloc(size_t size)
{
	m_pointer = phmalloc(size);
}

AutoMalloc::~AutoMalloc()
{
	if(m_pointer)
		phmfree(m_pointer);
}

char *AutoMalloc::ch_ptr(void)
{
	return (char *) m_pointer;
}

// Gets current time and strips newline off
char* timestr(char *buf)
{
	static char buffer[32];
	time_t now = time(0);
	char *t = ctime(&now);

	if(!buf)
		buf = buffer;

	char *s = buf;

	*s++='[';
	while(*t && *t!= 0x0a && *t !=0x0d)
		*s++ = *t++;
	*s++ = ']';
	*s=0;
	return buf;
}

// Create a zero leading hex number with no zero terminator.
Boolean zntox(unsigned int val, char *szret, int len)
{
	int nchars;
	char szstack[16];
	register char *stack = szstack;

	for(nchars=0; val; nchars++)
	{
		register unsigned n = (unsigned) (val % sizeof(_modec_array));
		val /= sizeof(_modec_array);
		*stack++ = (char)_modec_array_u[n];
	} 

	if(nchars > len)
		return FALSE;

	while(nchars < len)
	{
		*szret++ = '0';
		nchars++;
	}
	while(stack > szstack)
		*szret++ = *--stack;
	return TRUE;
}
Boolean getparam(char *ret, char *templ, char *str, int maxlen)
{
	char *p = strstr(str, templ);

	if(p)
	{
		int len = strlen(templ);
		p+= len;

		while(!isspace(*p) && *p && --maxlen)
			*ret++ = *p++;

		*ret = 0;
		return TRUE;
	}
	return FALSE;
}

static void __ntoa_ (unsigned long val, char *szret, char *szstack, unsigned radix, int is_neg)
{
	if(radix > 1) /* radix must be at least 2 */
	{
		register char *stack = szstack;

		if (is_neg)
		{
			*szret++ = '-';
			val = (unsigned long)(-(long)val);
		}

		do
		{
			register unsigned n = (unsigned) (val % radix);
			val /= radix;
			*stack++ = (char)_modec_array_u[n];

		} while (val);

		/* Unwind stack into ascii string. */

		while(stack > szstack)
			*szret++ = *--stack;
	}
	*szret=0;
}


char * _itoa ( int val, char *szret, int radix)
{
	char szstack[64];
	if (radix == 10 && val < 0)
		__ntoa_((unsigned long)val, szret, szstack, radix, 1);
	else
		__ntoa_((unsigned long)(unsigned int)val, szret, szstack, radix, 0);
	return szret;
}

char *  _ltoa (long val,char *szret,int radix)
{
	char szstack[64];
	__ntoa_((unsigned long)val, szret, szstack, radix, (radix == 10 && val < 0));
	return szret;
}

char *  _ultoa (unsigned long val, char *szret, int radix)
{
	char szstack[64];
	__ntoa_(val, szret, szstack, radix, 0);
	return szret;
}

#define SEC_PER_DAY	(60*60*24)
#define SEC_PER_HOUR	(60*60)
#define SEC_PER_MINUTE	(60)

char *elapsed(char *result, time_t start, time_t end)
{
	char *s = result;
	int seconds = end - start;
	int days = seconds / SEC_PER_DAY;
	seconds -= (days * SEC_PER_DAY);
	int hours = seconds / SEC_PER_HOUR;
	seconds -= (hours * SEC_PER_HOUR);
	int minutes = seconds / SEC_PER_MINUTE;
	seconds -= (minutes * SEC_PER_MINUTE);

	if(days)
	{
		sprintf(s, "%d Days ", days);
		s +=strlen(s);
	}

	sprintf(s, "%d:%d:%d", hours, minutes, seconds);
	return result;
}
int MITuple::intsortad(void *p1, void *p2)
{
	return ((ITUPLE *)p1)->a - ((ITUPLE *)p2)->a;
}
int MITuple::intsortbd(void *p1, void *p2)
{
	return ((ITUPLE *)p1)->b - ((ITUPLE *)p2)->b;
}
int MITuple::intsorta(void *p1, void *p2)
{
	return ((ITUPLE *)p2)->a - ((ITUPLE *)p1)->a;
}
int MITuple::intsortb(void *p1, void *p2)
{
	return ((ITUPLE *)p2)->b - ((ITUPLE *)p1)->b;
}

void MITuple::Sorta(Boolean fDecending)
{
	if(fDecending)
		shell(m_array, m_Count, (COMPARE) intsortad);
	else
		shell(m_array, m_Count, (COMPARE) intsorta);
}

void MITuple::Sortb(Boolean fDecending)
{
	if(fDecending)
		shell(m_array, m_Count, (COMPARE) intsortbd);
	else
		shell(m_array, m_Count, (COMPARE) intsortb);
}


DynamicBuffer::DynamicBuffer(int cb)
{
	m_jump = 256;
	m_data = phmalloc(cb);
	m_cb = (m_data) ? cb : 0;
	memset(m_data,0,cb);
}

DynamicBuffer::~DynamicBuffer()
{
	if(m_data)
		phmfree(m_data);
}

void *DynamicBuffer::GetPointer(int len)
{
	if(len >= m_cb)
	{
		int cb = m_cb += m_jump;
		if(cb < len)
			cb = (len & ~7) + 8;
		void * p = phmrealloc(m_data, cb);

		if(!p) //Try to malloc
		{
			p = phmalloc(cb);
			if(p)
				memcpy(p, m_data, m_cb);
			phmfree(m_data);
		}
		m_cb = cb;
		m_jump += m_jump;
		m_data = p;
	}
	assert(m_data);
	return m_data;
}

void *DynamicBuffer::GetPointer(int n, int cb)
{
	return (void *) GetPointer(n*cb);
}

void *DynamicBuffer::GetZPointer(int cb)
{
	void *p = GetPointer(cb);
	if(p)
		memset(p,0,cb);
	return p;
}
char *DynamicBuffer::Strcpy(char *str)
{
	int cb = strlen(str)+1;
	char *p = (char *)GetPointer(cb);
	memcpy(p, str, cb);
	return p;
}

#ifdef POSTGRESQL_INC

char *DynamicBuffer::GetPointer(text *text)
{
	int len = VARSIZE(text)-VARHDRSZ;
	char *buffer = GetPointer(len+1);
       	memcpy(buffer, VARDATA(text), len);
       	buffer[len]=0;
       	return buffer;
}

char *DynamicBuffer::SizePointer(text *text)
{
	int len = VARSIZE(text)-VARHDRSZ;
	return GetPointer(len+1);
}
#endif

#ifdef WIN32
clock_t times(struct tms *t)
{
	return GetTickCount();
}
#endif


#ifdef ATOM_STUBS
static MUTEX s_atomic;

// DO NOT USE THESE IN PRODUCTION!!!
// They use a single mutex.
// It is best to write the functions 
// specifically for your platform.
void atom_inc(volatile int *num)
{
	MutexLock(&s_atomic);
	(*num)++;
	MutexUnlock(&s_atomic);
}
void atom_dec(volatile int *num)
{
	MutexLock(&s_atomic);
	(*num)--;
	MutexUnlock(&s_atomic);
}
int atom_xchg(volatile int *m, int intval)
{
	MutexLock(&s_atomic);
	int n = *m;
	*m = intval;
	MutexUnlock(&s_atomic);
	return n;
}
void atom_add(volatile int *m, int num)
{
	MutexLock(&s_atomic);
	*m += num;
	MutexUnlock(&s_atomic);
}
void atom_sub(volatile int *m, int num)
{
	MutexLock(&s_atomic);
	*m -= num;
	MutexUnlock(&s_atomic);
}
#endif

#ifdef GCC
#ifndef DARWIN
void atom_inc(volatile int *num)
{
        __asm__ __volatile__ ( "lock incl %0" : "=m" (*num));
}
void atom_dec(volatile int *num)
{
        __asm__ __volatile__ ( "lock decl %0" : "=m" (*num));
}
int atom_xchg(volatile int *m, int inval)
{
	register int val = inval;
	__asm__ __volatile__ ( "lock xchg %1,%0" : "=m" (*m), "=r" (val) : "1" (inval));
        return val;
}
void atom_add(volatile int *m, int inval)
{
	register int val = inval;
	__asm__ __volatile__ ( "lock add %1,%0" : "=m" (*m), "=r" (val) : "1" (inval));
}
void atom_sub(volatile int *m, int inval)
{
	register int val = inval;
	__asm__ __volatile__ ( "lock sub %1,%0" : "=m" (*m), "=r" (val) : "1" (inval));
}
#endif
#endif

#ifdef WIN32
void atom_inc(volatile int *num)
{
	_asm 
	{ 	
			mov	esi,	num
		lock	inc	DWORD PTR [esi] 
	};
}
void atom_dec(volatile int *num)
{
	_asm 
	{ 		mov	esi,	num
		lock	dec	DWORD PTR [esi] 
	};
}
int atom_xchg(volatile int *m, int inval)
{
	_asm 
	{ 
			mov	eax,	inval
			mov	esi,	m
		lock	xchg	eax,	DWORD PTR [esi]
			mov	inval,	eax 	
	}
	return inval;
}
void atom_add(volatile int *num, int val)
{
	_asm 
	{ 		mov	esi,	num
			mov	eax, 	val
		lock	add	DWORD PTR [esi], eax
	};
}
void atom_sub(volatile int *num, int val)
{
	_asm 
	{ 		mov	esi,	num
			mov	eax, 	val
		lock	sub	DWORD PTR [esi], eax
	};
}
#endif

