/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to commercially license this library, the author
    can be reached at info@mohawksoft.com
*/ 
// #define TRACE_DEBUG 1

#include "Phoenix.h"

MWWW::MWWW()
{
	TRACE_FUNCT("MWWW:MWWW");
	m_agent=MWWW_DEF_AGENT;
	m_accept=MWWW_DEF_ACCEPT;
	m_language=MWWW_DEF_LANG;
	m_charset=MWWW_DEF_CHAR;
	m_cookie = NULL;
	m_debug=0;
}

void MWWW::progress(int type)
{
	TRACE_FUNCT("progress");
	if(m_debug>1)
		mferr.mfprintf("%d\n",type);
}

void MWWW::SetCookie(char *str)
{
	TRACE_FUNCT("SetCookie");
	m_cookie = str;
}
void MWWW::SetDebug(int debug)
{
	TRACE_FUNCT("SetDebug");
	m_debug = debug;
}
void MWWW::SetAgent(char *str)
{
	TRACE_FUNCT("SetAgent");
	m_agent = str;
}

void MWWW::SetAccept(char *str)
{
	TRACE_FUNCT("SetAccept");
	m_accept = str;
}

void MWWW::SetLanguage(char *str)
{
	TRACE_FUNCT("SetLanguage");
	m_language = str;
}

void MWWW::SetCharSet(char *str)
{
	TRACE_FUNCT("GetCharSet");
	m_charset = str;
}

int MWWW::GetCode(void)
{
	TRACE_FUNCT("GetCode");
	return m_code;
}

int MWWW::SplitURL(char *url, char *hostname, char *path, char *protocol)
{
	TRACE_FUNCT("SplitURL");
	int cb;
	int port=80;
	char *pe;
	char *p = trim(url);
	char tmpproto[32];

	if(!p)		// All spaces!!
		return 0;

	if(!protocol)
		protocol = tmpproto;
		
	// Does url have protocol spec?
	if ( (pe = strstr(p, ":/")) )
	{
		pe+=3;
		cb = pe-p;
		memcpy(protocol, url, cb);
		protocol[cb]=0;
		if(strstr(protocol,"ftp"))
			port = 21;
		else if(strstr(protocol, "https"))
			port = 443;
		else if(strstr(protocol, "http"))
			port = 80; // pipe dream, ssl?
		p = pe; // skip over protocol spec.
	}
	else
	{
		strcpy(protocol, "http://");
	}
	
	// p now should point to host name!!!!
	// Does it have a path?
	if ( (pe = strchr(p, '/')) )
	{
		if(path)
			strcpy(path, pe);
		cb = pe - p;
	}
	else
	{
		if(path)
		{
			path[0]='/';
			path[1]=0;
		}
		cb = strlen(p);
	}

	memcpy(hostname, p, cb);
	hostname[cb]=0;

	pe = strchr(hostname, ':');

	if(pe) // the port was given explicitly
	{
		port = atoi(&pe[1]);
		*pe=0; // Nuke it because it is not part of the host
	}
	return port;
}

int MWWW::ReadResponse(MSock *msock, MRandomFile *mf, char *buf, int cbbuf, int retries)
{
	TRACE_FUNCT("ReadResponse");
	int cbtotal=0;

	do	
	{
		int ret = msock->ReadSelect(1,0);

		progress(-1);

		if(ret> 0)
		{
			
       			int cb=msock->read(buf, 1);
			TRACE_VAL(cb);

			if(cb < 0)
				break;	// Error condition
			else if(cb)
				cbtotal += mf->write(buf,cb);
			else
				break;
		}
		else if(ret == 0)
		{
			while(retries > 0)
			{
				TRACE_VAL(retries);
				ret = msock->ReadSelect(1);
				if(ret > 0)
					break;
				TRACE_VAL(ret);
				retries--;
			}
			if(ret <= 0)
				break;
		}
		if(retries <= 0)
			break;
		
	}while(1);

	TRACE_STR("Done");

	msock->Linger(TRUE,0);
	return cbtotal;
}

int MWWW::ParseCode(MRandomFile *mf, char *buf, int cbbuf)
{
	TRACE_FUNCT("ParseCode");
	int code=0;
	mf->rewind();
	if(mf->read(buf, cbbuf) > 0)
	{
		static char http[]="HTTP/";
		static char http2[]="2HTTP/";

		char *p = strstr(buf, http2);

		if(p)	// http2 tag!
		{
			p+=sizeof(http2)-1;
			m_http = atof(p);

			while(*p && !isspace(*p))
				p++;

			while(*p && isspace(*p))
				p++;

			code = atoi(p);
		}

		if(code != 200)
		{
			p = strstr(buf, http);
			if(p)
			{
				p+=sizeof(http)-1;
				m_http = atof(p);

				while(*p && !isspace(*p))
					p++;

				while(*p && isspace(*p))
					p++;

				code = atoi(p);
			}
		}
	}
	mf->rewind(); // Set to point to the beginning
	return code;
}

int MWWW::WaitResponse(MSock *msock, int seconds)
{
	TRACE_FUNCT("WaitResponse");
	int i;
	for(i=0; i < seconds; i++)	
	{
		TRACE_STR("Waiting");
		TRACE_VAL(i);
		if(msock->ReadSelect(1,0))
			break; // got something.
		progress(i);
	}
	return i;
}

// This function gets a URL from a site and fills a buffer.
int MWWW::Get(MRandomFile *mf, char *hostname, int port, char *uri) 
{
	TRACE_FUNCT("Get");
	char 		buf[2048];
	int		cbtotal;

	char fmt[]= 	"GET %s HTTP/1.0\n"
			"User-Agent: %s\n"
			"Host: %s\n"
			"Accept: %s\n"
			"Accept-Language: %s\n"
			"Accept-Charset: %s\n";

	char reqfmt[]=	"%sConnection: Keep-Alive\n\n";

	char cookiefmt[]= "Cookie: %s\n";

	MSock		msock(port);

	m_code =0;
#ifndef WIN32  
	gettimeofday(&m_start, NULL);
#endif
	msock.SetHostByName(hostname);

	if(!msock.Connect())
		return MWWW_NOCONNECT;

	sprintf(buf, fmt, uri, m_agent, hostname, m_accept, m_language, m_charset);

	if(m_cookie)
	{
		int cb=strlen(buf);
		sprintf(&buf[cb], cookiefmt, m_cookie);
	}

	if(m_debug > 1)
 		mferr.mfprintf(reqfmt, buf);

 	if(msock.mfprintf(reqfmt, buf)<(ssize_t)strlen(reqfmt))
		return MWWW_NOWRITE;

	if(WaitResponse(&msock, 120) >= 120)
		return MWWW_NODATA;

	cbtotal = ReadResponse(&msock, mf, buf, sizeof(buf));
#ifndef WIN32  
	gettimeofday(&m_stop, NULL);
#endif

	m_code = ParseCode(mf, buf, sizeof(buf));

  	return cbtotal;
}
// This function a URL from a site and fills a buffer.
int MWWW::Post(MRandomFile *mf, char *hostname, int port, char *uri, char *postdata) 
{
	TRACE_FUNCT("Post");
	char 		buf[2048];
	char 		buffer[2048];
	int		cbtotal;

	char fmtnc[]= 	"POST %s HTTP/1.0\n"
			"User-Agent: %s\n"
			"Content-Type: application/x-www-form-urlencoded\n"
			"Content-Length: 000\n"
			"Host: %s\n"
			"Accept: %s\n"
			"Accept-Language: %s\n"
			"Accept-Charset: %s\n\n"
			"%s";

	char fmtc[]= 	"POST %s HTTP/1.0\n"
			"User-Agent: %s\n"
			"Content-Type: application/x-www-form-urlencoded\n"
			"Content-Length: 000\n"
			"Host: %s\n"
			"Accept: %s\n"
			"Accept-Language: %s\n"
			"Accept-Charset: %s\n"
			"Cookie: %s\n\n"
			"%s";

	MSock		msock(port);

	char *fmt = (m_cookie) ? fmtc : fmtnc;


	m_code =0;
#ifndef WIN32
	gettimeofday(&m_start, NULL);
#endif

	msock.SetHostByName(hostname);

	if(!msock.Connect())
		return MWWW_NOCONNECT;

	if(m_cookie)
  		sprintf(buffer, fmt, uri, m_agent, hostname, m_accept, m_language, m_charset, m_cookie, postdata);
	else
  		sprintf(buffer, fmt, uri, m_agent, hostname, m_accept, m_language, m_charset, postdata);

	char cl[]="Content-Length: ";
	char *p = strstr(buffer, cl);

	assert(p);
	p+=(sizeof(cl)-1);

	char test[6];
	sprintf(test, "%03d", strlen(postdata));

	memcpy(p, test, 3);

	msock.write(buffer, strlen(buffer));

	if(WaitResponse(&msock, 120) >= 120)
		return MWWW_NODATA;

	cbtotal = ReadResponse(&msock, mf, buf, sizeof(buf));

#ifndef WIN32
	gettimeofday(&m_stop, NULL);
#endif

	m_code = ParseCode(mf, buf, sizeof(buf));

  	return cbtotal;
}
