#include <Phoenix.h>
#ifndef DARWIN
#include <malloc.h>
#endif
#include <string.h>
#include <assert.h>
#include <stdio.h>

static int numhandles=0;
static int total = 0;

#ifdef DEBUG_MEMORY
#define MEM_DBG
#define VALID_HANDLE	0x0A
#define FREE_HANDLE	0x05

#define GET_FLAG(X)	( (X) & 0x0F)
#define GET_VAL(X)	( (X) & ~0x0F)
#define ROUND_UP(X)	( GET_VAL(X) + 0x10)

void * _phmalloc(unsigned int cb)
{
	ATOMIC_INC(numhandles);
	unsigned int newcb = ROUND_UP(cb+sizeof(int));
	unsigned int *mem = (unsigned int *)malloc(newcb);
	if(mem)
	{
		mem[0] = newcb | VALID_HANDLE;
		ATOMIC_ADD(total, newcb);
		return (void *) &mem[1];
	}
	return NULL;
}
void _phmfree(void *p)
{
	if(p)
	{
		ATOMIC_DEC(numhandles);
		unsigned int *mem = (unsigned int *)p;
		mem--;
		int cb = GET_VAL(mem[0]);
		int flag = GET_FLAG(mem[0]);
		assert(flag == VALID_HANDLE);
		ATOMIC_SUB(total, cb);
		memset(mem, FREE_HANDLE, cb);
		assert(total >= 0);
		free(mem);
	}
}	
void *_phmrealloc(void *p, unsigned int cb)
{
	if(p)
	{
		unsigned int *mem = (unsigned int *) p;
		cb += sizeof(unsigned int);
		unsigned int newcb = (cb & ~0x0F) + 0x10;
		mem--;
		int oldcb = GET_VAL(*mem);
		int flag = GET_FLAG(*mem);

		assert(flag == VALID_HANDLE);
		ATOMIC_SUB(total, oldcb);
		mem = (unsigned int *)realloc((void*)mem,newcb);
		if(mem)
		{
			ATOMIC_ADD(total, newcb);
			mem[0] = newcb | VALID_HANDLE;
			return (void *)&mem[1];
		}
		return NULL;
	}
	else
		return _phmalloc(cb);
}
void *_phmcalloc(unsigned int num, unsigned int cbelement)
{
	unsigned int cb = num*cbelement;
	void *p = phmalloc(cb);
	memset(p, 0, cb);
	return p;
}


#else
void * _phmalloc(unsigned int cb)
{
	return malloc(cb);
}

void _phmfree(void *p)
{
	free(p);
}

void *_phmcalloc(unsigned int num, unsigned int cbelement)
{
	return calloc(num,cbelement);
}

void *_phmrealloc(void *p, unsigned int cbnew)
{
	return realloc(p,cbnew);
}
#endif


void *operator new (size_t size)
{
	return phmalloc(size);
}

void operator delete (void *mem)
{
	phmfree(mem);
}
char *_phmstrdup(char *str)
{
	int cb = strlen(str)+1;
	char *p = (char *)phmalloc(cb);
	memcpy(p, str, cb);
	return p;
}
int phmtotal(void)
{
	return total +(numhandles * 4);
}
int phmcount(void)
{
	return numhandles;
}

int phmsystotal(void)
{
#ifdef LINUX
	struct mallinfo mi = mallinfo();
	return mi.uordblks;
#else
	return phmtotal();
#endif
	
}

int phmfrag(void)
{
#ifdef LINUX
	struct mallinfo mi = mallinfo();
	return mi.fordblks;
#else
	return 0;
#endif
	
}


#ifdef DEBUG_MEMORY
void * phmalloc_dbg(unsigned int cb, char *file, int line)
{
	void *p = _phmalloc(cb);

	if(!p)
		mferr.mfprintf("phmalloc(%d) failed in %d %s\n", cb, line, file);
	return p;
}
void phmfree_dbg(void *p, char *file, int line)
{
	if(!p)
		mferr.mfprintf("phmfree(%X) with NULL pointer in %d %s\n", p, line, file);
	else
	{
		unsigned int *mem = (unsigned int *) p;
		if((mem[-1] & 0x0F) == FREE_HANDLE)
			mferr.mfprintf("phmfree(%X) called with free pointer %d %s\n", p, line, file);
		else if((mem[-1] & 0x0F) != VALID_HANDLE)
			mferr.mfprintf("phmfree(%X) called without cookie %d %s\n", p, line, file);
		_phmfree(p);
	}
}

void *phmrealloc_dbg(void *p, unsigned int cb, char *file, int line)
{
	if(p)
	{
		unsigned int *mem = (unsigned int *) p;
		if((mem[-1] & 0x0F) != VALID_HANDLE)
			mferr.mfprintf("phmrealloc(%X, %d) called without cookie %d %s\n", p, cb,line, file);
	}
	p = _phmrealloc(p,cb);

	if(!p)
		mferr.mfprintf("phmrealloc(%X, %d) failed in %d %s\n", p, cb, line, file);

	return p;
}

void *phmcalloc_dbg(unsigned int num, unsigned int cbelement, char *file, int line)
{
	void *p = _phmcalloc(num,cbelement);
	if(!p)
		mferr.mfprintf("phmcalloc(%d, %d) failed %d %s\n", num, cbelement, line, file);
	return p;
}
char *phmstrdup_dbg(char *str, char *file, int line)
{
	if(!str)
		mferr.mfprintf("phmstrdup(%X) called with NULL %d %s\n", str, line, file);
	char *p = _phmstrdup(str);

	if(!p)
		mferr.mfprintf("phmstrdup(%X) failed %d %s\n", str, line, file);
	return p;
}

#endif

