/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to professionally license this library, the author
    can be reached at info@mohawksoft.com
*/ 

// NOTE: Make sure to check semaphore locking!!!
// #define TRACE_DEBUG 1

#include "session.h"
#include "reqclient.h"

unsigned int	g_defReqFmt = REQFMT_ASCII;

extern "C" REQB *AllocateRequestBuffer(unsigned  size)
{
	TRACE_FUNCT("AllocateRequestBuffer");
	unsigned int cb = sizeof(REQB) + size;
	REQB *reqb = (REQB *) phmalloc(cb);
	assert(reqb);
	reqb->size = size;
	reqb->type = REQB_DYNAMIC;
	return reqb;
}
extern "C" void FreeRequestBuffer(REQB *req)
{
	TRACE_FUNCT("FreeRequestBuffer");
	phmfree((void *)req);
}

extern "C" REQB *SizeRequestBuffer(REQB *reqb, unsigned int size)
{
	TRACE_FUNCT("SizeRequestBuffer");
	TRACE_VAL(size);

	if(reqb->size >= size)
		return reqb;

	if(reqb->type != REQB_DYNAMIC)
	{
		TRACE_STR("Tried to resize a static request buffer bigger than allocated");	
		return NULL;
	}

	unsigned int cb = sizeof(REQB) + size;

	REQB *reqbT = (REQB *)phmrealloc(reqb, cb);

	if(!reqbT)
	{
		TRACE_STR("realloc failed, creating new request buffer");
		reqbT = AllocateRequestBuffer(size);
		memcpy((void *)&reqbT->req, (void *)&reqb->req, reqb->size);
		phmfree(reqb);
	}
	return reqbT;
}

// Bottom Part of ReadRequest
// After the request hdr is read/translated, the rest of the request
// is read from the scoket
static Boolean ReadDatum(REQB **ppreq, MSock *sock, int timeout)
{
	TRACE_FUNCT("ReadDatum");
	REQB *preq = *ppreq;
	preq = SIZEREQB(preq, preq->req.len);

	// Fatal error
	assert(preq);

	*ppreq = preq;

	if(preq->req.len > 0)
	{
		if(sock->retryread((void *)preq->req.datum, preq->req.len) != preq->req.len)
		{
			mferr.mfprintf("Packet too short\n");
			preq->req.stat = REQ_ERR;
			preq->req.param = REQE_BADLEN;
			return FALSE;
		}
	}
	return TRUE;
}

// Middle part of ReadRequestTimout
// Checks for a binary packet, if it is there, it uses it.
static Boolean ReadBinaryRequest(REQB **ppreq, MSock *sock, int timeout)
{
	TRACE_FUNCT("ReadBinaryRequest");
	if((*ppreq)->req.version != REQV_VERSION && (*ppreq)->req.version != REQV_VMULTI)
	{
		TRACE_STR("Not a binary packet");
		return FALSE;
	}

	(*ppreq)->fmt = REQFMT_BINARY;
	
	return ReadDatum(ppreq, sock, timeout);
}

// Middle part of ReadRequestTimout
// Checks for an ascii packet, if it is there, converts it, then uses it.
static Boolean ReadAsciiRequest(REQB **ppreq, MSock *sock, int timeout)
{
	TRACE_FUNCT("ReadAsciiRequest");
	int version;

	if(strncmp((*ppreq)->reqsz.version, REQVSZ_VERSION, 4) == 0)
		version = REQV_VERSION;
	else if(strncmp((*ppreq)->reqsz.version, REQVSZ_MULTI, 4) == 0)
		version = REQV_VMULTI;
	else
	{
		TRACE_STR("Not an ascii packet");
		return FALSE;
	}


#ifdef TRACE_DEBUG
	int len = sizeof(REQSZ);
	char test[len+1];
	memcpy(test, &(*ppreq)->req, len);
	test[len]=0;
	TRACE_STRSTR("Ascii Request:", test);
#endif
	(*ppreq)->fmt = REQFMT_ASCII;

	// Change in place
	// If sizeof(int) is greater than 4 bytes, this could
	// be a problem.
	(*ppreq)->req.version = version;
	(*ppreq)->req.stat 	  = modectoin((*ppreq)->reqsz.stat,4);
	(*ppreq)->req.len     = modectoin((*ppreq)->reqsz.len,4);
	(*ppreq)->req.session = modectoin((*ppreq)->reqsz.session,4);
	(*ppreq)->req.name    = modectoin((*ppreq)->reqsz.name,4);
	(*ppreq)->req.value   = modectoin((*ppreq)->reqsz.value,4);
	(*ppreq)->req.param   = modectoin((*ppreq)->reqsz.param,4);

	return ReadDatum(ppreq, sock, timeout);
}
// Converts binary request to ascii
Boolean ReqToAscii(REQB **ppreq)
{
	TRACE_FUNCT("ReqToAscii");
	REQB *preq = *ppreq;


	if(preq->req.version == REQV_VERSION)
		memcpy(preq->reqsz.version, REQVSZ_VERSION,4);
	else if(preq->req.version == REQV_VMULTI)
		memcpy(preq->reqsz.version, REQVSZ_MULTI,4);
	else
		return FALSE;
		
	zntox(preq->req.stat,preq->reqsz.stat,4);
	zntox(preq->req.len,preq->reqsz.len,4);
	zntox(preq->req.session,preq->reqsz.session,4);
	zntox(preq->req.name,preq->reqsz.name,4);
	zntox(preq->req.value,preq->reqsz.value,4);
	zntox(preq->req.param,preq->reqsz.param,4);

#ifdef TRACE_DEBUG
	int len = sizeof(REQSZ);
	char test[len+1];
	memcpy(test, &preq->req, len);
	test[len]=0;
	TRACE_STRSTR("Ascii Request:", test);
#endif
	return TRUE;
}

extern "C" int ReadRequestTimeout(REQB **ppreq, MSock *sock, int timeout)
{
	int cb;
	TRACE_FUNCT("ReadRequestTimeout");
	if(sock->ReadSelect(timeout,0) <= 0)
	{
		// This is a recoverable error, I hope
		TRACE_STR("Timeout in ReadRequestTimeout");
		(*ppreq)->req.stat = REQ_ERR;
		(*ppreq)->req.param = REQE_BADCONN;
		return (*ppreq)->req.stat;
	}
	
	cb = sock->read((void *)&(*ppreq)->req, sizeof(REQ));

	if(cb <= 0)
	{
		// This is a recoverable error, I hope
		TRACE_STR("Server closed the socket ReadRequestTimeout");
		(*ppreq)->req.stat = REQ_ERR;
		(*ppreq)->req.param = REQE_BADCONN;
		return (*ppreq)->req.stat;
	}
	else if(cb != sizeof(REQ))
	{
		TRACE_STR("Read REQ header failed");
		(*ppreq)->req.stat = REQ_ERR;
		(*ppreq)->req.param = REQE_BADHDR;
		return (*ppreq)->req.stat;
	}

	if(!ReadBinaryRequest(ppreq, sock, timeout))
	{
		if(!ReadAsciiRequest(ppreq,sock,timeout))
		{
			TRACE_STR("Bad Version in read");
			(*ppreq)->req.stat = REQ_ERR;
			(*ppreq)->req.param = REQE_BADVER;
			return (*ppreq)->req.stat;
		}
	}
	return (*ppreq)->req.stat;
}
extern "C" int ReadRequest(REQB **ppreq, MSock *sock)
{
	TRACE_FUNCT("ReadRequest");
	return ReadRequestTimeout(ppreq,sock,30);
}

extern "C" int WriteRequest(REQB **ppreq, MSock *sock)
{
	TRACE_FUNCT("WriteRequest");
	TRACE_VAL((*ppreq)->req.len);
	int cb=(*ppreq)->req.len + sizeof(REQ);

	if((*ppreq)->fmt == REQFMT_ASCII)
		ReqToAscii(ppreq);

	return (sock->write((void *)&(*ppreq)->req, cb)==cb) ? TRUE : FALSE;
}

extern "C" int DoSingleRequest(char *hostname, int port, REQB **ppreq)
{
	TRACE_FUNCT("DoSingleRequest");
	MSock sock(port);
	sock.SetHostByName(hostname);

	if(sock.Connect())
	{
		WriteRequest(ppreq, &sock);
		ReadRequest(ppreq, &sock);
	}
	else
	{
		REQB *preq = *ppreq;
		preq->req.stat = REQ_ERR;
		preq->req.param = REQE_NOCONNECT;
		
	}
	return ((*ppreq)->req.stat == REQ_OK) ? REQE_NOERROR : (*ppreq)->req.param;
}

REQB *StaticRequestBuffer(char *buffer, unsigned int cb)
{
	REQB *reqb = (REQB *) buffer;
	reqb->type = REQB_STATIC;
	reqb->size = cb - sizeof(REQB);
	return reqb;
}
extern "C" Boolean OpenReqSock(REQB **reqb, MSock *sock, char *hostname)
{
	TRACE_FUNCT("OpenReqSock");
	int retries=3;
	sock->SetHostByName(hostname);
	Boolean fConnected=FALSE;

	while(fConnected == FALSE && retries-- != 0 )
		fConnected = sock->Connect();

	if(fConnected)
	{
		FormatRequest(reqb, REQ_POPEN, "","","",30);
		WriteRequest(reqb, sock);
		ReadRequest(reqb, sock);

		if((*reqb)->req.stat == REQ_OK)
			return TRUE;
	}
	sock->Close();
	return FALSE;
}
	
extern "C" void *OpenReqConn(char *hostname, int port)
{
	TRACE_FUNCT("OpenReqConn");
	STATIC_REQB(1024);

	MSock *sock= new MSock(port);

	if(OpenReqSock(&preq, sock, hostname))
		return sock;
	else
	{
		delete sock;
		return NULL;
	}
	return NULL;
}
extern "C" Boolean ReopenReqConn(void *conn)
{
	TRACE_FUNCT("ReopenReqConn");
	STATIC_REQB(1024);
	MSock *sock = (MSock *) conn;
	int retries=3;
	Boolean fConnected=FALSE;

	while(fConnected == FALSE && retries-- != 0 )
		fConnected = sock->Reconnect();

	if(fConnected)
	{
		FormatRequest(&preq, REQ_POPEN, "","","",0);
		WriteRequest(&preq, sock);
		ReadRequest(&preq, sock);
		if(preq->req.stat == REQ_OK)
			return TRUE;
	}
	sock->Close();
	return FALSE;
}
extern "C" void CloseReqSock(REQB **reqb, MSock *sock)
{
	TRACE_FUNCT("CloseReqSock");
	FormatRequest(reqb, REQ_PCLOSE, "","","",0);
	WriteRequest(reqb, sock);
}

extern "C" void CloseReqConn(void *conn)
{
	TRACE_FUNCT("CloseReqConn");
	STATIC_REQB(1024);
	if(conn)
	{
		MSock *sock = (MSock *) conn;
		CloseReqSock(&preq, sock);
		delete sock;
	}
}

extern "C" void DeleteReqConn(void *conn)
{
	TRACE_FUNCT("DeleteReqSock");
	MSock *sock = (MSock *) conn;
	delete sock;
}

extern "C" int DoRequest(void *conn, REQB **ppreq)
{
	TRACE_FUNCT("DoRequest");
	MSock *sock = (MSock *) conn;
	if(sock)
	{
		if(WriteRequest(ppreq, sock))
		{
			ReadRequest(ppreq, sock);
		}
		else
		{
			(*ppreq)->req.stat = REQ_ERR;
			(*ppreq)->req.param = REQE_NOSEND;
		}
	}
	else
	{
		(*ppreq)->req.stat = REQ_ERR;
		(*ppreq)->req.param = REQE_BADCONN;
	}
	return ((*ppreq)->req.stat == REQ_OK) ? REQE_NOERROR : (*ppreq)->req.param;
}

extern "C" char *ReqbErr(REQB *reqb)
{
	if(reqb->req.stat != REQ_ERR)
		return "No error";
	else if(reqb->req.len == 0)
		sprintf(reqb->req.datum, "Unknown Error (%d), no text provided from server", reqb->req.param);
	return reqb->req.datum;
}
/** This is depricated */
extern "C" char *ReqErr(int param)
{
	return "ReqErr() is depricated in this version of Phoenix, upgrage and use ReqbErr()";
}

extern "C" int FormatRequest(REQB **ppreq, int stat, char *session, char *name, char *value, int param)
{
	TRACE_FUNCT("FormatRequest");

	int cbsession	= (session)	? strlen(session)+1 : 0;
	int cbname	= (name)	? strlen(name)+1 : 0;
	int cbvalue	= (value)	? strlen(value)+1 : 0;
	int len 	= cbname + cbvalue + cbsession;

	REQB *preq = SIZEREQB((*ppreq), len);
	
	if(preq)
	{
#ifdef TRACE_DEBUG
		mferr.mfprintf("FormatRequest:[%s][%s][%s][%d]\n", session,name,value,param);
#endif
		*ppreq = preq;

		preq->fmt = g_defReqFmt;
		preq->req.version= REQV_VERSION;
		preq->req.stat 	= stat;
		preq->req.len 	= cbsession + cbname + cbvalue;
		preq->req.session= 0;
		preq->req.name 	= preq->req.session + cbsession;
		preq->req.value = preq->req.name + cbname;
		preq->req.param = param;

		if(cbsession)
			strcpy(&preq->req.datum[preq->req.session], session);
		if(cbname)
			strcpy(&preq->req.datum[preq->req.name], name);
		if(cbvalue)
			strcpy(&preq->req.datum[preq->req.value], value);
	}
	return len;
}

extern "C" int FormatRequestMulti(REQB **ppreq, int stat, char *session, int n, char **pairs, int param)
{
	TRACE_FUNCT("FormatRequest");
	REQB *preq = *ppreq;

	int cbsession= (session) ? strlen(session)+1 : 0;

	if(preq)
	{
		preq->fmt = g_defReqFmt;
		preq->req.version= REQV_VMULTI;
		preq->req.stat = stat;
		preq->req.len= 0;
		preq->req.session = 0;
		preq->req.name 	= cbsession;
		preq->req.value = 0;
		preq->req.param = param;

		int ndx = cbsession;
		int pi=0;
		int cb = cbsession;
		int i=0;

		if(cbsession)
			memcpy(&preq->req.datum[preq->req.session], session, cbsession);

		for(i=0, pi=0; i < n; i++)
		{
			cb += strlen(pairs[pi++])+1;
			cb += strlen(pairs[pi++])+1;
		}
		
		preq = SIZEREQB(preq, cb);
		if(!preq)
			return 0;
		else
			*ppreq = preq;

		for(i=0, pi=0; i < n; i++)
		{
			char *name = pairs[pi++];
			char *value = pairs[pi++];

			if(!name || !value)
				break;
#ifdef TRACE_DEBUG
			mferr.mfprintf("Set: %s = %s\n", name, value);
#endif
			int cbname = strlen(name)+1;
			int cbval = strlen(value)+1;

			cb += (cbname+cbval);
			
			memcpy(&preq->req.datum[ndx], name, cbname);
			ndx += cbname;

			memcpy(&preq->req.datum[ndx], value, cbval);
			ndx+=cbval;

			preq->req.datum[ndx]=0;
			preq->req.value++;
		}	
		preq->req.len = ndx;
	}
	return preq->req.len;
}

extern "C" int FormatRequestStrings(REQB **ppreq, int stat, char *session, int n, char **strings)
{
	TRACE_FUNCT("FormatRequest");
	REQB *preq = *ppreq;

	int cbsession= (session) ? strlen(session)+1 : 0;

	if(preq)
	{
		preq->fmt = g_defReqFmt;
		preq->req.version= REQV_VMULTI;
		preq->req.stat = stat;
		preq->req.len= 0;
		preq->req.session = 0;
		preq->req.name 	= cbsession;
		preq->req.value = 0;
		preq->req.param = n;

		int ndx = cbsession;
		int pi=0;
		int cb = cbsession;
		int i=0;

		if(cbsession)
			memcpy(&preq->req.datum[preq->req.session], session, cbsession);

		for(i=0, pi=0; i < n; i++)
			cb += strlen(strings[pi++])+1;
		
		preq = SIZEREQB(preq, cb);
		if(!preq)
			return 0;
		else
			*ppreq = preq;

		for(i=0, pi=0; i < n; i++)
		{
			char *value = strings[pi++];

			if(!value)
				break;

			int cbval = strlen(value)+1;

			memcpy(&preq->req.datum[ndx], value, cbval);
			ndx+= cbval;
			cb += cbval;

			preq->req.datum[ndx]=0;
		}	
		preq->req.len = ndx;
	}
	return preq->req.len;
}

