/*
    Mohawk Software Framework by Mohawk Software
    Copyright (C) 1998-2001 Mark L. Woodward
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
 
    If you want support or to commercially license this library, the author
    can be reached at info@mohawksoft.com
*/ 
// #define TRACE_DEBUG

#include "Phoenix.h"
// #include <sys/time.h>
// #include <sys/types.h>
// #include <unistd.h>


MMTServer::MMTServer(MSock *listener)
{
	TRACE_FUNCT("MMTServer::MMTServer");
	FD_ZERO(&m_fdsetSocket);
	FD_ZERO(&m_fdsetRead);

	m_count=0;
	m_lastSock = 0;
	m_nsock = 0;
	m_nmaxSock=0;
	m_fWait = FALSE;

	m_sock = listener;
	m_fContinue = TRUE;
}

// LOCK MUTEX BEFORE calling
void MMTServer::RemoveSocket(int sock)
{
	TRACE_FUNCT("MMTServer::RemoveSocket");
	m_nsock--;
	FD_CLR(sock, &m_fdsetSocket);

	// This is our maximum socket
	// Lets find our next highest
	// This is needed for select
	if(sock == m_nmaxSock)
	{
		for(int i=sock-1; i; i--)
		{
			if(FD_ISSET(i, &m_fdsetSocket))
			{
				m_nmaxSock = i;
				break;
			}
		}
	}
}

// LOCK MUTEX BEFORE calling
void MMTServer::AddSocket(int sock)
{
	TRACE_FUNCT("MMTServer::AddSocket");
	m_nsock++;
	FD_SET(sock, &m_fdsetSocket);

	if(sock > m_nmaxSock)
		m_nmaxSock = sock;
}
// This loop accepts sockets and puts them into the 
// Socket set.
//
void MMTServer::threadLoop(void)
{
	TRACE_FUNCT("MMTServer::threadLoop");
	// Just use a sock puppet to use the template version of msocket's Accept
	MSock SockPuppet;
	while(m_fContinue)
	{
		int ret = m_sock->ReadSelect(1);

		if(ret < 0)	// Error
			break; 

		if(ret == 0)
			continue;
		
		if(!m_sock->Accept(&SockPuppet))
			continue;

		m_mutex.lock();
		AddSocket(SockPuppet.GetRawSock());
		m_mutex.unlock();
		m_mutex.signal();
#ifdef TRACE_DEBUG
		mferr.mfprintf("Socket? [%d][%d][%d]\n", m_nsock, m_nmaxSock, m_count);
#endif
	}
}

// This API returns a socket to the calling application.
int MMTServer::GetNextSocket(void)
{
	TRACE_FUNCT("MMTServer::GetNextSocket");
	int sock=-1;

	while(m_fContinue && sock == -1)
	{
		if(m_fWait && m_fContinue)
			YieldTimeSlice();

		if(!m_fContinue)
			break;

		m_mutex.lock();

		// Nothing sitting in fdset, select again
		if(m_count <= 0)
		{
			// If there are no sockets, lets wait for some
			if(m_nsock == 0)
			{
				m_fWait = TRUE;
				m_mutex.wait();
				m_fWait = FALSE;
			}

			timeval	to; 
			m_fdsetRead = m_fdsetSocket;
			to.tv_sec = 0; 
			to.tv_usec = 100;

			int count = select(m_nmaxSock+1, &m_fdsetRead, NULL, NULL, &to);

			if(count > 0)
			{
				m_lastSock=0;
				m_count = count;
			}
		}

		// THese is stuff in the fdset, grab the next one
		if (m_count > 0)
		{
			while(m_lastSock <= m_nmaxSock)
			{
				int sockT = m_lastSock++;
				if(FD_ISSET(sockT, &m_fdsetRead))
				{
					m_count--;
					sock = sockT;
					RemoveSocket(sock);
					break;
				}
			}
			if(sock == -1)
				m_count = 0;
		}
		m_mutex.unlock();
	}

	return sock;
}

// Server thread calls this for each request.
MSock *MMTServer::Accept(MSock *ms)
{
	TRACE_FUNCT("MMTServer::Accept");
	int sock = GetNextSocket();

	if(sock != -1)
	{
		ms->SetRawSock(sock);
		return ms;
	}
	return NULL;
}

// Server calls this when it has finished with the request, but NOT with the socket.
void MMTServer::Release(MSock *ms)
{
	TRACE_FUNCT("MMTServer::Release");
	m_mutex.lock();
	AddSocket(ms->GetRawSock());
	m_mutex.unlock();
	
}
// Server calls this when the socket should be closed.
void MMTServer::Finish(MSock *ms)
{
	TRACE_FUNCT("MMTServer::Finish");
	ms->Close();
}
